//	KaleidoTileGraphics.c
//
//	Drawing code for KaleidoTile.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "KaleidoTile-Common.h"
#include <math.h>

#ifdef __APPLE__
#include <OpenGL/gl.h>
#endif
#ifdef __WIN32__
#include "GeometryGames-Win32-OpenGLEntryPoints.h"
#endif

//	The near clipping distance is 1/INVERSE_NEAR_CLIP.
#define INVERSE_NEAR_CLIP		128.0

//	How many times does the background texture repeat in each direction?
#define TEX_REPS				4

//	When we subdivide a quad into an n×n grid,
//	how big an n should we allow for?
#define MAX_SUBDIVISION			16

//	How small must a face be before we revert to default texture coordinates?
#define MIN_TEXURE_AXIS_LENGTH	0.001

//	How much of the 256-unit wide control panel should
//	the fundamental triangle fill?
#define TRIANGLE_PANEL_WIDTH	224.0

//	Set the scale for the little circle markers in the control panel.
#define TRIANGLE_MARKER_SIZE	0.08
#define COLOR_MARKER_SIZE		0.015


//	Each Vertex Buffer Object (VBO) will contain
//	the following data for each of its vertices.
//	The fundamental triangle will comprise three meshes,
//	each a quadrilateral.
typedef struct
{
	float	pos[3],	//	position (x,y,z)
			nor[3],	//	normal vector (nx,ny,nz)
			tex[2];	//	texture coordinates (u,v)
} VBOData;

//	Define a rectangle with floating point coordinates.
//	The origin sits at the lower-left.
typedef struct
{
	float	left,
			bottom,
			right,
			top;
} RectFloat;

//	How is the control panel laid out?
static const RectFloat
	gSymmetryPanel		= { 16.0, 932.0, 240.0, 972.0},
	gLabelText[5]		= {
							{  0.0, 980.0, 256.0, 1012.0},
							{  0.0, 884.0, 256.0,  916.0},
							{  0.0, 787.0, 256.0,  819.0},
							{  0.0, 532.0, 256.0,  564.0},
							{  0.0, 138.0, 256.0,  170.0}
						  },
	gSymmetryButton[5]	= {
							{ 16.0, 932.0,  56.0, 972.0},
							{ 62.0, 932.0, 102.0, 972.0},
							{108.0, 932.0, 148.0, 972.0},
							{154.0, 932.0, 194.0, 972.0},
							{200.0, 932.0, 240.0, 972.0}
						  },
	gStylePanel			= { 76.0, 836.0, 180.0, 876.0},
	gStyleButton[2]		= {
							{ 76.0, 836.0, 116.0, 876.0},
							{140.0, 836.0, 180.0, 876.0}
						  },
	gTrianglePanel		= {  0.0, 560.0, 256.0, 816.0},
	gFacePanel[4]		= {
							{  0.0, 414.0, 256.0, 524.0},
							{  0.0, 296.0, 256.0, 406.0},
							{  0.0, 178.0, 256.0, 288.0},
							{  0.0,  20.0, 256.0, 130.0}
						  },
	gFaceButton[5]		= {	//	in local gFacePanel coordinates
							{ 28.0, 91.0,  44.0, 107.0},
							{ 52.0, 91.0,  68.0, 107.0},
							{ 76.0, 91.0,  92.0, 107.0},
							{100.0, 91.0, 116.0, 107.0},
							{124.0, 91.0, 140.0, 107.0}
						  },
	gFacePaintWell		= { 28.0, 3.0, 228.0, 83.0};	//	in local gFacePanel coordinates


//	As we tile outwards, keep track of the region tiled so far
//	by maintaining its perimeter as circular doubly-linked list.
//	Each PerimeterSegment represents a single edge of a single triangle.
typedef struct PerimeterSegment
{
	//	Is the image of the fundamental triangle 
	//	positively or negatively oriented?
	bool					itsImageIsReflected;	//	true if negatively oriented
	
	//	Which side of the fundamental triangle does
	//	this PerimeterSegment represent?
	unsigned int			itsSideIndex;			//	= 0, 1 or 2
	
	//	Viewed from within the tiled region looking outwards,
	//	each PerimeterSegment connects to a neighbor on its left
	//	and a neighbor on its right.
	struct PerimeterSegment	*itsLeftNeighbor,
							*itsRightNeighbor;

	//	"Free valence" means how many additional images 
	//	of the fundamental triangle a given vertex may accommodate.
	//	For example, in the (2,3,7) triangle group,
	//	each image of the vertex of (rotational) order 3
	//	will be surrounded by a total of 3 + 3 = 6 images
	//	of the fundamental triangle (3 positively oriented
	//	and 3 negatively oriented).  If, at a given stage of the tiling,
	//	only 2 of those 6 potential images have been tiled,
	//	then that image of the vertex has valence 2
	//	and free valence 6 - 2 = 4.
	//
	//	What is the free valence of the vertex on the left
	//	(with "left" defined as above, viewing the PerimeterSegment
	//	from within the tiled region looking outwards) ?
	unsigned int			itsLeftFreeValence;
	
	//	Where is the image of the fundamental triangle
	//	to which this PerimeterSegment belongs?
	ktMatrix				itsPlacement;
	
	//	If the tiling algorithm cannot continue past
	//	this PerimeterSegment without placing an image
	//	of the fundamental triangle beyond the allowable radius,
	//	mark this PerimeterSegment as blocked.
	//	A blocked PerimeterSegment remains part of the perimeter,
	//	even though the tiling algorithm does not attempt to push
	//	beyond it.  The tiling algorithm will terminated when
	//	the perimeter contains no more unblocked PerimeterSegments.
	bool					itsBlockedFlag;

} PerimeterSegment;


static void				MakeTriangleVBOs(ModelData *md);
static ktTilingStyle	GetModifiedTilingStyle(ModelData *md);
static void				ChooseTextureCoordinatesAtCorners(ktVector aCornerLocations[4], ktGeometry aGeometry, double aCornerTextureCoords[4][2]);
static void				BindTriangleVBOs(unsigned int aVertexBufferName, unsigned int anIndexBufferName, bool aShadersFlag, ShaderIndex aShaderIndex);
static void				DrawTriangleVBOs(unsigned int aMeshSize);
static void				RenderContent(ModelData *md, EyeType aGeoWallEye);
static void				RenderMainPanelBackground(ModelData *md);
static void				RenderMainPanelTiling(ModelData *md, EyeType aGeoWallEye);
static void				SetProjectionMatrix(EyeType anEyeType, ModelData *md);
static void				DrawTheScene(ModelData *md);
static void				DrawTheSceneWithParity(ModelData *md, ktMatrix aBaseTrianglePlacement, bool aDrawPositiveImagesFlag, bool aDrawNegativeImagesFlag, bool aBaseTriangleOnlyFlag);
static void				SendTransformationsToShader(ModelData *md, bool aDirectionalLightingFlag);
static void				DrawTiling(ModelData *md, ktMatrix aBaseTrianglePlacement, double anInradius, double aDomainRadius, double aCoshTilingRadius, 
							unsigned int aMeshSize, bool aDrawPositiveImagesFlag, bool aDrawNegativeImagesFlag, bool aDirectionalLightingFlag);
static bool				TrianglePlacementIsCloseEnough(ktVector aTriangleCenter, double aDomainRadius, double aCoshTilingRadius);
static void				RenderArchimedeanName(ModelData *md);
static void				RenderControlPanel(ModelData *md);
static void				SetRectangleVertices(float aVertexArray[4][3], const RectFloat *aRectangle);
static void				SetTextureCoordinates(float aTexCoordArray[4][2], float aLeftSide, float aBottomSide, float aRightSide, float aTopSide);
static void				SetBevelVertices(float aVertexPointer[4][3], RectFloat aRect, unsigned int aWidth);
static void				HSLAtoRGBA(ColorHSLA *anHSLA, ColorRGBA *anRGBA);
static void				ClipToUnitSquare(RectFloat *p, RectFloat *u);
static bool				PointInRectangle(const double h, const double v, const RectFloat *aRect);


char *GetRendererName(void)
{
	return (char *)glGetString(GL_RENDERER);
}


void SetTextureImageFancy(
	unsigned int	aTextureName,
	unsigned int	aWidth,
	unsigned int	aHeight,
	unsigned int	aDepth,			//	Pass 4 for RGBA or 1 for alpha-only.
	Byte			*aPixelArray,	//	4 bytes per pixel or 1 byte per pixel
	bool			aShadersFlag)
{
	unsigned int	thePixelFormat;

	if (aTextureName == 0
	 || ! IsPowerOfTwo(aWidth )
	 || ! IsPowerOfTwo(aHeight)
	 || (aDepth != 1 && aDepth != 4))
		return;

	//	Select an appropriate pixel format.
	if (aDepth == 4)	//	RGBA texture
	{
		thePixelFormat = GL_RGBA;
	}
	else				//	alpha-only texture
	{
		//	Different versions of OpenGL support
		//	alpha-only textures in different ways.
		//
		//	OpenGL 1.1 expects a texture with internal format GL_ALPHA,
		//	and the fixed-function pipeline handles it correctly.
		//
		//	OpenGL 3.0 doesn't support GL_ALPHA in a forward-compatible context.
		//	Instead we may create a 1-component texture with internal format GL_RED,
		//	and let the shader interpret the sample's red component as an opacity.
		//
		//	OpenGL 2.0 offers a choice of GL_ALPHA as in OpenGL 1.1
		//	or a 1-component texture as in OpenGL 3.0, but with the catch that
		//	we must use internal format GL_LUMINANCE instead of GL_RED.
		//	We opt for GL_LUMINANCE over GL_ALPHA, so the fragment shader
		//	code may remain the same for OpenGL 2.0 and OpenGL 3.0.
		//
		if (aShadersFlag)
		{
			if (RUNNING_OPENGL_3)
				thePixelFormat = GL_RED;
			else
				thePixelFormat = GL_LUMINANCE;
		}
		else
		{
//			thePixelFormat = GL_ALPHA;
			thePixelFormat = GL_INTENSITY;
		}
	}
	
	glBindTexture(GL_TEXTURE_2D, aTextureName);
	
	//	glTexImage2D()'s default 4-byte row alignment is fine for all RGBA textures
	//	as well as for alpha-only textures with power-of-two width at least 4,
	//	but we must relax the row alignment for alpha-only textures of width 1 or 2.
	if (aDepth == 1 && aWidth < 4)
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	//	When completing the migration to OpenGL 3.0 and OpenGL ES 2.0,
	//	eliminate this call to glTexParameteri(…GL_GENERATE_MIPMAP…)
	//	in favor of the glGenerateMipmap() call below.
	if ( ! RUNNING_OPENGL_3)
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);

	glTexImage2D(	GL_TEXTURE_2D,
					0,
					thePixelFormat,
					aWidth,
					aHeight,
					0,
					(thePixelFormat == GL_INTENSITY ? GL_LUMINANCE : thePixelFormat),
					GL_UNSIGNED_BYTE,
					aPixelArray);

	//	To average two pixels correctly when generating mipmaps,
	//	we need to weight them according to their alpha values.
	//	With traditional (R,G,B,α) the formula is a bit messy
	//
	//							 α0*(R0,G0,B0) + α1*(R1,G1,B1)
	//		(Ravg, Gavg, Bavg) = -----------------------------
	//							            α0 + α1
	//
	//							 α0 + α1
	//		              αavg = -------
	//							    2
	//
	//	With premultiplied (αR,αG,αB,α) the formula becomes a simple average
	//
	//		(αavg*Ravg, αavg*Gavg, αavg*Bavg, αavg)
	//
	//			  (α0*R0, α0*G0, α0*B0, α0) + (α1*R1, α1*G1, α1*B1, α1)
	//			= -----------------------------------------------------
	//			                            2
	//
//	if (RUNNING_OPENGL_3)
//		glGenerateMipmap(GL_TEXTURE_2D);

	//	Restore the default row alignment, if necessary.
	if (aDepth == 1 && aWidth < 4)
		glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
}


bool IsPowerOfTwo(unsigned int n)
{
	if (n == 0)
		return false;

	while ((n & 0x00000001) == 0)
		n >>= 1;

	return (n == 0x00000001);
}


void ResizeGLView(
	ModelData		*md,
	unsigned int	aNewViewWidth,	//	in pixels
	unsigned int	aNewViewHeight)	//	in pixels
{
	GLint	theOldViewport[4],
			theOldViewWidth,
			theOldViewHeight;
	double	theHorizontalFactor,
			theVerticalFactor,
			theScaleFactor;

	//	Ignore calls with aNewViewWidth or aNewViewHeight equal to zero, 
	//	because they'll foul up itsWindowHalfWidth and itsWindowHalfHeight.
	//	(Win32 generates such calls when the window is minimized.)
	if (aNewViewWidth == 0 || aNewViewHeight == 0)
		return;
		
	//	Note the previous view size.
	glGetIntegerv(GL_VIEWPORT, theOldViewport);
	theOldViewWidth		= theOldViewport[2];
	theOldViewHeight	= theOldViewport[3];

	//	Let OpenGL know about the new view size.
	glViewport(0, 0, aNewViewWidth, aNewViewHeight);
	
	//	The old view size should never be zero, but check it anyhow.
	if (theOldViewWidth == 0 || theOldViewHeight == 0)
		return;
	
	//	By what factor has the view grown/shrunk in each direction?
	theHorizontalFactor	= (double) aNewViewWidth  / (double) theOldViewWidth;
	theVerticalFactor	= (double) aNewViewHeight / (double) theOldViewHeight;

	//	Update the view's approximate dimensions (typically in centimeters).
	md->itsWindowHalfWidth	*= theHorizontalFactor;
	md->itsWindowHalfHeight	*= theVerticalFactor;
	
	//	If the resized view maintains the same aspect ratio as before,
	//	we'll want to draw the same content, only larger or smaller.
	//
	//	If the aspect ratio changes, we must choose between
	//	
	//		preserving the horizontal field of view,
	//		preserving the vertical field of view,
	//		or some compromise method of altering both.
	//
	//	In KaleidoTile this choice has only a modest effect:
	//	it affects the perspective, but not the size or placement of the tiling.

//	theScaleFactor = theHorizontalFactor;
//	theScaleFactor = theVerticalFactor;
	theScaleFactor = sqrt( theHorizontalFactor * theVerticalFactor );

	md->itsWindowDistance	*= theScaleFactor;
	md->itsEyeOffset		*= theScaleFactor;
}


ErrorText ReadRGBAPixels(
	unsigned int	aWidth,
	unsigned int	aHeight,
	Byte			*aBuffer)
{
	//	I bracketed the glReadBuffer() call with #ifdef GL_VERSION_1_1,
	//	so that the call remains visible under standard OpenGL 
	//	but disappears under OpenGL ES.
#ifdef GL_VERSION_1_1
	//	Mac OS X seems to literally swap buffers, rather than copying
	//	the contents of a fixed back buffer to a fixed front buffer.
	//	As a result, reading the back buffer gives the contents
	//	of the frame previous to the one currently visible.
	//	Solution:  Read the front buffer instead.
	glReadBuffer(GL_FRONT);
#endif

	glReadPixels(	0, 0,
					(GLsizei) aWidth, (GLsizei) aHeight,
					GL_RGBA, GL_UNSIGNED_BYTE,
					(GLvoid *) aBuffer);
	
	return GetErrorString();
}


ErrorText Render(ModelData *md)
{
	GLint	theFullViewport[4],
			theFullWidth,
			theFullHeight,
			theHalfWidth,
			thePanelWidth;

	//	A transparent background lets the user superimpose
	//	copied-and-pasted or saved images on top of the background
	//	of his/her choice.
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT);
	
	//	KaleidoTile draws the scene in strict back-to-front order
	//	to get transparency effects right, so it needs no depth buffer.
	glDisable(GL_DEPTH_TEST);

	//	Face culling isn't needed, except when rendering a spherical tiling,
	//	at which point we'll enable it temporarily.
	glDisable(GL_CULL_FACE);

	//	Blending determines how the final fragment
	//	blends in with the previous color buffer contents.
	//	For the background we'll temporarily disable blending.
	//	For everything else we'll leave it enabled.
	//
	//	Rigorously correct blending requires
	//
	//				   αs*(Rs,Gs,Bs) + (1 - αs)*αd*(Rd,Gd,Bd)
	//		(R,G,B) = ----------------------------------------
	//				         αs      +      (1 - αs)*αd
	//
	//		      α = αs + (1 - αs)*αd
	//
	//	Replacing the traditional (R,G,B,α) with the premultiplied (αR,αG,αB,α)
	//	simplifies the formula to 
	//
	//		(αR, αG, αB) = (αs*Rs, αs*Gs, αs*Bs) + (1 - αs)*(αd*Rd, αd*Gd, αd*Bd)
	//
	//		           α = αs + (1 - αs)*αd
	//
	//	Because they share the same coefficients, we may merge the RGB and α parts
	//	into a single formula
	//
	//		(αR, αG, αB, α) = (αs*Rs, αs*Gs, αs*Bs, αs) + (1 - αs)*(αd*Rd, αd*Gd, αd*Bd, αd)
	//
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

	//	According to the formal specification, one may call
	//
	//			glEnable(GL_MULTISAMPLE)
	//	or
	//			glDisable(GL_MULTISAMPLE)
	//
	//	but in practice the pixel format (with or without a multisample buffer)
	//	wholly determines whether multisampling takes place.
	//	This latter approach makes sense, given that 
	//	multisampling into a single-sample buffer would be impossible, and
	//	single-sampling into a multisample buffer would be pointless.
	//	Nevertheless, let's call glEnable/glDisable(GL_MULTISAMPLE),
	//	just in case some future OpenGL graphics driver wants the call.
	if (md->itsMultisamplingAvailableFlag && md->itsMultisamplingEnabledFlag)
		glEnable(GL_MULTISAMPLE);
	else
		glDisable(GL_MULTISAMPLE);

	//	Disable fixed-function arrays when running shaders,
	//	and disable generic attribute arrays when running in fixed-function mode.
	//	This guards against the possibility that the wrong kinds of arrays
	//	were left enabled from a previous frame, when itsShadersEnabledFlag
	//	may have been different.
	//
	//	Note #1.  A forward-compatible OpenGL 3.0 context
	//	will not support the fixed-function calls.
	//
	//	Note #2.  This whole paragraph of code may be deleted 
	//	when fixed-function support gets removed.
	if (md->itsShadersAvailableFlag && ! RUNNING_OPENGL_3)
	{
		if (md->itsShadersEnabledFlag)
		{
			//	We'll use only the forward-compatible glVertexAttribPointer(),
			//	not the backward-compatible glVertexPointer(), glTexCoordPointer(), etc.
			glDisableClientState(GL_VERTEX_ARRAY		);
			glDisableClientState(GL_NORMAL_ARRAY		);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY	);
			glDisableClientState(GL_COLOR_ARRAY			);
		}
		else
		{
			//	We'll use only the backward-compatible glVertexPointer(), glTexCoordPointer(), etc.,
			//	not the forward-compatible glVertexAttribPointer().
			glDisableVertexAttribArray(ATTRIBUTE_POSITION			);
			glDisableVertexAttribArray(ATTRIBUTE_MVP_MATRIX_ROW_0	);
			glDisableVertexAttribArray(ATTRIBUTE_MVP_MATRIX_ROW_1	);
			glDisableVertexAttribArray(ATTRIBUTE_MVP_MATRIX_ROW_2	);
			glDisableVertexAttribArray(ATTRIBUTE_MVP_MATRIX_ROW_3	);
			glDisableVertexAttribArray(ATTRIBUTE_DIFFUSE_EVALUATOR	);
			glDisableVertexAttribArray(ATTRIBUTE_SPECULAR_EVALUATOR	);
			glDisableVertexAttribArray(ATTRIBUTE_NORMAL				);
			glDisableVertexAttribArray(ATTRIBUTE_TEX_COORD			);
			glDisableVertexAttribArray(ATTRIBUTE_COLOR_HSLA			);
		}
	}

	//	Generate buffer names for the meshes
	//	realizing the fundamental triangle's three regions.
	glGenBuffers(3, (GLuint *) md->itsVertexBufferNames);
	glGenBuffers(3, (GLuint *) md->itsIndexBufferNames );

	//	Create Vertex Buffer Objects for those three meshes.
	MakeTriangleVBOs(md);

	//	Draw the tiling background, the tiling, 
	//	the Archimedean name (if any) and the control panel.
	if (md->itsStereoMode != StereoGeoWall)
	{
		RenderContent(md, EyeOnly);	//	EyeOnly gets ignored without StereoGeoWall.
	}
	else
	{
		//	Get original viewport dimensions.
		glGetIntegerv(GL_VIEWPORT, theFullViewport);
		theFullWidth	= theFullViewport[2];
		theFullHeight	= theFullViewport[3];
		theHalfWidth	= theFullWidth / 2;					//	discards fractional part, typically 0
		thePanelWidth	= theHalfWidth - theHalfWidth%5;	//	rounds down to multiple of 5

		//	Draw the left  eye image into the left  half of the window.
		glViewport(0, 0, thePanelWidth, theFullHeight);
		RenderContent(md, EyeLeft);

		//	Draw the right eye image into the right half of the window.
		glViewport(theHalfWidth, 0, thePanelWidth, theFullHeight);
		RenderContent(md, EyeRight);
			
		//	Restore the original viewport.
		glViewport(0, 0, theFullWidth, theFullHeight);
	}
	

	//	Free the Vertex Buffer Objects and release their names.
	//
	//	Minor technical comment:  Not only is there no harm
	//	in creating and freeing the VBOs once per frame,
	//	the frame rate actually turned out slightly higher (!) this way.
	glDeleteBuffers(3, (GLuint *) md->itsVertexBufferNames);
	glDeleteBuffers(3, (GLuint *) md->itsIndexBufferNames );

	//	Return a string describing the first of any OpenGL errors
	//	that may have occurred.  If no errors occurred, return NULL.
	return GetErrorString();
}


static void MakeTriangleVBOs(
	ModelData	*md)
{
	ktMetric		theMetric;
	ktVector		theTriplePoint,
					theScaledVertices[3],
					theCrossingPoints[3],
					theReflectedImage;
	ktVector		theCorners[4];
	double			theCornerTextureCoordinates[4][2],
					theNormal[3],
					theSubdivisionVertices  [(MAX_SUBDIVISION+1)*(MAX_SUBDIVISION+1)][3],
					theSubdivisionNormals   [(MAX_SUBDIVISION+1)*(MAX_SUBDIVISION+1)][3],
					theSubdivisionTexCoords [(MAX_SUBDIVISION+1)*(MAX_SUBDIVISION+1)][2];
	VBOData			theSubdivisionInterlaced[(MAX_SUBDIVISION+1)*(MAX_SUBDIVISION+1)];
	unsigned short	theSubdivisionFaces[MAX_SUBDIVISION*MAX_SUBDIVISION][2][3];
	unsigned int	n,	//	subdivide each curved quad into an n×n grid
					i,
					j,
					k,
					l;
	
	//	Define the trivial subdivision of a quad in "triangle strip" ordering,
	//	just in case iPod Touch implementations of OpenGL ES really do
	//	find that more efficient (I read something to that effect, 
	//	but I'm not convinced it's really true).
	static const unsigned short	theTrivialSubdivisionFaces[2][3] =
								{
									{1, 2, 0},
									{0, 2, 3}
								};

	//	Note the relevant metric.
	theMetric = GetMetric(md->itsTiling.itsGeometry, MetricVertical);

	//	Convert md->itsTriplePointPosition
	//	from barycentric coordinates to real coordinates.
	VectorTimesMatrix(	md->itsTriplePointPosition,
						md->itsTiling.itsBarycentricBasis,
						theTriplePoint);
	VectorNormalize(theTriplePoint, theMetric);

	//	The vertices of the fundamental triangle never move, but
	//	we must adjust their length to be coplanar with the triple point.
	for (i = 0; i < 3; i++)
	{
		ConstantTimesVector(
			fabs(VectorInnerProduct(
				theTriplePoint,
				md->itsTiling.itsVertices[i],
				theMetric)),
			md->itsTiling.itsVertices[i],
			theScaledVertices[i]);
	}

	//	The crossing points lie midway between theTriplePoint
	//	and each of theTriplePoint's reflected images.
	for (i = 0; i < 3; i++)
	{
		VectorTimesMatrix(	theTriplePoint,
							md->itsTiling.itsGenerators[i],
							theReflectedImage);
		VectorAverage(	theTriplePoint,
						theReflectedImage,
						theCrossingPoints[i]);
	}

	for (i = 0; i < 3; i++)
	{
		//	Locate the corners of the basic quad.
		VectorCopy(theCorners[0], theScaledVertices[i]      );
		VectorCopy(theCorners[1], theCrossingPoints[(i+2)%3]);
		VectorCopy(theCorners[2], theTriplePoint            );
		VectorCopy(theCorners[3], theCrossingPoints[(i+1)%3]);

		//	To cut along mirror lines, shrink the fundamental domain
		//	in the x- and y- (but not the z-) directions to leave a little gap
		//	between its translates.
		if (md->itsMirrorLinesFlag)
		{
			for (j = 0; j < 4; j++)
				for (k = 0; k < 2; k++)
					theCorners[j][k] *= 0.98;
		}

		//	Assign aesthetically pleasing texture coordinates.
		ChooseTextureCoordinatesAtCorners(
			theCorners,
			md->itsTiling.itsGeometry,
			theCornerTextureCoordinates);

		//	In a custom texture the true pixels fill only a subset
		//	of the power-of-two OpenGL texture.
		//	Scale the texture coordinates accordingly.
		if (md->itsFacePainting[i].itsStyle == FacePaintTextureCustom)
		{
			for (j = 0; j < 4; j++)
				for (k = 0; k < 2; k++)
					theCornerTextureCoordinates[j][k] =
						(	(theCornerTextureCoordinates[j][k] - 0.5)
						  * md->itsFacePainting[i].itsCustomTextureFraction )
						+ 0.5;
		}

		//	Define the face.
		//
		//	Use GetModifiedTilingStyle() to force (2,2,n) prisms
		//	to render in StyleFlat, no matter what the user requested.
		//	Please see GetModifiedTilingStyle() for a more detailed explanation.
		if (GetModifiedTilingStyle(md) == StyleFlat)
		{
			//	Realize a flat face as a 1 × 1 mesh.
			n = 1;

			//	Copy the four corners, converting doubles to floats.
			for (j = 0; j < 4; j++)
				for (k = 0; k < 3; k++)
					theSubdivisionInterlaced[j].pos[k] = (float) theCorners[j][k];
			
			//	Use a single normal vector orthogonal to the face.
			VectorNormal(	md->itsTiling.itsVertices[i],
							md->itsTiling.itsGeometry,
							theNormal);
			for (j = 0; j < 4; j++)
				for (k = 0; k < 3; k++)
					theSubdivisionInterlaced[j].nor[k] = (float) theNormal[k];

			//	Copy the corners' texture coordinates.
			for (j = 0; j < 4; j++)
				for (k = 0; k < 2; k++)
					theSubdivisionInterlaced[j].tex[k] = (float) theCornerTextureCoordinates[j][k];
			
			//	Copy the statically defined face indices.
			for (j = 0; j < 2; j++)
				for (k = 0; k < 3; k++)
					theSubdivisionFaces[0][j][k] = theTrivialSubdivisionFaces[j][k];
		}
		else	//	GetModifiedTilingStyle(md) == StyleCurved
		{
			n = md->itsMeshSize;

			//	Just to be safe...
			if ( ! (1 <= n && n <= MAX_SUBDIVISION))
				continue;

			for (j = 0; j < n + 1; j++)
				for (k = 0; k < n + 1; k++)
				{
					for (l = 0; l < 3; l++)
						theSubdivisionVertices[(n+1)*j + k][l] =
							(	(n - j) * (n - k) * theCorners[0][l]
							  + (n - j) *    k    * theCorners[1][l]
							  +    j    *    k    * theCorners[2][l]
							  +    j    * (n - k) * theCorners[3][l]
							) / (n*n);

					VectorNormalize(theSubdivisionVertices[(n+1)*j + k], theMetric);

					VectorNormal(	theSubdivisionVertices[(n+1)*j + k],
									md->itsTiling.itsGeometry,
									theSubdivisionNormals[(n+1)*j + k]);

					for (l = 0; l < 2; l++)
						theSubdivisionTexCoords[(n+1)*j + k][l] =
							(	(n - j) * (n - k) * theCornerTextureCoordinates[0][l]
							  + (n - j) *    k    * theCornerTextureCoordinates[1][l]
							  +    j    *    k    * theCornerTextureCoordinates[2][l]
							  +    j    * (n - k) * theCornerTextureCoordinates[3][l]
							) / (n*n);
				}

			for (j = 0; j < n; j++)
				for (k = 0; k < n; k++)
				{
					theSubdivisionFaces[n*j + k][0][0] = (unsigned short) ( (n + 1)*j + k + 0           );
					theSubdivisionFaces[n*j + k][0][1] = (unsigned short) ( (n + 1)*j + k + 1           );
					theSubdivisionFaces[n*j + k][0][2] = (unsigned short) ( (n + 1)*j + k + (n + 1)     );

					theSubdivisionFaces[n*j + k][1][0] = (unsigned short) ( (n + 1)*j + k + 1           );
					theSubdivisionFaces[n*j + k][1][1] = (unsigned short) ( (n + 1)*j + k + (n + 1) + 1 );
					theSubdivisionFaces[n*j + k][1][2] = (unsigned short) ( (n + 1)*j + k + (n + 1)     );
				}

			//	Copy the parallel arrays of doubles to an interlaced array of floats.
			for (j = 0; j < n + 1; j++)
				for (k = 0; k < n + 1; k++)
				{
					for (l = 0; l < 3; l++)
						theSubdivisionInterlaced[(n+1)*j + k].pos[l] = (float) theSubdivisionVertices[(n+1)*j + k][l];

					for (l = 0; l < 3; l++)
						theSubdivisionInterlaced[(n+1)*j + k].nor[l] = (float) theSubdivisionNormals[(n+1)*j + k][l];

					for (l = 0; l < 2; l++)
						theSubdivisionInterlaced[(n+1)*j + k].tex[l] = (float) theSubdivisionTexCoords[(n+1)*j + k][l];
				}
		}

		//	Create the Vertex Buffer Objects and load in the data.

		glBindBuffer(GL_ARRAY_BUFFER, md->itsVertexBufferNames[VertexBufferMeshBase + i]);
		glBufferData(GL_ARRAY_BUFFER, (n+1)*(n+1)*sizeof(VBOData), theSubdivisionInterlaced, GL_STATIC_DRAW);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, md->itsIndexBufferNames[VertexBufferMeshBase + i]);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, n * n * 2 * 3 * sizeof(unsigned short), theSubdivisionFaces, GL_STATIC_DRAW);
	}
}

static ktTilingStyle GetModifiedTilingStyle(ModelData *md)
{
	//	Normally we respect the user's choice of tiling style.
	//	However, to keep the mesh projection clean and simple,
	//	we make an exception and render a (2,2,n) polyhedron
	//	as StyleFlat no matter what the user requested.
	//	Otherwise the geometry gets a lot more complicated because
	//	the flat polyhedron's faces approach the origin and don't project
	//	well onto the sphere.

	if (TilingIs22n(&md->itsTiling))
		return StyleFlat;
	else
		return md->itsTilingStyle;
}

static void ChooseTextureCoordinatesAtCorners(
	ktVector	aCornerLocations[4],		//	input
	ktGeometry	aGeometry,					//	input
	double		aCornerTextureCoords[4][2])	//	output
{
	unsigned int	i;
	ktMetric		theMetric;
	ktVector		theCornerLocations[4],
					theAxisU,
					theAxisV,
					theProjection,
					theRelativePosition;
	double			theLengthU,
					theLengthV,
					theFactor;

	//	We expect all four corners to lie in a single plane orthogonal to corner 0.
	//	Assign corner 0 the texture coordinate (1/2, 0).
	//	Assign corner 2 the texture coordinate (1/2, 1).
	//	Assign the remaining texture coordinate so as to respect aspect ratios.

	//	Make a local copy of aCornerLocations.
	for (i = 0; i < 4; i++)
		VectorCopy(theCornerLocations[i], aCornerLocations[i]);

	//	Note the relevant metric.
	theMetric = GetMetric(aGeometry, MetricHorizontal);

	//	How long is the diagonal from corner 0 to corner 2?
	VectorDifference(theCornerLocations[2], theCornerLocations[0], theAxisV);
	theLengthV = VectorLength(theAxisV, theMetric);
	if (theLengthV < MIN_TEXURE_AXIS_LENGTH)
		goto SetFallbackTextureCoordinates;

	//	Rescale everything to normalize the aforementioned diagonal to length one.
	theFactor = 1.0 / theLengthV;
	for (i = 0; i < 4; i++)
		ConstantTimesVector(theFactor, theCornerLocations[i], theCornerLocations[i]);
	ConstantTimesVector(theFactor, theAxisV, theAxisV);

	//	Find an orthogonal axis.
	VectorDifference(theCornerLocations[1], theCornerLocations[3], theAxisU);
	ConstantTimesVector(
		VectorInnerProduct(theAxisU, theAxisV, theMetric),
		theAxisV,
		theProjection);
	VectorDifference(theAxisU, theProjection, theAxisU);
	theLengthU = VectorLength(theAxisU, theMetric);
	if (theLengthU < MIN_TEXURE_AXIS_LENGTH)
		goto SetFallbackTextureCoordinates;
	ConstantTimesVector(
		1.0 / theLengthU,
		theAxisU,
		theAxisU);

	//	Compute the texture coordinates relative to the given axes.
	for (i = 0; i < 4; i++)
	{
		VectorDifference(theCornerLocations[i], theCornerLocations[0], theRelativePosition);

		aCornerTextureCoords[i][0] = 0.5 + VectorInnerProduct(theRelativePosition, theAxisU, theMetric);
		aCornerTextureCoords[i][1] = 0.0 + VectorInnerProduct(theRelativePosition, theAxisV, theMetric);
	}

	return;

SetFallbackTextureCoordinates:

	aCornerTextureCoords[0][0] = 0.5;
	aCornerTextureCoords[0][1] = 0.0;

	aCornerTextureCoords[1][0] = 1.0;
	aCornerTextureCoords[1][1] = 0.5;

	aCornerTextureCoords[2][0] = 0.5;
	aCornerTextureCoords[2][1] = 1.0;

	aCornerTextureCoords[3][0] = 0.0;
	aCornerTextureCoords[3][1] = 0.5;
}

static void BindTriangleVBOs(
	unsigned int	aVertexBufferName,
	unsigned int	anIndexBufferName,
	bool			aShadersFlag,
	ShaderIndex		aShaderIndex)
{
	glBindBuffer(GL_ARRAY_BUFFER, aVertexBufferName);

	if (aShadersFlag)
	{
		glEnableVertexAttribArray(ATTRIBUTE_POSITION);
		glVertexAttribPointer(ATTRIBUTE_POSITION,  3, GL_FLOAT, GL_FALSE, sizeof(VBOData), (void *)offsetof(VBOData, pos));

		if (aShaderIndex == ShaderColorWithLighting || aShaderIndex == ShaderTextureWithLighting)
		{
			glEnableVertexAttribArray(ATTRIBUTE_NORMAL);
			glVertexAttribPointer(ATTRIBUTE_NORMAL, 3, GL_FLOAT, GL_FALSE, sizeof(VBOData), (void *)offsetof(VBOData, nor));
		}
		else
			glDisableVertexAttribArray(ATTRIBUTE_NORMAL);

		if (aShaderIndex == ShaderTexture || aShaderIndex == ShaderTextureWithLighting)
		{
			glEnableVertexAttribArray(ATTRIBUTE_TEX_COORD);
			glVertexAttribPointer(ATTRIBUTE_TEX_COORD, 2, GL_FLOAT, GL_FALSE, sizeof(VBOData), (void *)offsetof(VBOData, tex));
		}
		else
		{
			glDisableVertexAttribArray(ATTRIBUTE_TEX_COORD);
		}

		glDisableVertexAttribArray(ATTRIBUTE_COLOR_HSLA);
	}
	else
	{
		//	Note #4
		//		Internet folklore claims that glVertexPointer() should be called last,
		//		after glTexCoordPointer(), glColorPointer(), etc., but my own testing
		//		revealed no difference.  [THIS COMMENT APPLIES TO DEPRECATED FUNCTIONS ONLY.]

		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, sizeof(VBOData), (void *)offsetof(VBOData, pos));

//	The fallback fixed-function code provides no lighting.
//
//		if (aGeometry == GeometrySpherical)
//		{
//			glEnableClientState(GL_NORMAL_ARRAY);
//			glNormalPointer(GL_FLOAT, sizeof(VBOData), (void *)offsetof(VBOData, nor));
//		}
//		else
//			glDisableClientState(GL_NORMAL_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);

		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, sizeof(VBOData), (void *)offsetof(VBOData, tex));

		glDisableClientState(GL_COLOR_ARRAY);
	}

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, anIndexBufferName);
}

static void DrawTriangleVBOs(
	unsigned int	aMeshSize)	//	must agree with mesh size defined in MakeTriangleVBOs above
{
	unsigned int	theNumVertices,
					theNumFaces;

	theNumVertices	= (aMeshSize + 1) * (aMeshSize + 1);
	theNumFaces		= 2 * aMeshSize * aMeshSize;

	glDrawRangeElements(	GL_TRIANGLES,
							0,
							theNumVertices - 1,
							theNumFaces * 3,
							GL_UNSIGNED_SHORT,
							0);
}


static void RenderContent(
	ModelData	*md,
	EyeType		aGeoWallEye)
{
	GLint	theViewport[4],
			theViewportOriginX,
			theViewportOriginY,
			theViewportWidth,
			theViewportHeight,
			theMainPanelWidth,
			theControlPanelWidth;

	//	We'd like to divide the viewport into a square region 
	//	for the main tiling alongside a control panel.   
	//	In the Mac OS X version the full viewport always maintains 
	//	an aspect ratio of 5:4, so there's no problem alloting 80%
	//	of the window to the square tiling panel, with 20% remaining
	//	for the control panel.  In the Win32 version the user may
	//	vary the aspect ratio, but even so we pretend it's 5:4 and
	//	hope for the best.  This isn't a perfect solution, but it
	//	will have to do until I figure out how to enforce the 5:4 ratio.
	glGetIntegerv(GL_VIEWPORT, theViewport);
	theViewportOriginX		= theViewport[0];
	theViewportOriginY		= theViewport[1];
	theViewportWidth		= theViewport[2];
	theViewportHeight		= theViewport[3];
	theMainPanelWidth		= 4 * theViewportWidth / 5;
	theControlPanelWidth	= theViewportWidth - theMainPanelWidth;

	//	Draw the main panel.
	glViewport(	theViewportOriginX,
				theViewportOriginY,
				theMainPanelWidth,
				theViewportHeight);
	RenderMainPanelBackground(md);
	RenderMainPanelTiling(md, aGeoWallEye);
	RenderArchimedeanName(md);

	//	Draw the control panel.
	glViewport(	theViewportOriginX + theMainPanelWidth,
				theViewportOriginY,
				theControlPanelWidth,
				theViewportHeight);
	RenderControlPanel(md);

	//	Restore the original viewport.
	glViewport(	theViewportOriginX,
				theViewportOriginY,
				theViewportWidth,
				theViewportHeight);
}


static void RenderMainPanelBackground(ModelData *md)
{
	ColorRGBA		theRGBA;
	ShaderIndex		theShaderIndex;
	GLint			theShaderProgram;
	double			theAspectRatio;
	unsigned int	i;

	float			theTexCoords[4][2] =
					{	//	may be overridden below
						{  0.0,      0.0   },
						{TEX_REPS,   0.0   },
						{TEX_REPS, TEX_REPS},
						{  0.0,    TEX_REPS}
					};

	static const float	theVertices[4][3] =
						{
							{-1.0, -1.0, 0.0},
							{+1.0, -1.0, 0.0},
							{+1.0, +1.0, 0.0},
							{-1.0, +1.0, 0.0}
						};

	//	In stereoscopic 3D it's difficult to fuse the left- and right-eye
	//	images of the background.  Better to leave the background black.
	//	(Code to draw the background in stereo has been archived
	//	to the file "stereo 3D background.c".)
	if (md->itsStereoMode != StereoNone)
		return;

	//	The background needs no blending.
	glDisable(GL_BLEND);
	
	//	Disable Vertex Buffer Objects.
	//	We draw a single quad only once, so it's just as easy
	//	to pass the arrays directly from the CPU.
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	//	Set an appropriate color or texture.
	switch (md->itsFacePainting[BACKGROUND_IMAGE].itsStyle)
	{
		case FacePaintColorHS:
		case FacePaintColorLA:
			HSLAtoRGBA(&md->itsFacePainting[BACKGROUND_IMAGE].itsHSLA, &theRGBA);
			if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
			{
				theShaderIndex		= ShaderColor;
				theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
				glUseProgram(theShaderProgram);

				glUniform4f(glGetUniformLocation(theShaderProgram, "uniColor"), theRGBA.r, theRGBA.g, theRGBA.b, theRGBA.a);

				glEnableVertexAttribArray(ATTRIBUTE_POSITION);
				glVertexAttribPointer(ATTRIBUTE_POSITION,  3, GL_FLOAT, GL_FALSE, 0, theVertices);
				
				glDisableVertexAttribArray(ATTRIBUTE_NORMAL);
				
				glDisableVertexAttribArray(ATTRIBUTE_TEX_COORD);

				glDisableVertexAttribArray(ATTRIBUTE_COLOR_HSLA);
			}
			else
			{
				glColor4f(theRGBA.r, theRGBA.g, theRGBA.b, theRGBA.a);
				glDisable(GL_TEXTURE_2D);

				glEnableClientState(GL_VERTEX_ARRAY);
				glVertexPointer(3, GL_FLOAT, 0, theVertices);

				glDisableClientState(GL_NORMAL_ARRAY);

				glDisableClientState(GL_TEXTURE_COORD_ARRAY);

				glDisableClientState(GL_COLOR_ARRAY);
			}
			break;

		case FacePaintTextureCustom:
			theAspectRatio = md->itsFacePainting[BACKGROUND_IMAGE].itsCustomTextureAspectRatio;
			if (theAspectRatio > 0.0 && theAspectRatio <= 8.0)	//	something reasonable?
			{
				//	accommodate the (reasonable) aspect ratio
				for (i = 0; i < 4; i++)
				{
					theTexCoords[i][0] /= sqrt(theAspectRatio);
					theTexCoords[i][1] *= sqrt(theAspectRatio);
				}
			}
			else
			{
				//	ignore the (unreasonable) aspect ratio
			}
			//	Fall through to next case...
		case FacePaintTextureStandard:
			glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureTilingBackground]);
			if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
			{
				theShaderIndex		= ShaderTexture;
				theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
				glUseProgram(theShaderProgram);
				
				glEnableVertexAttribArray(ATTRIBUTE_POSITION);
				glVertexAttribPointer(ATTRIBUTE_POSITION,  3, GL_FLOAT, GL_FALSE, 0, theVertices);
				
				glDisableVertexAttribArray(ATTRIBUTE_NORMAL);
				
				glEnableVertexAttribArray(ATTRIBUTE_TEX_COORD);
				glVertexAttribPointer(ATTRIBUTE_TEX_COORD, 2, GL_FLOAT, GL_FALSE, 0, theTexCoords);

				glDisableVertexAttribArray(ATTRIBUTE_COLOR_HSLA);
			}
			else
			{
				glColor4f(1.0, 1.0, 1.0, 1.0);
				glEnable(GL_TEXTURE_2D);
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

				glEnableClientState(GL_VERTEX_ARRAY);
				glVertexPointer(3, GL_FLOAT, 0, theVertices);

				glDisableClientState(GL_NORMAL_ARRAY);

				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
				glTexCoordPointer(2, GL_FLOAT, 0, theTexCoords);

				glDisableClientState(GL_COLOR_ARRAY);
			}
			break;

		case FacePaintInvisible:
			//	Draw nothing.
			return;

		default:
			//	Should never occur.
			return;
	}

	//	No fancy transformations are required.
	Matrix44Identity(md->itsModelViewMatrix);
	Matrix44Identity(md->itsProjectionMatrix);
	SendTransformationsToShader(md, false);

	//	Draw the background, which will precisely fill the viewport.
	glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
}


static void RenderMainPanelTiling(
	ModelData	*md,
	EyeType		aGeoWallEye)	//	used only for StereoGeoWall
{
	//	Blending is essential for partially transparent surfaces.
	glEnable(GL_BLEND);

	//	Use shaders if possible, otherwise fall back to the fixed-function pipeline.
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
		//	forward-compatible code, using shaders
	}
	else
	{
		//	backward-compatible code, which may someday be eliminated
		glEnable(GL_TEXTURE_2D);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	}

	//	Draw the tiling.
	switch (md->itsStereoMode)
	{
		case StereoNone:

			//	Draw a full color image for a single eye.
			SetProjectionMatrix(EyeOnly, md);
			DrawTheScene(md);

			break;

		case StereoAnaglyphic:

			//	Restrict to the red channel.
			glColorMask(true, false, false, true);

			//	Draw the left eye image.
			SetProjectionMatrix(EyeLeft, md);
			DrawTheScene(md);

			//	Restrict to the "cyan channel".
			glColorMask(false, true, true, true);

			//	Draw the right eye image.
			SetProjectionMatrix(EyeRight, md);
			DrawTheScene(md);

			//	Re-enable all color channels.
			glColorMask(true, true, true, true);

			break;

		case StereoGeoWall:

			//	The viewport has already been set to one half
			//	of the window or the other.  Draw the tiling
			//	as seen by aGeoWallEye.
			SetProjectionMatrix(aGeoWallEye, md);
			DrawTheScene(md);

			break;
	}
}


static void SetProjectionMatrix(
	EyeType				anEyeType,
	ModelData			*md)
{
	//	FlatSpaceProjectionMatrices.pdf explains this construction
	//	of the projection matrix in full detail.

	double	(*m)[4],
			d,
			w,
			h,
			e,
			c,
			theIntrinsicWindowDistance,
			theTranslation[4][4];
	
	m = md->itsProjectionMatrix;

	d = md->itsWindowDistance;
	w = md->itsWindowHalfWidth;
	if (md->itsStereoMode == StereoGeoWall)
		w *= 0.5;
	h = md->itsWindowHalfHeight;

	switch (md->itsTiling.itsGeometry)
	{
		case GeometrySpherical:
			//	Size to the panel's smaller dimension, 
			//	so nothing gets omitted.
			c = (w <= h ? w : h) / TILING_PANEL_HALF_SIZE_INTRINSIC_SPH;	//	centimeters per intrinsic unit
			break;
		
		case GeometryEuclidean:
			//	Size to the panel's larger dimension, 
			//	to ensure that the tiling continues all the way
			//	to (and slightly past) the panel's edge.
			c = (w >= h ? w : h) / TILING_PANEL_HALF_SIZE_INTRINSIC_EUC;	//	centimeters per intrinsic unit
			break;
		
		case GeometryHyperbolic:

			//	Size to the panel's smaller dimension, 
			//	so nothing gets omitted.
			c = (w <= h ? w : h) / TILING_PANEL_HALF_SIZE_INTRINSIC_HYP;	//	centimeters per intrinsic unit

			//	For stereo viewing, provide an extra margin,
			//	so the shifted views don't extend past the edge of the view.
			//	(I'm sure there'd be some rigorous way to work out
			//	the required factor exactly, but for now this fudge factor
			//	seems to do the job, at least with the default settings.)
			if (anEyeType != EyeOnly)
				c *= 0.9375;

			break;
		
		default:	//	to suppress compiler warnings
			c = 1.0;
			break;
	}

	switch (anEyeType)
	{
		case EyeOnly:	e = 0;					break;
		case EyeLeft:	e = +md->itsEyeOffset;	break;
		case EyeRight:	e = -md->itsEyeOffset;	break;
		default:		e = 0;					break;	//	unused
	}

	//	In the hyperbolic case, view the hyperboloid from z = -1,
	//	so the user sees the Poincaré disk model.
	if (md->itsTiling.itsGeometry == GeometryHyperbolic)
	{
		//	Keep stereo effects proportional to new viewing distance.
		if (d > 0)
			e *= c / d;

		//	View the hyperboloid from z = -d/c = -1.
		d = c;
	}

	theIntrinsicWindowDistance = d / c;

	//	KaleidoTile uses no depth buffer, so we may safely 
	//	set the near clip to 0 and the far clip to infinity.
//	n = 0;
//	f = ∞;

	if (w == 0 || h == 0 || c == 0)	//	should never occur
	{
		Matrix44Identity(m);
		return;
	}

	//	Technical Note:  In theory we can set m[2][2] to the values 
	//	you get with n = 0 and f = ∞, and in practice this works fine
	//	on my computer.  But I'm a little nervous about having all vertices
	//	end up on the far face of the viewing volume (at z = w),
	//	just in case someday some non-compliant OpenGL implementation
	//	clips against -w < z < +w instead of the prescribed -w ≤ z ≤ +w.
	//	Therefore, for added safely, I've artificially set m[2][2] = 0.0
	//	to force all vertices to the middle plane of the viewing volume.

	m[0][0] = d/w;
	m[0][1] = 0;
	m[0][2] = 0;
	m[0][3] = 0;

	m[1][0] = 0;
	m[1][1] = d/h;
	m[1][2] = 0;
	m[1][3] = 0;

	m[2][0] = -e/w;
	m[2][1] = 0;
	m[2][2] = 0;	//	(f+n)/(f-n) = 1 when n=0 and f=∞, but see Technical Note above 
	m[2][3] = 1;

	m[3][0] = (e*d)/(c*w);
	m[3][1] = 0;
	m[3][2] = 0;	//	-2*f*n/(f-n) = 0 when n=0 and f=∞
	m[3][3] = 0;

	//	Just before applying the projection, translate the scenery
	//	so that the model may sit at the origin in world space and 
	//	still have the projection come out right.
	//	In the Euclidean case, move the Euclidean plane one extra unit
	//	towards the observer, so it coincides with the plane of the screen.
	Matrix44Identity(theTranslation);
	switch (md->itsTiling.itsGeometry)
	{
		case GeometrySpherical:   theTranslation[3][2] = theIntrinsicWindowDistance + 0.0;  break;
		case GeometryEuclidean:   theTranslation[3][2] = theIntrinsicWindowDistance - 1.0;  break;
		case GeometryHyperbolic:  theTranslation[3][2] = theIntrinsicWindowDistance + 0.0;  break;
	}
	Matrix44Product(theTranslation, m, m);
}


static void DrawTheScene(ModelData *md)
{
	ktMatrix	theBaseTrianglePlacement;

	//	Compute the base triangle's present position.
	IntegrateVelocity(1.0, md->itsTilingPosition, md->itsTiling.itsGeometry, theBaseTrianglePlacement);
	
	//	Flat and hyperbolic tilings may be rendered in one pass,
	//	but spherical tilings require two passes 
	//	(first the back faces, then the front faces).
	if (md->itsTiling.itsGeometry != GeometrySpherical)
	{
		//	Draw all tiles.
		DrawTheSceneWithParity(md, theBaseTrianglePlacement, md->itsPositiveImagesFlag, md->itsNegativeImagesFlag, false);
	}
	else	//	GeometrySpherical
	{
		//	Draw the scene in back-to-front order to get the transparency right.
		
		//	Temporarily enable face culling.
		glEnable(GL_CULL_FACE);

		//	Draw all front-facing polygons.
		//
		//	Note:  The sphere, the Euclidean plane and the hyperbolic plane
		//	are all tangent at the north pole (0,0,1), with their "front faces"
		//	facing towards the origin.  In particular, on the sphere
		//	the front faces face the sphere's interior while its back faces
		//	face its exterior.  It follows that we want to draw front faces
		//	before back faces.
		glCullFace(GL_BACK);	//	ignore back-facing polygons
		if (md->itsPositiveImagesFlag)
		{
			glFrontFace(GL_CCW);
			DrawTheSceneWithParity(md, theBaseTrianglePlacement, true, false, false);
		}
		if (md->itsNegativeImagesFlag)
		{
			glFrontFace(GL_CW);
			DrawTheSceneWithParity(md, theBaseTrianglePlacement, false, true, false);
		}

		//	Draw all back-facing polygons.
		glCullFace(GL_FRONT);	//	ignore front-facing polygons
		if (md->itsPositiveImagesFlag)
		{
			glFrontFace(GL_CCW);
			DrawTheSceneWithParity(md, theBaseTrianglePlacement, true, false, false);
		}
		if (md->itsNegativeImagesFlag)
		{
			glFrontFace(GL_CW);
			DrawTheSceneWithParity(md, theBaseTrianglePlacement, false, true, false);
		}

		//	Disable face culling, so the 2D control panels 
		//	won't need to worry about chirality.
		glDisable(GL_CULL_FACE);
	}
}

static void DrawTheSceneWithParity(
	ModelData	*md,
	ktMatrix	aBaseTrianglePlacement,	//	May be NULL when aBaseTriangleOnlyFlag == true
										//	The code below assumes itsModelViewMatrix and
										//	itsProjectionMatrix have been set, but not
										//	necessarily sent to the GPU.
	bool		aDrawPositiveImagesFlag,
	bool		aDrawNegativeImagesFlag,
	bool		aBaseTriangleOnlyFlag)	//	true  = draw a single fixed image of base triangle
										//	false = draw the full tiling
{
	unsigned int	theVertexIndex;
	ColorRGBA		theRGBA;
	ShaderIndex		theShaderIndex				= (ShaderIndex) 0;
	GLint			theShaderProgram			= 0;
	bool			theDirectionalLightingFlag;

	static double	theReflection[4][4] =
					{
						{ 1.0,  0.0,  0.0,  0.0 },
						{ 0.0,  1.0,  0.0,  0.0 },
						{ 0.0,  0.0, -1.0,  0.0 },
						{ 0.0,  0.0,  0.0,  1.0 }
					};
	
	for (theVertexIndex = 0; theVertexIndex < 3; theVertexIndex++)
	{
		//	Is this face invisible, colored or textured?
		switch (md->itsFacePainting[theVertexIndex].itsStyle)
		{
			case FacePaintInvisible:
				//	Skip this face and continue with the for() loop.
				continue;

			case FacePaintColorHS:
			case FacePaintColorLA:
				HSLAtoRGBA(&md->itsFacePainting[theVertexIndex].itsHSLA, &theRGBA);
				if (md->itsStereoMode == StereoAnaglyphic)
					ConvertColorToGreyscale(&theRGBA);
				if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
				{
					if (md->itsTiling.itsGeometry == GeometrySpherical && ! aBaseTriangleOnlyFlag )
						theShaderIndex = ShaderColorWithLighting;
					else
						theShaderIndex = ShaderColor;
				}
				else
				{
					glColor4f(theRGBA.r, theRGBA.g, theRGBA.b, theRGBA.a);
					glDisable(GL_TEXTURE_2D);
				}
				break;

			case FacePaintTextureStandard:
			case FacePaintTextureCustom:
				glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureFaceBase + theVertexIndex]);
				if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
				{
					if (md->itsTiling.itsGeometry == GeometrySpherical && ! aBaseTriangleOnlyFlag )
						theShaderIndex = ShaderTextureWithLighting;
					else
						theShaderIndex = ShaderTexture;
				}
				else
				{
					glColor4f(1.0, 1.0, 1.0, 1.0);
					glEnable(GL_TEXTURE_2D);
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
				}
				break;
		}
		
		//	Bind the shader program.
		if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
		{
			theShaderProgram = md->itsShaderPrograms[theShaderIndex];
			glUseProgram(theShaderProgram);
		}
			
		//	Bind uniform variables, as needed.
		if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
		{
			if (theShaderIndex == ShaderColor || theShaderIndex == ShaderColorWithLighting)
				glUniform4f(glGetUniformLocation(theShaderProgram, "uniColor"), theRGBA.r, theRGBA.g, theRGBA.b, theRGBA.a);
		}

		//	Bind the Vertex Buffer Objects and use them
		//	to bind vertex attribute arrays, as needed.
		BindTriangleVBOs(	md->itsVertexBufferNames[VertexBufferMeshBase + theVertexIndex],
							md->itsIndexBufferNames [VertexBufferMeshBase + theVertexIndex],
							md->itsShadersAvailableFlag && md->itsShadersEnabledFlag,
							theShaderIndex);

		//	Draw either the base triangle or the whole tiling, as requested.
		theDirectionalLightingFlag	=	theShaderIndex == ShaderColorWithLighting
									 || theShaderIndex == ShaderTextureWithLighting;
		if (aBaseTriangleOnlyFlag)
		{
			//	Draw the triangle once.
			//
			//	In the spherical case, theReflection pushes the triangle
			//	to the near side of the sphere (z < 0), with its normal vectors
			//	pointing towards the observer (which is relevant only when
			//	theShaderIndex includes lighting, which by default it does not).
			if (md->itsTiling.itsGeometry == GeometrySpherical)
				Matrix44Copy(md->itsModelViewMatrix, theReflection);
			else
				Matrix44Identity(md->itsModelViewMatrix);
			SendTransformationsToShader(md, theDirectionalLightingFlag);
			DrawTriangleVBOs(GetModifiedTilingStyle(md) == StyleCurved ? md->itsMeshSize : 1);
		}
		else
		{
			//	Draw the whole tiling.
			DrawTiling(	md,
						aBaseTrianglePlacement,
						md->itsTiling.itsInradius,				//	used in Euclidean case only
						md->itsTiling.itsDomainRadius,			//	used in Euclidean case only
						cosh(md->itsHyperbolicTilingRadius),	//	used in hyperbolic case only
						GetModifiedTilingStyle(md) == StyleCurved ? md->itsMeshSize : 1,
						aDrawPositiveImagesFlag,
						aDrawNegativeImagesFlag,
						theDirectionalLightingFlag);
		}
	}
}

static void SendTransformationsToShader(
	ModelData	*md,
	bool		aDirectionalLightingFlag)	//	Provide normal evaluators in support of lighting?
{
	double	theModelViewProjectionMatrix[4][4];
	float	theModelViewProjectionMatrixFloat[4][4];
	double	theDiffuseEvaluator[4],
			theSpecularEvaluator[4];

	//	The direction of maximal specular reflection is given by 
	//	the pre-computed vector
	//
	//		normalize(theLightDirection + theViewerDirection)
	//
	static double	theLightDirection[4]	= {-0.27, +0.27, -0.92,  0.00},	//	points towards the light source
				//	theViewerDirection[4]	= { 0.00,  0.00, -1.00,  0.00},	//	points towards the observer
					theSpecularDirection[4]	= {-0.14, +0.14, -0.98,  0.00};	//	direction of maximal specular reflection


	Matrix44Product(md->itsModelViewMatrix, md->itsProjectionMatrix, theModelViewProjectionMatrix);
	Matrix44DoubleToFloat(theModelViewProjectionMatrixFloat, theModelViewProjectionMatrix);

	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
		glVertexAttrib4fv(ATTRIBUTE_MVP_MATRIX_ROW_0, theModelViewProjectionMatrixFloat[0]);
		glVertexAttrib4fv(ATTRIBUTE_MVP_MATRIX_ROW_1, theModelViewProjectionMatrixFloat[1]);
		glVertexAttrib4fv(ATTRIBUTE_MVP_MATRIX_ROW_2, theModelViewProjectionMatrixFloat[2]);
		glVertexAttrib4fv(ATTRIBUTE_MVP_MATRIX_ROW_3, theModelViewProjectionMatrixFloat[3]);

		if (aDirectionalLightingFlag)
		{
			//	The light direction is an honest vector, which
			//	we write as a row vector for consistency
			//	with KaleidoTile's left-to-right matrix actions.
			//	By contrast a surface normal is, strictly speaking,
			//	a 1-form, which we write as a column
			//	for consistency with the left-to-right matrix actions.
			//	To compute the intensity of a light, we take
			//	the matrix product of the light direction
			//	with the surface normal:
			//
			//			          (Nx)
			//			(Lx Ly Lz)(Ny)
			//			          (Nz)
			//
			//	(Note:  If we were instead using right-to-left matrix actions,
			//	we'd take the transpose of everything in sight and
			//	write vectors as columns and 1-forms as rows.)
			//
			//	If we rotate the surface and the light source simultaneously
			//	using a matrix M, the light direction vector transforms
			//	as L → LM while the normal transforms as N → M⁻¹N,
			//	leaving their product
			//
			//			          (     )(     )(Nx)
			//			(Lx Ly Lz)(  M  )( M⁻¹ )(Ny)
			//			          (     )(     )(Nz)
			//
			//	invariant, as it must be.
			//
			//	In the present case, however, we want to leave
			//	the light source fixed while we rotate the polyhedron
			//	and its normals.  The product thus becomes
			//
			//			          (     )(Nx)
			//			(Lx Ly Lz)( M⁻¹ )(Ny)
			//			          (     )(Nz)
			//
			//	Conceptually we group the factors as L(M⁻¹N),
			//	but to minimize the computational load on the GPU,
			//	it's convenient to regroup the factors as (LM⁻¹)N
			//	and pre-compute the product LM⁻¹, which we call
			//	a "normal evaluator".
			//
			//	In general we'd need to invert the matrix M to compute LM⁻¹.
			//	In O(3), however, the inverse is just the transpose.
			//	So instead of computing LM⁻¹ we compute its transpose
			//
			//		transpose(LM⁻¹) = transpose(L transpose(M))
			//						= M transpose(L)
			//
			Matrix44TimesColumnVector(md->itsModelViewMatrix, theLightDirection, theDiffuseEvaluator);
			glVertexAttrib3f(	ATTRIBUTE_DIFFUSE_EVALUATOR,
								(float) theDiffuseEvaluator[0],
								(float) theDiffuseEvaluator[1],
								(float) theDiffuseEvaluator[2]);

			//	The same logic applies to compute theSpecularEvaluator.
			Matrix44TimesColumnVector(md->itsModelViewMatrix, theSpecularDirection, theSpecularEvaluator);
			glVertexAttrib3f(	ATTRIBUTE_SPECULAR_EVALUATOR,
								(float) theSpecularEvaluator[0],
								(float) theSpecularEvaluator[1],
								(float) theSpecularEvaluator[2]);
		}
	}
	else	//	no shaders
	{
		glMatrixMode(GL_MODELVIEW);
		glLoadMatrixf((float *)theModelViewProjectionMatrixFloat);
	}
}

static void DrawTiling(
	ModelData		*md,
	ktMatrix		aBaseTrianglePlacement,
	double			anInradius,			//	used in Euclidean case only
	double			aDomainRadius,		//	used in Euclidean case only
	double			aCoshTilingRadius,	//	used in hyperbolic case only
	unsigned int	aMeshSize,
	bool			aDrawPositiveImagesFlag,
	bool			aDrawNegativeImagesFlag,
	bool			aDirectionalLightingFlag)
{
	unsigned int		*thePQR,
						p,
						q,
						r;
	double				theDiskArea			= 0.0,
						theTriangleArea		= 0.0;
	unsigned int		theMaxNumTriangles,
						theMaxNumSegments,
						theNumAllocatedSegments,	//	number of segments allocated from our memory pool
						theNumActiveSegments,		//	number of unblocked segments in the current perimeter
						i;
	PerimeterSegment	*thePerimeterSegments,		//	our memory pool
						*theSegment;				//	the PerimeterSegment currently under consideration
	ktMatrix			theNewPlacement;
	bool				theNewImageIsReflected;
	PerimeterSegment	*theNewLeftSegment,
						*theNewRightSegment;

	//	Read the defining parameters of the (p,q,r) triangle group.
	thePQR = md->itsTiling.itsPQR;
	p = thePQR[0];
	q = thePQR[1];
	r = thePQR[2];
					
	//	We'll maintain a circular doubly-linked list representing 
	//	the perimeter of the region tiled so far.
	//	The algorithm will then go round and round the list, 
	//	adding more tiles and thus expanding the region outward.

	//	To avoid a lot of traffic with the memory manager,
	//	allocate one big block of memory and manage it ourselves.
	//	[I'm not sure whether this makes a difference or not,
	//	but there's no harm in playing it safe.]
	//
	//	How many PerimeterSegments will we need?
	//	Well, the initial triangle has 3 perimeter segments.
	//	Thereafter every time we add a triangle, we add 2 PerimeterSegments
	//	(which may or may not cancel with neighboring PerimeterSegments,
	//	but from the point of view of memory management that's irrelevant
	//	because we abandon cancelled PerimeterSegments, we don't recycle them).
	//	So the total number of PerimeterSegments is 2t + 1, where t is
	//	the number of triangles in the tiling.
	//
	//	How many triangles will the tiling contain?
	//	A simple area computation gives the answer.
	//	Break into cases according to the geometry.
	switch (md->itsTiling.itsGeometry)
	{
		case GeometrySpherical:
			//	The sphere as a whole has area 4π.
			//	Each triangle has area (π/p + π/q + π/r) - π.
			//	Therefore the number of triangles is
			//
			//					 4π
			//			---------------------
			//			(π/p + π/q + π/r) - π
			//	or
			//					4pqr
			//			--------------------
			//			(qr + rp + pq) - pqr
			//
			theMaxNumTriangles = 4*p*q*r / (q*r + r*p + p*q - p*q*r);
			break;

		case GeometryEuclidean:

			//	The tiling will be wholly contained within a disk of radius
			//	r = TILING_PANEL_RADIUS_EUC + 2*aDomainRadius, which has area πr².
			theDiskArea = PI * (TILING_PANEL_RADIUS_EUC + 2*aDomainRadius) * (TILING_PANEL_RADIUS_EUC + 2*aDomainRadius);

			//	We can deduce the fundamental triangle's area from its inradius
			//	by considering separately the three cases (3,3,3), (2,4,4) and (2,3,6).
			//	The parameters (p,q,r) may arrive permuted -- for example
			//	(3,6,2) is equivalent to (2,3,6) -- so for simplicity we recognize
			//	a (p,q,r) group by its sum p + q + r.
			switch (p + q + r)
			{
				case  9:	//	(3,3,3) triangle group
					theTriangleArea = (3*ROOT3) * anInradius * anInradius;
					break;

				case 10:	//	(2,4,4) triangle group
					theTriangleArea = (3 + 2*ROOT2) * anInradius * anInradius;
					break;

				case 11:	//	(2,3,6) triangle group
					theTriangleArea = (3 + 2*ROOT3) * anInradius * anInradius;
					break;

				default:
					return;	//	should never occur
			}
			
			//	Just to be safe...
			if (theTriangleArea <= 1e-4)
				return;	//	should never occur

			//	How many triangles should we allow for?
			theMaxNumTriangles = (unsigned int) floor(theDiskArea / theTriangleArea);

			break;

		case GeometryHyperbolic:
			//	A disk of radius r has area 2π(cosh(r) - 1).
			//	Each triangle has area π - (π/p + π/q + π/r).
			//	Therefore the number of triangles is approximately
			//
			//			   2π(cosh(r) - 1)
			//			---------------------
			//			π - (π/p + π/q + π/r)
			//	or
			//			 2pqr(cosh(r) - 1)
			//			--------------------
			//			pqr - (qr + rp + pq)
			//
			//	To allow for discretization effects around the boundary,
			//	pad that estimate by an extra 5% (which in practice seems
			//	to be more than sufficient).
			theMaxNumTriangles = (unsigned int) ceil( 1.05 * 
				2*p*q*r*(aCoshTilingRadius - 1.0) / (p*q*r - (q*r + r*p + p*q)));
			break;
		
		default:
			return;	//	should never occur
	}
	
	theMaxNumSegments = 2*theMaxNumTriangles + 1;

	thePerimeterSegments = (PerimeterSegment *) GET_MEMORY(theMaxNumSegments * sizeof(PerimeterSegment));
	if (thePerimeterSegments == NULL)
		goto CleanUpDrawTiling;
	theNumAllocatedSegments = 0;
	
	//	Draw the fundamental triangle's base image...
	if (aDrawPositiveImagesFlag)
	{
		MatrixToMatrix44(aBaseTrianglePlacement, md->itsModelViewMatrix);
		SendTransformationsToShader(md, aDirectionalLightingFlag);
		DrawTriangleVBOs(aMeshSize);
	}
	
	//	...and use its three sides to initialize the perimeter.
	if (theMaxNumSegments < 3)
		goto CleanUpDrawTiling;
	for (i = 0; i < 3; i++)
	{
		thePerimeterSegments[i].itsImageIsReflected	= false;
		thePerimeterSegments[i].itsSideIndex		= i;
		thePerimeterSegments[i].itsLeftNeighbor		= &thePerimeterSegments[(i+1)%3];
		thePerimeterSegments[i].itsRightNeighbor	= &thePerimeterSegments[(i+2)%3];
		thePerimeterSegments[i].itsLeftFreeValence	= 2*thePQR[(i+2)%3] - 1;
		MatrixCopy(thePerimeterSegments[i].itsPlacement, aBaseTrianglePlacement);
		thePerimeterSegments[i].itsBlockedFlag		= false;
	}
	theNumAllocatedSegments	+= 3;
	theNumActiveSegments	= 3;
	
	//	As long as the perimeter contains active ( = unblocked ) PerimeterSegments,
	//	keep pushing the perimeter outwards.
	theSegment = &thePerimeterSegments[0];
	while (theNumActiveSegments > 0)
	{
		if ( ! theSegment->itsBlockedFlag )
		{
			//	If we push outward across theSegment,
			//	where will the next image of the fundamental triangle lie?
			MatrixProduct(	md->itsTiling.itsGenerators[theSegment->itsSideIndex],
							theSegment->itsPlacement,
							theNewPlacement);
			theNewImageIsReflected = ! theSegment->itsImageIsReflected;
			
			//	If theNewPlacement is close enough to the origin...
			if (TrianglePlacementIsCloseEnough(theNewPlacement[2], aDomainRadius, aCoshTilingRadius))
			{
				//	Draw the corresponding image of the fundamental triangle.
				if (theNewImageIsReflected ? aDrawNegativeImagesFlag : aDrawPositiveImagesFlag)
				{
					MatrixToMatrix44(theNewPlacement, md->itsModelViewMatrix);
					SendTransformationsToShader(md, aDirectionalLightingFlag);
					DrawTriangleVBOs(aMeshSize);
				}

				//	Push the perimeter across this image of the fundamental triangle.
				
				//	Have we got two more PerimeterSegments in our memory pool?
				//	If not, terminate the tiling immediately.
				if (theNumAllocatedSegments + 2 > theMaxNumSegments)
					break;	//	break from enclosing while() loop
					
				//	Create the two new segments and install them
				//	into the perimeter in place of theSegment.

				theNewLeftSegment	= &thePerimeterSegments[theNumAllocatedSegments + 0];
				theNewRightSegment	= &thePerimeterSegments[theNumAllocatedSegments + 1];
				theNumAllocatedSegments	+= 2;
				theNumActiveSegments	+= (2 - 1);	//	two new segments added, one old one removed

				theNewLeftSegment->itsImageIsReflected	= theNewImageIsReflected;
				theNewRightSegment->itsImageIsReflected	= theNewImageIsReflected;

				if (theSegment->itsImageIsReflected)
				{
					theNewLeftSegment->itsSideIndex		= (theSegment->itsSideIndex + 2) % 3;
					theNewRightSegment->itsSideIndex	= (theSegment->itsSideIndex + 1) % 3;
				}
				else
				{
					theNewLeftSegment->itsSideIndex		= (theSegment->itsSideIndex + 1) % 3;
					theNewRightSegment->itsSideIndex	= (theSegment->itsSideIndex + 2) % 3;
				}

				theNewLeftSegment->itsLeftNeighbor						= theSegment->itsLeftNeighbor;
				theNewLeftSegment->itsLeftNeighbor->itsRightNeighbor	= theNewLeftSegment;
				theNewLeftSegment->itsRightNeighbor						= theNewRightSegment;

				theNewRightSegment->itsLeftNeighbor						= theNewLeftSegment;
				theNewRightSegment->itsRightNeighbor					= theSegment->itsRightNeighbor;
				theNewRightSegment->itsRightNeighbor->itsLeftNeighbor	= theNewRightSegment;

				theNewLeftSegment->itsLeftFreeValence	= theSegment->itsLeftFreeValence - 1;
				theNewRightSegment->itsLeftFreeValence	= 2*thePQR[theSegment->itsSideIndex] - 1;
				theNewRightSegment->itsRightNeighbor->itsLeftFreeValence--;

				MatrixCopy(theNewLeftSegment->itsPlacement,  theNewPlacement);
				MatrixCopy(theNewRightSegment->itsPlacement, theNewPlacement);

				theNewLeftSegment->itsBlockedFlag	= false;
				theNewRightSegment->itsBlockedFlag	= false;
				
				//	If a free valence reaches zero,
				//	the perimeter has folded in on itself.
				//	Remove the two "cancelling" edges from the perimeter.
				//	
				//	Note #1.  The perimeter length (before pushing aross
				//	the current triangle image) should always be at least 3.
				//	It should (I think) be exactly 3 only when the tiling first begins
				//	and, in the spherical case, when the tilings ends
				//	at the antipodal triangle.  The latter case gets detected
				//	in the code immediately below, which checks whether
				//	the tiling folds in on both sides simultaneously.
				//
				//	Note #2.  In a more generic tiling you'd have to worry
				//	about the tiled region overlapping itself in less friendly ways.
				//	Fortunately for the (p,q,r) triangle groups the only cases
				//	that arise are simple fold-ins and the insertion of the
				//	antipodal triangle in the spherical triangle tilings.
				
				if (theNewLeftSegment->itsLeftFreeValence == 0
				 && theNewRightSegment->itsRightNeighbor->itsLeftFreeValence == 0)
				{
					//	If both sides are folding in simultaneously,
					//	this means (I think) that we've reached the antipodal triangle
					//	in a spherical tiling.  If this assumption is correct,
					//	we can stop here.
					break;	//	break from enclosing while() loop
				}
				
				if (theNewLeftSegment->itsLeftFreeValence == 0)
				{
					theNewRightSegment->itsLeftFreeValence += theNewLeftSegment->itsLeftNeighbor->itsLeftFreeValence;
					theNewRightSegment->itsLeftFreeValence -= 2*thePQR[theSegment->itsSideIndex];

					theNewRightSegment->itsLeftNeighbor = theNewLeftSegment->itsLeftNeighbor->itsLeftNeighbor;
					theNewRightSegment->itsLeftNeighbor->itsRightNeighbor = theNewRightSegment;
					
					if ( ! theNewLeftSegment->itsBlockedFlag )					//	test is definitely unnecessary
						theNumActiveSegments--;
					if ( ! theNewLeftSegment->itsLeftNeighbor->itsBlockedFlag )	//	test is almost surely unnecessary
						theNumActiveSegments--;
					
					theSegment = theNewRightSegment->itsLeftNeighbor;	//	consider this segment on the next pass through the loop
				}
				else
				if (theNewRightSegment->itsRightNeighbor->itsLeftFreeValence == 0)
				{
					theNewRightSegment->itsRightNeighbor->itsRightNeighbor->itsLeftFreeValence
						+= theNewRightSegment->itsLeftFreeValence;
					theNewRightSegment->itsRightNeighbor->itsRightNeighbor->itsLeftFreeValence
						-= 2*thePQR[theSegment->itsSideIndex];
					
					theNewLeftSegment->itsRightNeighbor = theNewRightSegment->itsRightNeighbor->itsRightNeighbor;
					theNewLeftSegment->itsRightNeighbor->itsLeftNeighbor = theNewLeftSegment;

					if ( ! theNewRightSegment->itsBlockedFlag )						//	test is definitely unnecessary
						theNumActiveSegments--;
					if ( ! theNewRightSegment->itsRightNeighbor->itsBlockedFlag )	//	test is almost surely unnecessary
						theNumActiveSegments--;
					
					theSegment = theNewLeftSegment->itsLeftNeighbor;	//	consider this segment on the next pass through the loop
				}
				else
					theSegment = theNewLeftSegment->itsLeftNeighbor;	//	consider this segment on the next pass through the loop
			}
			else	//	TrianglePlacementIsCloseEnough() returned 'false'
			{
				//	theSegment is blocked.  Mark it and move on.
				theSegment->itsBlockedFlag = true;
				theNumActiveSegments--;
				theSegment = theSegment->itsLeftNeighbor;
			}
		}
		else	//	theSegment->itsBlockedFlag == true
		{
			//	Move on to the next segment.
			theSegment = theSegment->itsLeftNeighbor;
		}
	}

CleanUpDrawTiling:
	//	Free our supply of PerimeterSegments.
	FREE_MEMORY_SAFELY(thePerimeterSegments);
}

static bool TrianglePlacementIsCloseEnough(
	ktVector	aTriangleCenter,
	double		aDomainRadius,		//	used in Euclidean case only
	double		aCoshTilingRadius)	//	used in hyperbolic case only
{
	bool	theReturnValue;
	double	theTilingRadius;

	if (aTriangleCenter[2] > 1.0)	//	hyperbolic case
	{
		//	Accept a triangle placement iff the triangle's center
		//	lies below the cutoff circle.
		theReturnValue = (aTriangleCenter[2] < aCoshTilingRadius);
	}
	else
	if (aTriangleCenter[2] == 1.0)	//	flat case (or origin in hyperbolic or spherical case)
	{
		//	Accept a triangle placment iff aTriangleCenter lies
		//	within a disk of radius TILING_PANEL_RADIUS_EUC + aDomainRadius.
		theTilingRadius = TILING_PANEL_RADIUS_EUC + aDomainRadius;
		theReturnValue = (
			aTriangleCenter[0]*aTriangleCenter[0] + aTriangleCenter[1]*aTriangleCenter[1]
		  < theTilingRadius*theTilingRadius);
	}
	else							//	spherical case
	{
		//	In the spherical case, every triangle placements is close enough.
		theReturnValue = true;
	}

	return theReturnValue;
}


static void RenderArchimedeanName(ModelData *md)
{
	const float		*theColor;
	ShaderIndex		theShaderIndex;
	GLint			theShaderProgram;

	static const float	theVertices[4][3] =
						{
							{-1.0, -0.85 - 0.125, 0.0},
							{+1.0, -0.85 - 0.125, 0.0},
							{+1.0, -0.85,         0.0},
							{-1.0, -0.85,         0.0}
						},
						theTexCoords[4][2] =
						{
							{0.0, 0.0},
							{1.0, 0.0},
							{1.0, 1.0},
							{0.0, 1.0}
						};
	static const float	theBlackColor[4]	= {0.0, 0.0, 0.0, 1.0},
						theWhiteColor[4]	= {1.0, 1.0, 1.0, 1.0};

	if (md->itsArchimedeanKey != NULL)
	{
		//	Blend the Archimedean name onto the existing image.
		glEnable(GL_BLEND);
	
		//	Disable Vertex Buffer Objects.
		//	We draw a single quad only once, so it's just as easy
		//	to pass the arrays directly from the CPU.
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

		//	Bind the texture containing the name of the Archimedean solid.
		glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureRegularName]);

		//	Display the name in black and hope for the best,
		//	unless the background is known to be black
		//	(namely when using Stereoscopic 3D)
		//	in which case display the name in white letters.
		if (md->itsStereoMode != StereoNone)
			theColor = theWhiteColor;
		else
			theColor = theBlackColor;

		//	Bind the arrays.
		if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
		{
			theShaderIndex		= ShaderColorWithAlphaMask;
			theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
			glUseProgram(theShaderProgram);

			glUniform4fv(glGetUniformLocation(theShaderProgram, "uniColor"), 1, theColor);

			glEnableVertexAttribArray(ATTRIBUTE_POSITION);
			glVertexAttribPointer(ATTRIBUTE_POSITION,  3, GL_FLOAT, GL_FALSE, 0, theVertices);
				
			glDisableVertexAttribArray(ATTRIBUTE_NORMAL);

			glEnableVertexAttribArray(ATTRIBUTE_TEX_COORD);
			glVertexAttribPointer(ATTRIBUTE_TEX_COORD, 2, GL_FLOAT, GL_FALSE, 0, theTexCoords);

			glDisableVertexAttribArray(ATTRIBUTE_COLOR_HSLA);
		}
		else
		{
			glEnable(GL_TEXTURE_2D);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

			glColor4fv(theColor);

			glEnableClientState(GL_VERTEX_ARRAY);
			glVertexPointer(3, GL_FLOAT, 0, theVertices);

			glDisableClientState(GL_NORMAL_ARRAY);

			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2, GL_FLOAT, 0, theTexCoords);

			glDisableClientState(GL_COLOR_ARRAY);
		}

		//	No fancy transformations are required.
		Matrix44Identity(md->itsModelViewMatrix);
		Matrix44Identity(md->itsProjectionMatrix);
		SendTransformationsToShader(md, false);

		//	Draw the name.
		glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
	}
}


static void RenderControlPanel(ModelData *md)
{
	//	Design Note:  To keep the program logic simple,
	//	redraw the control every frame whether it needs it or not.
	//	The control panel takes only about 10% of the total rendering time,
	//	so it's not worth complicating the program logic for such
	//	a small speed improvement.  Besides, the fundamental triangle
	//	would have to be drawn every frame anyhow.

	ShaderIndex		theShaderIndex;
	GLint			theShaderProgram	= 0;
	float			theVertices[4][3],
					theTexCoords[4][2];
	unsigned int	i,
					j;
	double			theWellOffset[4][4];
	ColorRGBA		theRGBA;
	ColorHSLA		theHSLA;
	float			theColor[4];
	ktVector		theTriplePoint;
	double			x,
					y,
					dx,
					dy;
	ktFacePainting	*thePainting;
	double			thePaintingPanelOffset[4][4],
					theLightness,
					c,
					t,
					theOpacity,
					theHalfWidth,
					theHalfHeight;
	RectFloat		p,
					u;
	float			theHSLAArray[4][4];
	float			theLightnessAlphaColors[6][4] =
					{
						{-1.0, -1.0, -1.0,  0.0},	//	RGB to be overwritten with current hue-saturation
						{ 1.0,  1.0,  1.0,  0.0},
						{ 1.0,  1.0,  1.0,  1.0},
						{-1.0, -1.0, -1.0,  1.0},	//	RGB to be overwritten with current hue-saturation
						{ 0.0,  0.0,  0.0,  1.0},
						{ 0.0,  0.0,  0.0,  0.0}
					};
	double			theTrianglePanelOffset[4][4],
					theTrianglePanelScale[4][4],
					theDomainOffset[4][4];

	static double		theControlPanelTransformation[4][4] =
						{
							{2.0/256.0,    0.0,     0.0, 0.0},
							{   0.0,    2.0/1024.0, 0.0, 0.0},
							{   0.0,       0.0,     0.0, 0.0},
							{  -1.0,      -1.0,     0.0, 1.0}
						};

	static const float	theHueVertices[14][3] =
						{
							{ 0.0 / 6.0, 0.0, 0.0 },
							{ 0.0 / 6.0, 1.0, 0.0 },
							{ 1.0 / 6.0, 0.0, 0.0 },
							{ 1.0 / 6.0, 1.0, 0.0 },
							{ 2.0 / 6.0, 0.0, 0.0 },
							{ 2.0 / 6.0, 1.0, 0.0 },
							{ 3.0 / 6.0, 0.0, 0.0 },
							{ 3.0 / 6.0, 1.0, 0.0 },
							{ 4.0 / 6.0, 0.0, 0.0 },
							{ 4.0 / 6.0, 1.0, 0.0 },
							{ 5.0 / 6.0, 0.0, 0.0 },
							{ 5.0 / 6.0, 1.0, 0.0 },
							{ 6.0 / 6.0, 0.0, 0.0 },
							{ 6.0 / 6.0, 1.0, 0.0 }
						},
						theHueColors[14][4] =
						{
							{1, 0, 0, 1},
							{1, 0, 0, 1},
							{1, 1, 0, 1},
							{1, 1, 0, 1},
							{0, 1, 0, 1},
							{0, 1, 0, 1},
							{0, 1, 1, 1},
							{0, 1, 1, 1},
							{0, 0, 1, 1},
							{0, 0, 1, 1},
							{1, 0, 1, 1},
							{1, 0, 1, 1},
							{1, 0, 0, 1},
							{1, 0, 0, 1}
						},
						theSaturationColors[4][4] =
						{
							{0.5, 0.5, 0.5, 1.0},
							{0.5, 0.5, 0.5, 1.0},
							{0.5, 0.5, 0.5, 0.0},
							{0.5, 0.5, 0.5, 0.0}
						},
						theLightnessAlphaVertices[6][3] =
						{
							{0.5, 0.0, 0.0},
							{1.0, 0.0, 0.0},
							{1.0, 1.0, 0.0},
							{0.5, 1.0, 0.0},
							{0.0, 1.0, 0.0},
							{0.0, 0.0, 0.0}
						};

	//	Blend each successive element onto the existing image.
	//	(Some elements require this and some do not,
	//	but it may be just as fast to leave blending enabled the whole time.)
	glEnable(GL_BLEND);

	//	Disable Vertex Buffer Objects for drawing the various rectangles.
	//	We draw each rectangle only once, so it's just as easy
	//	to pass the arrays directly from the CPU.
	//	Exception:  The code that draws the fundamental triangle
	//	will temporarily re-activate Vertex Buffer Objects.
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	//	For the most part we'll use vertex and texture coordinate arrays,
	//	but not normal vectors or colors.  Individual panels may temporarily
	//	override these defaults, just so they restore them again afterwards.
	//
	//	Use fixed arrays for theVertices and theTexCoords,
	//	changing their contents as needed.
	//
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
		theShaderIndex		= ShaderTexture;
		theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
		glUseProgram(theShaderProgram);

		glEnableVertexAttribArray(ATTRIBUTE_POSITION);
		glVertexAttribPointer(ATTRIBUTE_POSITION,  3, GL_FLOAT, GL_FALSE, 0, theVertices);

		glDisableVertexAttribArray(ATTRIBUTE_NORMAL);

		glEnableVertexAttribArray(ATTRIBUTE_TEX_COORD);
		glVertexAttribPointer(ATTRIBUTE_TEX_COORD, 2, GL_FLOAT, GL_FALSE, 0, theTexCoords);

		glDisableVertexAttribArray(ATTRIBUTE_COLOR_HSLA);
	}
	else
	{
		glEnable(GL_TEXTURE_2D);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, 0, theVertices);

		glDisableClientState(GL_NORMAL_ARRAY);

		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, 0, theTexCoords);

		glDisableClientState(GL_COLOR_ARRAY);
	}
	
	//	Initialize theTexCoords to their default values.
	//	Blocks of code may alter these values,
	//	but they must restore the defaults when they're done.
	SetTextureCoordinates(theTexCoords, 0.0, 0.0, 1.0, 1.0);

	//	Load an orthogonal projection matrix that establishes
	//	a 256×1024 coordinate system with the origin at the lower left.
	//	theControlPanelTransformation also collapses the z-coordinate to zero,
	//	to avoid clipping issues when drawing the fundamental triangle.
	Matrix44Identity(md->itsModelViewMatrix);
	Matrix44Copy(md->itsProjectionMatrix, theControlPanelTransformation);
	SendTransformationsToShader(md, false);

	//	Draw the background
	glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureControlBackground]);
	SetRectangleVertices(theVertices, &(RectFloat){0, 0, 256, 1024});
	glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

	//	Draw the labels.
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
		theShaderIndex		= ShaderColorWithAlphaMask;
		theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
		glUseProgram(theShaderProgram);

		glUniform4f(glGetUniformLocation(theShaderProgram, "uniColor"), 0.0, 0.0, 0.0, 1.0);
	}
	else
	{
		glColor4f(0.0, 0.0, 0.0, 1.0);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	}
	for (i = 0; i < 5; i++)
	{
		glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureLabelBase + i]);
		SetRectangleVertices(theVertices, &gLabelText[i]);
		glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
	}

	//	The symmetry buttons and the style buttons
	//	will use the same shader program.
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
		theShaderIndex		= ShaderColorWithTexture;
		theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
		glUseProgram(theShaderProgram);
	}
	else
	{
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	}

	//	Draw the symmetry buttons.
	for (i = 0; i < 5; i++)
	{
		if (md->itsTiling.itsPQR[0] == 2
		 && md->itsTiling.itsPQR[1] == 3
		 && md->itsTiling.itsPQR[2] == 3 + i)
		{
			//	Even if the face is textured, use its color anyhow.
			theHSLA.h = md->itsFacePainting[2].itsHSLA.h;
			theHSLA.s = md->itsFacePainting[2].itsHSLA.s;
			theHSLA.l = md->itsFacePainting[2].itsHSLA.l;
			theHSLA.a = 1.0;
			HSLAtoRGBA(&theHSLA, &theRGBA);
			theColor[0] = theRGBA.r;
			theColor[1] = theRGBA.g;
			theColor[2] = theRGBA.b;
			theColor[3] = theRGBA.a;
		}
		else
		{
			theColor[0] = 0.75;
			theColor[1] = 0.75;
			theColor[2] = 0.75;
			theColor[3] = 1.00;
		}
		if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
		{
			glUniform4fv(glGetUniformLocation(theShaderProgram, "uniColor"), 1, theColor);
		}
		else
		{
			glColor4fv(theColor);
		}
		glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureSymmetryBase + i]);
		SetRectangleVertices(theVertices, &gSymmetryButton[i]);
		glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
	}

	//	Draw the style buttons.
	for (i = 0; i < 2; i++)
	{
		if ((md->itsTilingStyle == StyleFlat) == (i == 0))
		{
			theColor[0] = 0.00;
			theColor[1] = 1.00;
			theColor[2] = 0.00;
			theColor[3] = 1.00;
		}
		else
		{
			theColor[0] = 0.75;
			theColor[1] = 0.75;
			theColor[2] = 0.75;
			theColor[3] = 1.00;
		}
		if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
		{
			glUniform4fv(glGetUniformLocation(theShaderProgram, "uniColor"), 1, theColor);
		}
		else
		{
			glColor4fv(theColor);
		}
		glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureStyleBase + i]);
		SetRectangleVertices(theVertices, &gStyleButton[i]);
		glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
	}

	//	Draw the color/texture boxes for the faces.
	Matrix44Identity(thePaintingPanelOffset);
	Matrix44Identity(theWellOffset);
	for (i = 0; i < 4; i++)
	{
		thePaintingPanelOffset[3][0] = gFacePanel[i].left;
		thePaintingPanelOffset[3][1] = gFacePanel[i].bottom;
		Matrix44Copy(md->itsModelViewMatrix, thePaintingPanelOffset);
		SendTransformationsToShader(md, false);
		
		thePainting = &md->itsFacePainting[i];

		//	After checking that thePainting->itsStyle is a valid array index
		//	(it always is, but good to check before accessing arrays!)...
		if ((unsigned int)thePainting->itsStyle < NUM_SELECTION_BUTTONS)
		{
			//	Draw a bevel for the selected face paint style button.
			//	Use a 50% opacity rectangle to realize the desired shading.
			glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureBevel]);
			if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
			{
				theShaderIndex		= ShaderTexture;
				theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
				glUseProgram(theShaderProgram);
			}
			else
			{
				glColor4f(1.0, 1.0, 1.0, 1.0);	//	irrelevant but safe
				glEnable(GL_TEXTURE_2D);
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
			}
			SetBevelVertices(theVertices, gFaceButton[thePainting->itsStyle], 3);
			glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

			//	Draw a background for the selected face paint style button.
			theColor[0] = 0.6;
			theColor[1] = 0.6;
			theColor[2] = 0.6;
			theColor[3] = 1.0;
			if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
			{
				theShaderIndex		= ShaderColor;
				theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
				glUseProgram(theShaderProgram);

				glUniform4fv(glGetUniformLocation(theShaderProgram, "uniColor"), 1, theColor);

				glDisableVertexAttribArray(ATTRIBUTE_TEX_COORD);
			}
			else
			{
				glColor4fv(theColor);
				glDisable(GL_TEXTURE_2D);

				glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			}
			SetRectangleVertices(theVertices, &gFaceButton[thePainting->itsStyle]);
			glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
			if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
			{
				glEnableVertexAttribArray(ATTRIBUTE_TEX_COORD);
			}
			else
			{
				glEnable(GL_TEXTURE_2D);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			}
		}
		
		//	The next two items use plain textures.
		if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
		{
			theShaderIndex		= ShaderTexture;
			theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
			glUseProgram(theShaderProgram);
		}
		else
		{
			glColor4f(1.0, 1.0, 1.0, 1.0);	//	irrelevant but safe
			glEnable(GL_TEXTURE_2D);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		}

		//	Draw the face paint style buttons.
		//	(ShaderTexture is already enabled.)
		for (j = 0; j < 5; j++)
		{
			glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureButtonBase + j]);
			SetRectangleVertices(theVertices, &gFaceButton[j]);
			glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
		}
		
		//	If a well is present, draw a bevel for it.
		//	(ShaderTexture is already enabled.)
		if (thePainting->itsStyle != FacePaintInvisible)
		{
			glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureBevel]);
			SetBevelVertices(theVertices, gFacePaintWell, 3);
			glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
		}
		
		//	Draw the contents of the well.
		
		//	Translate the origin to the well's lower-left corner.
		theWellOffset[3][0] = gFacePaintWell.left;
		theWellOffset[3][1] = gFacePaintWell.bottom;
		Matrix44Product(theWellOffset, md->itsModelViewMatrix, md->itsModelViewMatrix);
		
		//	Translate the origin to the well's lower-left corner,
		//	then re-scale to [0,1] × [0,1].
		md->itsModelViewMatrix[0][0] *= gFacePaintWell.right - gFacePaintWell.left;
		md->itsModelViewMatrix[1][1] *= gFacePaintWell.top   - gFacePaintWell.bottom;

		//	Send the new matrix to the GPU.
		SendTransformationsToShader(md, false);
		
		//	We're now drawing to a unit square.
		SetRectangleVertices(theVertices, &(RectFloat){0.0, 0.0, 1.0, 1.0});
	
		//	Draw a color gradient, a texture palette or nothing, as appropriate.
		switch (thePainting->itsStyle)
		{
			case FacePaintColorHS:
			case FacePaintColorLA:

				if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
				{
					theShaderIndex		= ShaderHSLAWithTexture;
					theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
					glUseProgram(theShaderProgram);

					glEnableVertexAttribArray(ATTRIBUTE_COLOR_HSLA);
					glVertexAttribPointer(ATTRIBUTE_COLOR_HSLA, 4, GL_FLOAT, GL_FALSE, 0, theHSLAArray);
					
					if (thePainting->itsStyle == FacePaintColorHS)
					{
						//	Draw a color plane with variable hue and saturation,
						//	but fixed lightness and opacity.

						theHSLAArray[0][0] = 0.0;
						theHSLAArray[0][1] = 0.0;
						theHSLAArray[0][2] = thePainting->itsHSLA.l;
						theHSLAArray[0][3] = thePainting->itsHSLA.a;
						
						theHSLAArray[1][0] = 1.0;
						theHSLAArray[1][1] = 0.0;
						theHSLAArray[1][2] = thePainting->itsHSLA.l;
						theHSLAArray[1][3] = thePainting->itsHSLA.a;
						
						theHSLAArray[2][0] = 1.0;
						theHSLAArray[2][1] = 1.0;
						theHSLAArray[2][2] = thePainting->itsHSLA.l;
						theHSLAArray[2][3] = thePainting->itsHSLA.a;
						
						theHSLAArray[3][0] = 0.0;
						theHSLAArray[3][1] = 1.0;
						theHSLAArray[3][2] = thePainting->itsHSLA.l;
						theHSLAArray[3][3] = thePainting->itsHSLA.a;
					}
					else	//	thePainting->itsStyle == FacePaintColorLA
					{
						//	Draw a color plane with fixed hue and saturation,
						//	but variable lightness and opacity.

						theHSLAArray[0][0] = thePainting->itsHSLA.h;
						theHSLAArray[0][1] = thePainting->itsHSLA.s;
						theHSLAArray[0][2] = 0.0;
						theHSLAArray[0][3] = 0.0;
						
						theHSLAArray[1][0] = thePainting->itsHSLA.h;
						theHSLAArray[1][1] = thePainting->itsHSLA.s;
						theHSLAArray[1][2] = 1.0;
						theHSLAArray[1][3] = 0.0;
						
						theHSLAArray[2][0] = thePainting->itsHSLA.h;
						theHSLAArray[2][1] = thePainting->itsHSLA.s;
						theHSLAArray[2][2] = 1.0;
						theHSLAArray[2][3] = 1.0;
						
						theHSLAArray[3][0] = thePainting->itsHSLA.h;
						theHSLAArray[3][1] = thePainting->itsHSLA.s;
						theHSLAArray[3][2] = 0.0;
						theHSLAArray[3][3] = 1.0;
					}

					//	Use a checkboard texture for the background.
					glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureCheckerboard]);

					//	Use custom texture coordinates for the opacity
					//	to set the number of checkboard repetitions and
					//	to get the aspect ratio right.
					SetTextureCoordinates(theTexCoords, 0.0, 0.0, 5.0, 2.0);
				
					//	Draw the rectangle.
					glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

					//	Restore the default texture coordinates.
					SetTextureCoordinates(theTexCoords, 0.0, 0.0, 1.0, 1.0);

					//	Disable the color array to restore the default state.
					glDisableVertexAttribArray(ATTRIBUTE_COLOR_HSLA);
				}
				else	//	! aShadersFlag
				{
					//	TO ENSURE THAT THE ARCHIVED OPENGL 1.X COPY OF KALEIDO WORKS,
					//	WE NEED TO TEMPORARILY RESTORE THE NON-PREMULTIPLIED ALPHA MODE
					//	TO GET THIS LITTLE HACK TO WORK.
					glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
					
					//	The following fixed-function code is a bit complicated,
					//	but happily it will all go away in a couple years
					//	when I eliminate all fixed-function code in favor
					//	of programmable shaders only.

					if (thePainting->itsStyle == FacePaintColorHS)
					{
						//	Even though the color's green intensity varies
						//	linearly from 0 at (h,s) = (0, 1) up to 1 at (h,s) = (1/6, 1),
						//	and also varies linearly from 1/2 at (h,s) = (0, 0),
						//	across to 1/2 at (h,s) = (1/6, 0), it is *not* linear
						//	over the whole rectangle spanned by those four points.
						//	(This example assumes the lightness l = 1/2 and the
						//	opacity α = 1, just to keep things simple while illustrating
						//	the issue at hand.)  The account for the non-linearity,
						//	we must rending the color well in several passes.
						//	The following code makes 4 passes, for hue, then saturation,
						//	then lightness, then opacity.  It might be possible
						//	to condense things into only 3 passes, but for the moment
						//	I prefer to stick with the 4-pass system because its logic
						//	is simpler.  In any case, this code might not make a huge
						//	difference to KaleidoTile's overall frame rate.

						//	Temporarily disable texturing and enable color interpolation.
						glDisable(GL_TEXTURE_2D);
						glShadeModel(GL_SMOOTH);
						
						//	Enable/disable the appropriate arrays.
						glDisableClientState(GL_TEXTURE_COORD_ARRAY);
						glEnableClientState(GL_COLOR_ARRAY);

						//	Set the vertex and color arrays to realize a rainbow palette.
						glVertexPointer(3, GL_FLOAT, 0, theHueVertices);
						glColorPointer(4, GL_FLOAT, 0, theHueColors);

						//	Draw the hue, which varies horizontally.
						//	Note that the colors are all fully opaque at this point. 
						glDrawArrays(GL_TRIANGLE_STRIP, 0, 14);

						//	Restore the default vertex array.
						glVertexPointer(3, GL_FLOAT, 0, theVertices);
						
						//	The remaining three passes will be simple unit squares.
						SetRectangleVertices(theVertices, &(RectFloat){0.0, 0.0, 1.0, 1.0});
						
						//	Multiply in the saturation, which varies vertically.
						//	In effect the saturation interpolates linearly
						//	from fully saturated colors along the well's top edge
						//	to a 50% grey along the well's bottom edge.
						//	Note that here the color's alpha component is simply
						//	a blending factor -- it has nothing to do with the opacity
						//	of the final color (which we'll get to momentarily).
						glColorPointer(4, GL_FLOAT, 0, theSaturationColors);
						glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

						//	We're done with color interpolation, so disable it.
						glShadeModel(GL_FLAT);
						
						//	We're done with the color array.
						glDisableClientState(GL_COLOR_ARRAY);

						//	Multiply in the lightness, which is a constant value
						//	equal to the lightness of thePainting's current color.
						theLightness = thePainting->itsHSLA.l;
						if (theLightness != 0.5)	//	Ignore the common case of lightness 1/2, which has no effect.
						{
							if (theLightness > 0.5)
							{	//	Lighter than average colors interpolate toward...
								c = 1.0;						//	...pure white
								t = 2.0 * (theLightness - 0.5);	//	...by this factor.
							}
							else
							{	//	Darker  than average colors interpolate toward...
								c = 0.0;						//	...pure black
								t = 2.0 * (0.5 - theLightness);	//	...by this factor.
							}
							
							//	Again the color's alpha component is simply a blending factor.
							glColor4f(c, c, c, t);
							glDrawArrays(GL_TRIANGLE_FAN, 0, 4);	//	unit square still loaded
						}

						//	Re-enable texturing.
						glEnable(GL_TEXTURE_2D);
						
						//	Re-enable the texture array.
						glEnableClientState(GL_TEXTURE_COORD_ARRAY);
						
						//	Multiply in the opacity.
						//	In other words, take the color well as drawn so far
						//	and blend it with a background image,
						//	according to thePainting's current opacity.
						//	Conceptually the background image sits underneath
						//	the color palette, so it's a little counterintuitive
						//	drawing it last.  But technically it's easier this way
						//	because we got to use OpenGL's standard color buffer
						//	(rather than an offscreen buffer) to blend together
						//	the hue, saturation and lightness.  And of course
						//	just so we blend the color palette and the background
						//	image in the right proportions, it doesn't matter
						//	which one we draw first.
						theOpacity = thePainting->itsHSLA.a;
						if (theOpacity != 1.0)	//	Ignore the common case of opacity 1, which has no effect.
						{
							//	Use custom texture coordinates for the opacity
							//	to set the number of checkboard repetitions and
							//	to get the aspect ratio right.
							SetTextureCoordinates(theTexCoords, 0.0, 0.0, 5.0, 2.0);
						
							glColor4f(1.0, 1.0, 1.0, 1.0 - theOpacity);
							glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureCheckerboard]);
							glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
							glDrawArrays(GL_TRIANGLE_FAN, 0, 4);	//	unit square still loaded

							//	Restore the default texture coordinates.
							SetTextureCoordinates(theTexCoords, 0.0, 0.0, 1.0, 1.0);
						}
					}
					else	//	thePainting->itsStyle == FacePaintColorLA
					{
						//	Draw an opaque background pattern.
						SetRectangleVertices(theVertices, &(RectFloat){0.0, 0.0, 1.0, 1.0});	//	unit square
						SetTextureCoordinates(theTexCoords, 0.0, 0.0, 5.0, 2.0);	//	proper checkerboard size and proportions
						glColor4f(1.0, 1.0, 1.0, 1.0);	//	irrelevant but safe
						glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureCheckerboard]);
						glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
						glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
						SetTextureCoordinates(theTexCoords, 0.0, 0.0, 1.0, 1.0);	//	restore default texture coordinates
								
						//	Fill the well with a color whose lightness varies horizontally,
						//	whose opacity varies vertically, and whose hue and saturation
						//	match those of thePainting's current color.
						theHSLA.h = thePainting->itsHSLA.h;
						theHSLA.s = thePainting->itsHSLA.s;
						theHSLA.l = 0.5;
						theHSLA.a = 1.0;
						HSLAtoRGBA(&theHSLA, &theRGBA);
						theLightnessAlphaColors[0][0] = theLightnessAlphaColors[3][0] = theRGBA.r;
						theLightnessAlphaColors[0][1] = theLightnessAlphaColors[3][1] = theRGBA.g;
						theLightnessAlphaColors[0][2] = theLightnessAlphaColors[3][2] = theRGBA.b;
						glDisable(GL_TEXTURE_2D);
						glShadeModel(GL_SMOOTH);
						glDisableClientState(GL_TEXTURE_COORD_ARRAY);
						glEnableClientState(GL_COLOR_ARRAY);
						glVertexPointer(3, GL_FLOAT, 0, theLightnessAlphaVertices);
						glColorPointer(4, GL_FLOAT, 0, theLightnessAlphaColors);
						for (j = 0; j < 2; j++)	//	Repeat to exagerate the opacity [as α(2 - α)] for greater visibility.
							glDrawArrays(GL_TRIANGLE_FAN, 0, 6);
						glVertexPointer(3, GL_FLOAT, 0, theVertices);
						glDisableClientState(GL_COLOR_ARRAY);
						glEnableClientState(GL_TEXTURE_COORD_ARRAY);
						glShadeModel(GL_FLAT);
						glEnable(GL_TEXTURE_2D);
					}

					//	RESTORE PREMULTIPLIED ALPHA MODE.
					glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
				}

				break;

			case FacePaintTextureStandard:

				glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[i < 3 ? TextureFacePalette : TextureBackgroundPalette]);
				if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
				{
					theShaderIndex		= ShaderTexture;
					theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
					glUseProgram(theShaderProgram);
				}
				else
				{
					glColor4f(1.0, 1.0, 1.0, 1.0);	//	irrelevant but safe
					glEnable(GL_TEXTURE_2D);
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
				}
				SetRectangleVertices(theVertices, &(RectFloat){0.0, 0.0, 1.0, 1.0});	//	unit square
				//	Horizontal coordinates are scaled by 5/8 to accommodate 
				//	5 little squares in a texture with power-of-two dimensions.
				SetTextureCoordinates(theTexCoords, 0.0, 0.0, 5.0/8.0, 1.0);
				glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
				SetTextureCoordinates(theTexCoords, 0.0, 0.0, 1.0, 1.0);	//	Restore default texture coordinates.

				break;

			case FacePaintTextureCustom:
			
				//	Prepare for plain texturing.
				if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
				{
					theShaderIndex		= ShaderTexture;
					theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
					glUseProgram(theShaderProgram);
				}
				else
				{
					glColor4f(1.0, 1.0, 1.0, 1.0);	//	irrelevant but safe
					glEnable(GL_TEXTURE_2D);
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
				}
				SetRectangleVertices(theVertices, &(RectFloat){0.0, 0.0, 1.0, 1.0});

				//	Draw an opaque background pattern.
				glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureCheckerboard]);
				SetTextureCoordinates(theTexCoords, 0.0, 0.0, 5.0, 2.0);
				glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
				
				//	Draw a portion of the texture itself
				//	(which may be partially transparent).
				glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureFaceBase + i]);
				theHalfWidth  = 0.5 * thePainting->itsCustomTextureFraction;
				theHalfHeight = (2.0/5.0) * theHalfWidth;
				SetTextureCoordinates(	theTexCoords,
										0.5 - theHalfWidth,
										0.5 - theHalfHeight,
										0.5 + theHalfWidth,
										0.5 + theHalfHeight);
				glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
				
				//	Restore default texture coordinates.
				SetTextureCoordinates(theTexCoords, 0.0, 0.0, 1.0, 1.0);

				break;

			case FacePaintInvisible:
				break;
		}

		//	Draw the marker symbol.
		switch (thePainting->itsStyle)
		{
			case FacePaintColorHS:
			case FacePaintColorLA:

				theHSLA.h = thePainting->itsHSLA.h;
				theHSLA.s = thePainting->itsHSLA.s;
				theHSLA.l = thePainting->itsHSLA.l;
				theHSLA.a = 1.0;
				HSLAtoRGBA(&theHSLA, &theRGBA);
				if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
				{
					theShaderIndex		= ShaderColorWithTexture;
					theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
					glUseProgram(theShaderProgram);

					glUniform4f(glGetUniformLocation(theShaderProgram, "uniColor"), theRGBA.r, theRGBA.g, theRGBA.b, theRGBA.a);
				}
				else
				{
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

					glColor4f(theRGBA.r, theRGBA.g, theRGBA.b, theRGBA.a);
				}
				glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureRoundMarker]);
				if (thePainting->itsStyle == FacePaintColorHS)
				{
					x = thePainting->itsHSLA.h;
					y = thePainting->itsHSLA.s;
				}
				else	//	FacePaintColorLA
				{
					x = thePainting->itsHSLA.l;
					y = thePainting->itsHSLA.a;
				}
				dx = 2 * COLOR_MARKER_SIZE;	//	The 200×80 pixel box has proportions 5:2
				dy = 5 * COLOR_MARKER_SIZE;	//		so scale the marker size inversely.
				p.left		= x - dx;	//	marker position
				p.bottom	= y - dy;
				p.right		= x + dx;
				p.top		= y + dy;
				u.left		= 0.0;		//	marker texture coordinates
				u.bottom	= 0.0;
				u.right		= 1.0;
				u.top		= 1.0;
				ClipToUnitSquare(&p, &u);
				SetRectangleVertices(theVertices, &p);
				SetTextureCoordinates(theTexCoords, u.left, u.bottom, u.right, u.top);
				glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
				SetTextureCoordinates(theTexCoords, 0.0, 0.0, 1.0, 1.0);	//	Restore default texture coordinates.
				
				break;

			case FacePaintTextureStandard:
			
				//	ShaderTexture is already set.

				dx = 1.0/5.0;
				dy = 1.0/2.0;

				x = dx * ((thePainting->itsStandardTextureID - 1) % 5);
				y = dy * (thePainting->itsStandardTextureID <= 5 ? 1 : 0);

				SetRectangleVertices(theVertices, &(RectFloat){x, y, x + dx, y + dy});
				glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureSquareMarker]);
				glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

				break;
			
			case FacePaintTextureCustom:
			case FacePaintInvisible:
				break;
		}
					
		//	Note:  We're leaving the modelview matrix set 
		//	to the well's normalized coordinate system,
		//	but that's OK because we're going to reset it anyhow,
		//	whether or not we're done with the loop.
	}


	//	Draw the fundamental triangle.

	//	Pretending we're in curved mode, set theFundamentalTrianglePlacement
	//	so that the fundamental domain just fits within a 240×240(?) box
	//	centered in gTrianglePanel (which is a 256×256 box).
	//	Fold these changes into itsProjectionMatrix.
	//	Let DrawTheSceneWithParity() set itsModelViewMatrix to its satisfaction,
	//	to ensure proper normal vector processing (if any).

	Matrix44Identity(theTrianglePanelOffset);
	theTrianglePanelOffset[3][0] = 0.5 * (gTrianglePanel.left + gTrianglePanel.right);
	theTrianglePanelOffset[3][1] = 0.5 * (gTrianglePanel.bottom + gTrianglePanel.top);
	Matrix44Product(theTrianglePanelOffset, md->itsProjectionMatrix, md->itsProjectionMatrix);
	
	Matrix44Identity(theTrianglePanelScale);
	theTrianglePanelScale[0][0] = (0.5*TRIANGLE_PANEL_WIDTH) / md->itsTiling.itsDomainScale;
	theTrianglePanelScale[1][1] = (0.5*TRIANGLE_PANEL_WIDTH) / md->itsTiling.itsDomainScale;
	Matrix44Product(theTrianglePanelScale, md->itsProjectionMatrix, md->itsProjectionMatrix);

	Matrix44Identity(theDomainOffset);
	theDomainOffset[3][0] = md->itsTiling.itsDomainOffset;
	Matrix44Product(theDomainOffset, md->itsProjectionMatrix, md->itsProjectionMatrix);

	//	Draw the fundamental triangle.
	//	DrawTheSceneWithParity() will call SendTransformationsToShader(),
	//	and will also temporarily reactivate Vertex Buffer Objects.
	DrawTheSceneWithParity(md, NULL, false, false, true);

	//	Re-disable Vertex Buffer Objects.
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	
	//	Prepare to draw a plain texture.
	//	Restore default array usage.
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
		theShaderIndex		= ShaderTexture;
		theShaderProgram	= md->itsShaderPrograms[theShaderIndex];
		glUseProgram(theShaderProgram);

		glEnableVertexAttribArray(ATTRIBUTE_POSITION);
		glVertexAttribPointer(ATTRIBUTE_POSITION,  3, GL_FLOAT, GL_FALSE, 0, theVertices);

		glDisableVertexAttribArray(ATTRIBUTE_NORMAL);

		glEnableVertexAttribArray(ATTRIBUTE_TEX_COORD);
		glVertexAttribPointer(ATTRIBUTE_TEX_COORD, 2, GL_FLOAT, GL_FALSE, 0, theTexCoords);

		glDisableVertexAttribArray(ATTRIBUTE_COLOR_HSLA);
	}
	else
	{
		glEnable(GL_TEXTURE_2D);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, 0, theVertices);

		glDisableClientState(GL_NORMAL_ARRAY);

		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, 0, theTexCoords);

		glDisableClientState(GL_COLOR_ARRAY);
	}

	//	Draw the control point.
	//	itsProjectionMatrix remains set from above.
	//	Set itsModelViewMatrix to the identity.
	Matrix44Identity(md->itsModelViewMatrix);
	SendTransformationsToShader(md, false);
	glBindTexture(GL_TEXTURE_2D, md->itsTextureNames[TextureRoundMarker]);
	VectorTimesMatrix(	md->itsTriplePointPosition,
						md->itsTiling.itsBarycentricBasis,
						theTriplePoint);
	VectorNormalize(theTriplePoint, GetMetric(md->itsTiling.itsGeometry, MetricVertical));
	x = theTriplePoint[0];
	y = theTriplePoint[1];
	dx = TRIANGLE_MARKER_SIZE * md->itsTiling.itsDomainScale;
	dy = TRIANGLE_MARKER_SIZE * md->itsTiling.itsDomainScale;
	SetRectangleVertices(theVertices, &(RectFloat){x - dx, y - dy, x + dx, y + dy});
	glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
}

static void SetRectangleVertices(
	float			aVertexArray[4][3],	//	output
	const RectFloat	*aRectangle)		//	input
{
	aVertexArray[0][0] = aRectangle->left;
	aVertexArray[0][1] = aRectangle->bottom;
	aVertexArray[0][2] = 0.0;

	aVertexArray[1][0] = aRectangle->right;
	aVertexArray[1][1] = aRectangle->bottom;
	aVertexArray[1][2] = 0.0;

	aVertexArray[2][0] = aRectangle->right;
	aVertexArray[2][1] = aRectangle->top;
	aVertexArray[2][2] = 0.0;

	aVertexArray[3][0] = aRectangle->left;
	aVertexArray[3][1] = aRectangle->top;
	aVertexArray[3][2] = 0.0;
}

static void SetTextureCoordinates(
	float	aTexCoordArray[4][2],	//	output
	float	aLeftSide,				//	input
	float	aBottomSide,			//	input
	float	aRightSide,				//	input
	float	aTopSide)				//	input
{
	aTexCoordArray[0][0] = aLeftSide;
	aTexCoordArray[0][1] = aBottomSide;

	aTexCoordArray[1][0] = aRightSide;
	aTexCoordArray[1][1] = aBottomSide;

	aTexCoordArray[2][0] = aRightSide;
	aTexCoordArray[2][1] = aTopSide;

	aTexCoordArray[3][0] = aLeftSide;
	aTexCoordArray[3][1] = aTopSide;
}

static void SetBevelVertices(
	float			aVertexPointer[4][3],	//	output
	RectFloat		aRect,					//	input
	unsigned int	aWidth)					//	input
{
	//	Filling the whole rectangle in order to create a border
	//	seems a little wasteful.  On the other hand, this code is simpler
	//	than code that would create the border using lines,
	//	and avoids issues concerning what lines widths are supported.
	//	(Also, hardware support for polygons may be better than for lines???)
	
	//	Expand aRect to enclose the desired border.
	aRect.left		-= aWidth;
	aRect.bottom	-= aWidth;
	aRect.right		+= aWidth;
	aRect.top		+= aWidth;
	
	//	Transfer aRect to aVertexPointer[][], 
	//	which we assume has already been enabled.
	SetRectangleVertices(aVertexPointer, &aRect);
}


static void HSLAtoRGBA(
	ColorHSLA	*anHSLA,	//	input
	ColorRGBA	*anRGBA)	//	output
{
	double	h,	//	hue
			s,	//	saturation
			l,	//	lightness
			r,	//	red
			g,	//	green
			b,	//	blue
			a,	//	alpha (opaqueness)
			t,
			c;

	//	Basic error checking.
	if (anHSLA->h < 0.0 || anHSLA->h > 1.0
	 || anHSLA->s < 0.0 || anHSLA->s > 1.0
	 || anHSLA->l < 0.0 || anHSLA->l > 1.0
	 || anHSLA->a < 0.0 || anHSLA->a > 1.0)
	{
		anRGBA->r = 0.0;
		anRGBA->g = 0.0;
		anRGBA->b = 0.0;
		anRGBA->a = 1.0;

		return;
	}

	//	For easy readability.
	h = anHSLA->h;
	s = anHSLA->s;
	l = anHSLA->l;
	a = anHSLA->a;

	//	To understanding what's going on here, make yourself a sketch
	//	of the unit cube in red-green-blue (RGB) space.

	//	First create a "pure color" of the given hue.
	//	The "pure colors" lie along the hexagon with vertices at
	//		(1,0,0)		(pure red)
	//		(1,1,0)		(pure yellow)
	//		(0,1,0)		(pure green)
	//		(0,1,1)		(pure cyan)
	//		(0,0,1)		(pure blue)
	//		(1,0,1)		(pure magenta)
	//	on the color cube.

	h *= 6.0;	//	for convenience in the six different cases

	if (h < 1.0)
	{
		r = 1.0;
		g = h;
		b = 0.0;
	}
	else
	if (h < 2.0)
	{
		r = 2.0 - h;
		g = 1.0;
		b = 0.0;
	}
	else
	if (h < 3.0)
	{
		r = 0.0;
		g = 1.0;
		b = h - 2.0;
	}
	else
	if (h < 4.0)
	{
		r = 0.0;
		g = 4.0 - h;
		b = 1.0;
	}
	else
	if (h < 5.0)
	{
		r = h - 4.0;
		g = 0.0;
		b = 1.0;
	}
	else
	{
		r = 1.0;
		g = 0.0;
		b = 6.0 - h;
	}

	//	To take into account the saturation, average the pure hue
	//	with a medium grey (0.5, 0.5, 0.5).
	r = s*r + (1.0 - s)*0.5;
	g = s*g + (1.0 - s)*0.5;
	b = s*b + (1.0 - s)*0.5;

	//	To take into account the lightness, average the color
	//	with pure white or pure black.
	if (l > 0.5)
	{
		t = 2.0 * (1.0 - l);
		c = 1.0;
	}
	else
	{
		t = 2.0 * (l - 0.0);
		c = 0.0;
	}
	r = t*r + (1.0 - t)*c;
	g = t*g + (1.0 - t)*c;
	b = t*b + (1.0 - t)*c;
	
	//	ColorRGBA requires premultiplied alpha.
	r *= a;
	g *= a;
	b *= a;

	//	Copy the results to the output variable.
	anRGBA->r = r;
	anRGBA->g = g;
	anRGBA->b = b;
	anRGBA->a = a;
}


static void ClipToUnitSquare(
	RectFloat	*p,	//	a quad's position on unit square
	RectFloat	*u)	//	a quad's texture coordinates
{
	//	Assume the quad's center lies within the unit square (possibly on the boundary)
	//	so in particular nothing too degenerate can happen.

	if (p->left < 0.0)
	{
		u->left = u->right + (u->left - u->right) * (0.0 - p->right) / (p->left - p->right);
		p->left = 0.0;
	}
	if (p->bottom < 0.0)
	{
		u->bottom = u->top + (u->bottom - u->top) * (0.0 - p->top) / (p->bottom - p->top);
		p->bottom = 0.0;
	}
	if (p->right > 1.0)
	{
		u->right = u->left + (u->right - u->left) * (1.0 - p->left) / (p->right - p->left);
		p->right = 1.0;
	}
	if (p->top > 1.0)
	{
		u->top = u->bottom + (u->top - u->bottom) * (1.0 - p->bottom) / (p->top - p->bottom);
		p->top = 1.0;
	}
}


MouseTarget InterpretMousePosition(
	double	*aMouseH,		//	input and output
	double	*aMouseV,		//	input and output
	double	aDomainScale,	//	input, used only for fundamental domain hits
	double	aDomainOffset)	//	input, used only for fundamental domain hits
{
	unsigned int	i,
					j;
	
	//	Determine which subpanel (*aMouseH, *aMouseV) lies in,
	//	and adjust (*aMouseH, *aMouseV) to the subpanel's local coordinates.
	
	//	Clamp input to [0,1].
	if (*aMouseH < 0.0)
		*aMouseH = 0.0;
	if (*aMouseH > 1.0)
		*aMouseH = 1.0;
	if (*aMouseV < 0.0)
		*aMouseV = 0.0;
	if (*aMouseV > 1.0)
		*aMouseV = 1.0;
	
	//	Interpret the mouse position.
	if (*aMouseH <= 0.8)	//	In the left 80% of the display?
	{
		//	main tiling panel
		
		//	Convert horizontal coordinate from [0, 4/5] to [0,1].
		*aMouseH *= 1.25;
		
		//	Convert coordinates from [0,1] to [-1,+1].
		*aMouseH = (*aMouseH * 2.0) - 1.0;
		*aMouseV = (*aMouseV * 2.0) - 1.0;
		
		return MouseTargetTiling;
	}
	else	//	In the right 20% of the display.
	{
		//	control panel
		
		//	Convert coordinates from [4/5, 1] × [0, 1] to [0, 256] × [0, 1024].
		*aMouseH = (*aMouseH * 5.0 - 4.0) * 256.0;
		*aMouseV *= 1024.0;
		
		//	symmetry buttons
		for (i = 0; i < 5; i++)
		{
			if (PointInRectangle(*aMouseH, *aMouseV, &gSymmetryButton[i]))
			{
				*aMouseH = 0.0;
				*aMouseV = 0.0;
				return MouseTargetSymmetryBase + i;
			}
		}
		
		//	style buttons
		if (PointInRectangle(*aMouseH, *aMouseV, &gStyleButton[0]))
		{
			*aMouseH = 0.0;
			*aMouseV = 0.0;
			return MouseTargetStyleFlat;
		}
		if (PointInRectangle(*aMouseH, *aMouseV, &gStyleButton[1]))
		{
			*aMouseH = 0.0;
			*aMouseV = 0.0;
			return MouseTargetStyleCurved;
		}
		
		//	fundamental triangle
		if (PointInRectangle(*aMouseH, *aMouseV, &gTrianglePanel))
		{
			*aMouseH -= (gTrianglePanel.left + gTrianglePanel.right) / 2.0;
			*aMouseV -= (gTrianglePanel.bottom + gTrianglePanel.top) / 2.0;

			*aMouseH *= aDomainScale / (TRIANGLE_PANEL_WIDTH/2.0);
			*aMouseV *= aDomainScale / (TRIANGLE_PANEL_WIDTH/2.0);

			*aMouseH -= aDomainOffset;

			return MouseTargetTriangle;
		}
		
		//	color panel / background panel
		for (i = 0; i < 4; i++)
		{
			if (PointInRectangle(*aMouseH, *aMouseV, &gFacePanel[i]))
			{
				//	Convert to panel coordinates.
				*aMouseH -= gFacePanel[i].left;
				*aMouseV -= gFacePanel[i].bottom;
				
				//	Selection button hit?
				for (j = 0; j < NUM_SELECTION_BUTTONS; j++)
				{
					if (PointInRectangle(*aMouseH, *aMouseV, &gFaceButton[j]))
					{
						*aMouseH = 0.0;
						*aMouseV = 0.0;
						
						return (MouseTarget)(MouseTargetSelectBase + i*NUM_SELECTION_BUTTONS + j);
					}
				}
				
				//	Color/texture palette hit?
				if (PointInRectangle(*aMouseH, *aMouseV, &gFacePaintWell))
				{
					*aMouseH -= gFacePaintWell.left;
					*aMouseV -= gFacePaintWell.bottom;
					
					*aMouseH /= (gFacePaintWell.right - gFacePaintWell.left);
					*aMouseV /= (gFacePaintWell.top - gFacePaintWell.bottom);
					
					return (MouseTarget)(MouseTargetPaintBase + i);
				}
				
				//	The click hit the panel's background.
				return MouseTargetNone;
			}
		}
		
		return MouseTargetNone;
	}
}

static bool PointInRectangle(
	const double	h,
	const double	v,
	const RectFloat	*aRect)
{
	return h >= aRect->left
		&& h <= aRect->right
		&& v >= aRect->bottom
		&& v <= aRect->top;
}


ErrorText GetErrorString(void)
{
	switch (glGetError())
	{
		case GL_NO_ERROR:
			return NULL;
		case GL_INVALID_ENUM:
			return "GL_INVALID_ENUM:  GLenum argument out of range.";
		case GL_INVALID_VALUE:
			return "GL_INVALID_VALUE:  Numeric argument out of range.";
		case GL_INVALID_OPERATION:
			return "GL_INVALID_OPERATION:  Operation illegal in current state.";
		case GL_STACK_OVERFLOW:
			return "GL_STACK_OVERFLOW:  Command would cause a stack overflow.";
		case GL_STACK_UNDERFLOW:
			return "GL_STACK_UNDERFLOW:  Command would cause a stack underflow.";
		case GL_OUT_OF_MEMORY:
			return "GL_OUT_OF_MEMORY:  Not enough memory left to execute command.";
		default:
			return "Unknown OpenGL error.";
	}
}
