//	KaleidoTileMouse.c
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "KaleidoTile-Common.h"
#include <math.h>


//	How many seconds must elapse to compute a reasonable velocity?
#define MIN_MOUSE_TIME				0.1

//	How many seconds must elapse for the mouse to be considered still?
#define FREEZE_TIME					0.3

//	Does let the tiling or triple point velocities get too extreme.
#define TILING_SPEED_LIMIT			0.4
#define TRIPLE_POINT_SPEED_LIMIT	0.2

//	How close must the triple point be to a regular point
//	in order to snap to it?
#define SNAP_TO_REGULAR_EPSILON		0.1

//	If we snap to a regular point and show the polyhedron's name,
//	how many seconds should we show it for?
#define REGULAR_NAME_DURATION		4.0


//	In the case of a spherical tiling, the user sees the southern hemisphere
//	in the main tiling window, but the northern hemisphere (near the north pole (0,0,1))
//	in the control panel's fundamental triangle.
typedef enum
{
	HemisphereSouthern,
	HemisphereNorthern
} Hemisphere;


static void	ConvertClickPoint(double x, double y, ktGeometry aGeometry, Hemisphere aHemisphere, ktVector aPoint);
static void	SnapToRegular(ModelData *md, double aTimeStamp);
static void	SortPQR(unsigned int anUnsortedPQR[3], unsigned int aSortedPQR[3]);
static void	PaintPanelClick(ktFacePainting *aPainting, double aMouseH, double aMouseV, bool aBackgroundImageFlag, bool *aPreparedTexturesFlag);


void MouseDown(
	ModelData	*md,
	double		aMouseH,	//	horizontal coordinate ∈ [0,1] (directed rightward)
	double		aMouseV,	//	 vertical  coordinate ∈ [0,1] (directed upward)
	double		aTimeStamp)	//	in seconds since system startup
{
	ktVector			theTriplePoint;
	unsigned int		i,
						thePaintingIndex,
						thePaintStyleIndex;
	ktFacePainting		*thePainting;
	ktFacePaintStyle	thePaintStyle;

	//	InterpretMousePosition() locates the subpanel in which the mouse click falls,
	//	and adjusts (aMouseH, aMouseV) to the subpanel's local coordinates.
	md->itsMouseTarget = InterpretMousePosition(&aMouseH, &aMouseV, md->itsTiling.itsDomainScale, md->itsTiling.itsDomainOffset);

	switch (md->itsMouseTarget)
	{
		case MouseTargetTiling:

			//	Store one copy for keeping the tiling or triple point position
			//	in sync with the mouse, which happens every time the mouse moves.
			md->itsPreviousMousePH = aMouseH;
			md->itsPreviousMousePV = aMouseV;

			//	Store a second copy for updating the velocity,
			//	which happens only after MIN_MOUSE_TIME seconds.
			md->itsPreviousMouseVH = aMouseH;
			md->itsPreviousMouseVV = aMouseV;

			//	Note the starting time.
			md->itsPreviousMouseVTime = aTimeStamp;

			//	Initialize the drag.
			md->itsTilingVelocity		= (ktVelocity) {0.0, 0.0, 0.0};
			md->itsNewTilingVelocity	= (ktVelocity) {0.0, 0.0, 0.0};

			break;
		
		case MouseTargetSymmetry3:
			ChangeTiling(md, 2, 3, 3);
			break;
		
		case MouseTargetSymmetry4:
			ChangeTiling(md, 2, 3, 4);
			break;
		
		case MouseTargetSymmetry5:
			ChangeTiling(md, 2, 3, 5);
			break;
		
		case MouseTargetSymmetry6:
			ChangeTiling(md, 2, 3, 6);
			break;
		
		case MouseTargetSymmetry7:
			ChangeTiling(md, 2, 3, 7);
			break;
		
		case MouseTargetStyleFlat:
			ChangeStyle(md, StyleFlat);
			break;
		
		case MouseTargetStyleCurved:
			ChangeStyle(md, StyleCurved);
			break;
		
		case MouseTargetTriangle:

			//	Store one copy for keeping the tiling or triple point position
			//	in sync with the mouse, which happens every time the mouse moves.
			md->itsPreviousMousePH = aMouseH;
			md->itsPreviousMousePV = aMouseV;

			//	Store a second copy for updating the velocity,
			//	which happens only after MIN_MOUSE_TIME seconds.
			md->itsPreviousMouseVH = aMouseH;
			md->itsPreviousMouseVV = aMouseV;

			//	Note the starting time.
			md->itsPreviousMouseVTime = aTimeStamp;

			//	Initialize the drag.

			ConvertClickPoint(	md->itsPreviousMousePH,
								md->itsPreviousMousePV,
								md->itsTiling.itsGeometry,
								HemisphereNorthern,
								theTriplePoint);
			ConvertToBarycentric(	theTriplePoint,
									md->itsTiling.itsBarycentricBasis,
									md->itsTriplePointPosition);
			ClampToFundamentalTriangle(md->itsTriplePointPosition);

			for (i = 0; i < 3; i++)
			{
				md->itsTriplePointVelocity[i]		= 0.0;
				md->itsNewTriplePointVelocity[i]	= 0.0;
			}

			break;
		

		case MouseTargetSelectAColorHS:
		case MouseTargetSelectAColorLA:
		case MouseTargetSelectATextureStandard:
		case MouseTargetSelectATextureCustom:
		case MouseTargetSelectAInvisible:
		case MouseTargetSelectBColorHS:
		case MouseTargetSelectBColorLA:
		case MouseTargetSelectBTextureStandard:
		case MouseTargetSelectBTextureCustom:
		case MouseTargetSelectBInvisible:
		case MouseTargetSelectCColorHS:
		case MouseTargetSelectCColorLA:
		case MouseTargetSelectCTextureStandard:
		case MouseTargetSelectCTextureCustom:
		case MouseTargetSelectCInvisible:
		case MouseTargetSelectBackgroundColorHS:
		case MouseTargetSelectBackgroundColorLA:
		case MouseTargetSelectBackgroundTextureStandard:
		case MouseTargetSelectBackgroundTextureCustom:
		case MouseTargetSelectBackgroundInvisible:
		
			thePaintingIndex	= (md->itsMouseTarget - MouseTargetSelectABase) / NUM_SELECTION_BUTTONS;
			thePaintStyleIndex	= (md->itsMouseTarget - MouseTargetSelectABase) % NUM_SELECTION_BUTTONS;
			
			thePainting		= &md->itsFacePainting[thePaintingIndex];
			thePaintStyle	= (ktFacePaintStyle)thePaintStyleIndex;
			
			//	Adjust the paint style accordingly.
			switch (thePaintStyle)
			{
				case FacePaintColorHS:
					SetFaceColorHS(thePainting, -1.0, -1.0, -1.0, -1.0);
					break;

				case FacePaintColorLA:
					SetFaceColorLA(thePainting, -1.0, -1.0, -1.0, -1.0);
					break;

				case FacePaintTextureStandard:
					SetFaceTextureStandard(thePainting, 0, &md->itsPreparedTextures);
					break;

				case FacePaintTextureCustom:
					SetFaceTextureCustom(thePainting, thePaintingIndex == 3, false, &md->itsPreparedTextures);
					break;

				case FacePaintInvisible:
					SetFaceInvisible(&md->itsFacePainting[thePaintingIndex]);
					break;
			}
	
			break;
		
		case MouseTargetPaintA:
		case MouseTargetPaintB:
		case MouseTargetPaintC:
		case MouseTargetPaintBackground:
			PaintPanelClick(&md->itsFacePainting[md->itsMouseTarget - MouseTargetPaintBase], aMouseH, aMouseV, md->itsMouseTarget == MouseTargetPaintBackground, &md->itsPreparedTextures);
			break;
		
		default:
			break;
	}
	
	//	Ask the idle-time routine to redraw the scene.
	md->itsRedrawRequestFlag = true;
}


void MouseMove(
	ModelData	*md,
	double		aMouseH,	//	horizontal coordinate ∈ [0,1] (directed rightward)
	double		aMouseV,	//	 vertical  coordinate ∈ [0,1] (directed upward)
	double		aTimeStamp)	//	in seconds since system startup
{
	ktVector		theOldPoint,
					theNewPoint,
					theTriplePoint,
					theOldBarycentricPoint,
					theNewBarycentricPoint;
	ktMatrix		theIncrement,
					thePosition;
	double			theElapsedTime,
					theFactor;
	ktFacePainting	*thePainting;

	//	InterpretMousePosition() locates the subpanel in which the mouse cursor lies,
	//	and adjusts (aMouseH, aMouseV) to the subpanel's local coordinates.
	//	Mouse motion outside the original subpanel (where the click began)
	//	gets ignored.
	if (InterpretMousePosition(&aMouseH, &aMouseV, md->itsTiling.itsDomainScale, md->itsTiling.itsDomainOffset) != md->itsMouseTarget)
	{
		//	If we wanted to terminate the drag, we'd set
		//		md->itsMouseTarget = MouseTargetNone;
		//	before returning.
		return;
	}

	switch (md->itsMouseTarget)
	{
		case MouseTargetTiling:

			//	If the mouse has moved, update the polygon's position
			//	(whether or not the clock has advanced).
			if (aMouseH != md->itsPreviousMousePH
			 || aMouseV != md->itsPreviousMousePV)
			{
				ConvertClickPoint(	md->itsPreviousMousePH,
									md->itsPreviousMousePV,
									md->itsTiling.itsGeometry,
									HemisphereSouthern,
									theOldPoint);
				ConvertClickPoint(	aMouseH,
									aMouseV,
									md->itsTiling.itsGeometry,
									HemisphereSouthern,
									theNewPoint);
				MatrixParallelTransport(theOldPoint, theNewPoint, theIncrement, md->itsTiling.itsGeometry);

				IntegrateVelocity(1.0, md->itsTilingPosition, md->itsTiling.itsGeometry, thePosition);

				MatrixProduct(thePosition, theIncrement, thePosition);

				DifferentiatePosition(thePosition, md->itsTiling.itsGeometry, &md->itsTilingPosition);

				md->itsPreviousMousePH = aMouseH;
				md->itsPreviousMousePV = aMouseV;
			}

			//	If at least MIN_MOUSE_TIME seconds have elapsed,
			//	compute a tiling velocity (whether or not the mouse has moved).
			theElapsedTime = aTimeStamp - md->itsPreviousMouseVTime;
			if (theElapsedTime > MIN_MOUSE_TIME)
			{
				//	How did the tiling move during theElapsedTime?
				ConvertClickPoint(	md->itsPreviousMouseVH,
									md->itsPreviousMouseVV,
									md->itsTiling.itsGeometry,
									HemisphereSouthern,
									theOldPoint);
				ConvertClickPoint(	aMouseH,
									aMouseV,
									md->itsTiling.itsGeometry,
									HemisphereSouthern,
									theNewPoint);
				MatrixParallelTransport(theOldPoint, theNewPoint, theIncrement, md->itsTiling.itsGeometry);

				//	What velocity, integrated over unit time,
				//	would account for the given motion?
				DifferentiatePosition(theIncrement, md->itsTiling.itsGeometry, &md->itsNewTilingVelocity);

				//	Rescale the velocity so that it achieves the given motion
				//	over theElapsedTime (rather than over unit time).
				theFactor = 1.0 / theElapsedTime;
				md->itsNewTilingVelocity.x *= theFactor;
				md->itsNewTilingVelocity.y *= theFactor;
				md->itsNewTilingVelocity.r *= theFactor;

				//	Prepare for the next iteration.
				md->itsPreviousMouseVH		= aMouseH;
				md->itsPreviousMouseVV		= aMouseV;
				md->itsPreviousMouseVTime	= aTimeStamp;
			}

			break;
		
		case MouseTargetTriangle:

			//	If the mouse has moved, update the triple point
			//	(whether or not the clock has advanced).
			if (aMouseH != md->itsPreviousMousePH
			 || aMouseV != md->itsPreviousMousePV)
			{
				ConvertClickPoint(	aMouseH,
									aMouseV,
									md->itsTiling.itsGeometry,
									HemisphereNorthern,
									theTriplePoint);
				ConvertToBarycentric(	theTriplePoint,
										md->itsTiling.itsBarycentricBasis,
										md->itsTriplePointPosition);
				ClampToFundamentalTriangle(md->itsTriplePointPosition);

				md->itsPreviousMousePH = aMouseH;
				md->itsPreviousMousePV = aMouseV;
			}

			//	If at least MIN_MOUSE_TIME milliseconds have elapsed,
			//	compute a triple point velocity (whether or not the mouse has moved).
			//
			//	Note.  We compute itsNewTriplePointVelocity here,
			//	but we don't actually install it as itsTriplePointVelocity
			//	until the MouseUp event arrives.  Instead we keep
			//	itsTriplePointVelocity at zero and manually move
			//	itsTriplePointPosition in direct response to mouse motion.
			theElapsedTime = aTimeStamp - md->itsPreviousMouseVTime;
			if (theElapsedTime > MIN_MOUSE_TIME)
			{
				//	How did the triple point move during theElapsedTime?

				ConvertClickPoint(	md->itsPreviousMouseVH,
									md->itsPreviousMouseVV,
									md->itsTiling.itsGeometry,
									HemisphereNorthern,
									theOldPoint);
				ConvertToBarycentric(	theOldPoint,
										md->itsTiling.itsBarycentricBasis,
										theOldBarycentricPoint);
				ClampToFundamentalTriangle(theOldBarycentricPoint);

				ConvertClickPoint(	aMouseH,
									aMouseV,
									md->itsTiling.itsGeometry,
									HemisphereNorthern,
									theNewPoint);
				ConvertToBarycentric(	theNewPoint,
										md->itsTiling.itsBarycentricBasis,
										theNewBarycentricPoint);
				ClampToFundamentalTriangle(theNewBarycentricPoint);

				//	Compute itsNewTriplePointVelocity.
				VectorDifference(	theNewBarycentricPoint,
									theOldBarycentricPoint,
									md->itsNewTriplePointVelocity);

				//	Rescale itsNewTriplePointVelocity to give the motion per unit time.
				ConstantTimesVector(1.0 / theElapsedTime,
									md->itsNewTriplePointVelocity,
									md->itsNewTriplePointVelocity);

				//	Prepare for the next iteration.
				md->itsPreviousMouseVH		= aMouseH;
				md->itsPreviousMouseVV		= aMouseV;
				md->itsPreviousMouseVTime	= aTimeStamp;
			}

			break;
		
		case MouseTargetPaintA:
		case MouseTargetPaintB:
		case MouseTargetPaintC:
		case MouseTargetPaintBackground:
			//	Only the Hue-Saturation and Lightness-Opacity panels respond to mouse drags.
			thePainting = &md->itsFacePainting[md->itsMouseTarget - MouseTargetPaintBase];
			if (thePainting->itsStyle == FacePaintColorHS
			 || thePainting->itsStyle == FacePaintColorLA)
				PaintPanelClick(thePainting, aMouseH, aMouseV, md->itsMouseTarget == MouseTargetPaintBackground, &md->itsPreparedTextures);
			break;

		default:
			break;
	}
	
	//	Ask the idle-time routine to redraw the scene.
	md->itsRedrawRequestFlag = true;
}


void MouseUp(
	ModelData	*md,
	double		aMouseH,	//	horizontal coordinate ∈ [0,1] (directed rightward)
	double		aMouseV,	//	 vertical  coordinate ∈ [0,1] (directed upward)
	double		aTimeStamp)	//	in seconds since system startup
{
	double		theSpeed;

	UNUSED_PARAMETER(aMouseH);
	UNUSED_PARAMETER(aMouseV);

	switch (md->itsMouseTarget)
	{
		case MouseTargetTiling:

			//	If the mouse has been still for more than FREEZE_TIME seconds,
			//	the velocity should be set to zero.  [MouseMove() won't have noticed,
			//	because MouseMove() gets called only when the mouse is *not* still.]
			if (aTimeStamp - md->itsPreviousMouseVTime > FREEZE_TIME)
				md->itsNewTilingVelocity = (ktVelocity) {0.0, 0.0, 0.0};

			//	Enforce a speed limit, so the tiling doesn't spin wildly.
			theSpeed = sqrt(md->itsNewTilingVelocity.x * md->itsNewTilingVelocity.x
						  + md->itsNewTilingVelocity.y * md->itsNewTilingVelocity.y
						  + md->itsNewTilingVelocity.r * md->itsNewTilingVelocity.r);
			if (theSpeed > TILING_SPEED_LIMIT)
			{
				md->itsNewTilingVelocity.x *= TILING_SPEED_LIMIT / theSpeed;
				md->itsNewTilingVelocity.y *= TILING_SPEED_LIMIT / theSpeed;
				md->itsNewTilingVelocity.r *= TILING_SPEED_LIMIT / theSpeed;
			}

			//	Install the new tiling velocity.
			md->itsTilingVelocity = md->itsNewTilingVelocity;

			break;

		case MouseTargetTriangle:

			//	If the mouse has been still for more than FREEZE_TIME milliseconds,
			//	the triple point velocity should be set to zero.  [MouseMove() won't have noticed,
			//	because MouseMove() gets called only when the mouse is *not* still.]
			if (aTimeStamp - md->itsPreviousMouseVTime > FREEZE_TIME)
			{
				md->itsNewTriplePointVelocity[0] = 0.0;
				md->itsNewTriplePointVelocity[1] = 0.0;
				md->itsNewTriplePointVelocity[2] = 0.0;
			}

			//	Enforce a speed limit, so the triple point doesn't bounce around wildly.
			theSpeed = VectorLength(md->itsNewTriplePointVelocity, MetricEuclidean);
			if (theSpeed > TRIPLE_POINT_SPEED_LIMIT)
			{
				md->itsNewTriplePointVelocity[0] *= TRIPLE_POINT_SPEED_LIMIT / theSpeed;
				md->itsNewTriplePointVelocity[1] *= TRIPLE_POINT_SPEED_LIMIT / theSpeed;
				md->itsNewTriplePointVelocity[2] *= TRIPLE_POINT_SPEED_LIMIT / theSpeed;
			}

			//	Install the new triple point velocity.
			VectorCopy(md->itsTriplePointVelocity, md->itsNewTriplePointVelocity);

			//	If the triple point is near a vertex, edge midpoint or center,
			//	snap to it and zero out the velocity.
			if (md->itsSnapToArchimedeanFlag)
				SnapToRegular(md, aTimeStamp);

			break;

		default:
			break;
	}
	
	//	Ask the idle-time routine to redraw the scene.
	md->itsRedrawRequestFlag = true;
}


static void ConvertClickPoint(
	double		x,				//	input horizontal coordinate in intrinsic units
	double		y,				//	input  vertical  coordinate in intrinsic units
	ktGeometry	aGeometry,		//	input geometry
	Hemisphere	aHemisphere,	//	input desired hemisphere (relevant only in spherical geometry)
	ktVector	aPoint)			//	output point on S², E² or H²
{
	double	theLength;

	theLength = sqrt(x*x + y*y);

	switch (aGeometry)
	{
		case GeometrySpherical:
			if (theLength < 1.0)
			{
				aPoint[0] = x;
				aPoint[1] = y;
				aPoint[2] = (aHemisphere == HemisphereNorthern ? +1.0 : - 1.0)
							* sqrt(1.0 - theLength*theLength);
			}
			else
			{
				aPoint[0] = x / theLength;
				aPoint[1] = y / theLength;
				aPoint[2] = 0.0;
			}
			break;

		case GeometryEuclidean:
			aPoint[0] = x;
			aPoint[1] = y;
			aPoint[2] = 1.0;
			break;

		case GeometryHyperbolic:
			aPoint[0] = x;
			aPoint[1] = y;
			aPoint[2] = sqrt(1.0 + theLength*theLength);
			break;
	}
}


static void SnapToRegular(
	ModelData	*md,
	double		aTimeStamp)
{
	unsigned int	i;
	ktVector		theErrorVector;
	unsigned int	theSortedPQR[3],
					theFamilyIndex;

	static ktVector	theRegularPoints[7] =
					{
						{1.0/3.0, 1.0/3.0, 1.0/3.0},	//	center
						{0.0, 0.5, 0.5},				//	edge midpoints
						{0.5, 0.0, 0.5},
						{0.5, 0.5, 0.0},
						{1.0, 0.0, 0.0},				//	vertices
						{0.0, 1.0, 0.0},
						{0.0, 0.0, 1.0},
					};
	static const char	*thePolyhedronName[3][7] =
					{
						{
							"truncated octahedron",
							"cuboctahedron",
							"truncated tetrahedron",
							"truncated tetrahedron",
							"octahedron",
							"tetrahedron",
							"tetrahedron"
						},
						{
							"truncated cuboctahedron",
							"rhombicuboctahedron",
							"truncated octahedron",
							"truncated cube",
							"cuboctahedron",
							"cube",
							"octahedron"
						},
						{
							"truncated icosidodecahedron",
							"rhombicosidodecahedron",
							"truncated icosahedron",
							"truncated dodecahedron",
							"icosidodecahedron",
							"dodecahedron",
							"icosahedron"
						}
					};
	static ktVector	theNullVelocity = {0.0, 0.0, 0.0};


	for (i = 0; i < 7; i++)
	{
		VectorDifference(	md->itsTriplePointPosition,
							theRegularPoints[i],
							theErrorVector);

		if (VectorLength(theErrorVector, MetricEuclidean) < SNAP_TO_REGULAR_EPSILON)
		{
			VectorCopy(md->itsTriplePointPosition, theRegularPoints[i]);
			VectorCopy(md->itsTriplePointVelocity, theNullVelocity);

			SortPQR(md->itsTiling.itsPQR, theSortedPQR);

			if (theSortedPQR[0] == 2
			 && theSortedPQR[1] == 3
			 && theSortedPQR[2] >= 3
			 && theSortedPQR[2] <= 5)
			{
				theFamilyIndex = theSortedPQR[2] - 3;

				md->itsArchimedeanKey					= thePolyhedronName[theFamilyIndex][i];
				md->itsArchimedeanNameDisplayEndTime	= aTimeStamp + REGULAR_NAME_DURATION;
				md->itsPreparedTextures					= false;
			}

			PlayWAV("click.wav");
		}
	}
}


static void SortPQR(
	unsigned int	anUnsortedPQR[3],	//	input
	unsigned int	aSortedPQR[3])		//	output
{
	unsigned int	i,
					j,
					theSwapValue;

	for (i = 0; i < 3; i++)
		aSortedPQR[i] = anUnsortedPQR[i];

	for (i = 0; i < 3; i++)
		for (j = i + 1; j < 3; j++)
			if (aSortedPQR[j] < aSortedPQR[i])
			{
				theSwapValue	= aSortedPQR[j];
				aSortedPQR[j]	= aSortedPQR[i];
				aSortedPQR[i]	= theSwapValue;
			}
}


static void PaintPanelClick(
	ktFacePainting	*aPainting,
	double			aMouseH,
	double			aMouseV,
	bool			aBackgroundImageFlag,	//	true for background images, false for face images
	bool			*aPreparedTexturesFlag)
{
	unsigned int	theRow,
					theCol;

	//	Handle the click according to the paint style.
	switch (aPainting->itsStyle)
	{
		case FacePaintColorHS:
			SetFaceColorHS(aPainting, aMouseH, aMouseV, -1.0, -1.0);
			break;
		
		case FacePaintColorLA:
			SetFaceColorLA(aPainting, -1.0, -1.0, aMouseH, aMouseV);
			break;
		
		case FacePaintTextureStandard:

			theRow = (unsigned int) floor(2.0 * (1.0 - aMouseV));
			if (theRow > 1)
				theRow = 1;

			theCol = (unsigned int) floor(5.0 * aMouseH);
			if (theCol > 4)
				theCol = 4;

			SetFaceTextureStandard(aPainting, 1 + 5*theRow + theCol, aPreparedTexturesFlag);

			break;
		
		case FacePaintTextureCustom:
			SetFaceTextureCustom(aPainting, aBackgroundImageFlag, true, aPreparedTexturesFlag);
			break;
		
		case FacePaintInvisible:
			break;
	}
}
