//	KaleidoTileUtilities.c
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "KaleidoTile-Common.h"
#include <math.h>	//	for floor()


static inline double	Luminance(double aRedComponent, double aGreenComponent, double aBlueComponent);


ErrorText GetAbsolutePath(
	const char		*aDirectory,	//	input,  zero-terminated UTF-8 string, may be NULL
	const char		*aFileName,		//	input,  zero-terminated UTF-8 string, may be NULL
	char			*aPath,			//	output, zero-terminated UTF-8 string
	unsigned int	aPathSize)		//	output buffer size, including space for terminating zero
{
	//	Assemble an absolute path of the form
	//
	//		<base path>/<directory name>/<file name>
	//
	//	where
	//
	//		<base path> says where the application lives,
	//			and won't be known until runtime,
	//
	//		<directory name> specifies the directory of interest,
	//			for example "Languages", "Sounds" or "Textures".
	//
	//		<file name> specifies the particular file,
	//			for example "KaleidoTile-ja.txt" or "Clouds.rgba".

	char	*theTerminator,
			theChar;

	//	Get the base path.
	if ( ! GetBasePath(aPath, aPathSize) )
	{
		aPath[0] = 0;
		return "Couldn't get base path in GetAbsolutePath()";
	}
	
	//	Locate the terminating zero (which GetBasePath() guarantees).
	theTerminator = aPath;
	while (*theTerminator != 0)
		theTerminator++;

	//	Append aDirectory.
	if (aDirectory != NULL)
	{
		while ((theChar = *aDirectory++) != 0)
		{
			*theTerminator = theChar;
			if (++theTerminator < aPath + aPathSize)
				*theTerminator = 0;
			else
			{
				aPath[0] = 0;
				return "Path name too long (1) in  in GetAbsolutePath()";
			}
		}
		
		//	Append a path separator.
		*theTerminator = '/';
		if (++theTerminator < aPath + aPathSize)
			*theTerminator = 0;
		else
		{
			aPath[0] = 0;
			return "Path name too long (2) in  in GetAbsolutePath()";
		}
	}

	//	Append aFileName.
	if (aFileName != NULL)
	{
		while ((theChar = *aFileName++) != 0)
		{
			*theTerminator = theChar;
			if (++theTerminator < aPath + aPathSize)
				*theTerminator = 0;
			else
			{
				aPath[0] = 0;
				return "Path name too long (3) in  in GetAbsolutePath()";
			}
		}
	}

	return NULL;
}


void ConvertColorToGreyscale(
	ColorRGBA	*anRGBAColor)
{
	double	theLuminance;

	theLuminance = Luminance(anRGBAColor->r, anRGBAColor->g, anRGBAColor->b);

	anRGBAColor->r = theLuminance;
	anRGBAColor->g = theLuminance;
	anRGBAColor->b = theLuminance;
}

void ConvertPixelsToGreyscale(
	PixelRGBA		*aPixelArray,
	unsigned int	aLength)
{
	unsigned int	i;
	Byte			theLuminance;

	for (i = 0; i < aLength; i++)
	{
		theLuminance = (Byte) floor(0.5 +
			Luminance((double)aPixelArray[i].r, (double)aPixelArray[i].g, (double)aPixelArray[i].b) );

		aPixelArray[i].r = theLuminance;
		aPixelArray[i].g = theLuminance;
		aPixelArray[i].b = theLuminance;
	}
}

static inline double Luminance(
	double	aRedComponent,
	double	aGreenComponent,
	double	aBlueComponent)
{
	//	The formula
	//
	//		luminance = 30% red + 59% green + 11% blue
	//
	//	appears widely on the internet, but with little explanation.
	//	Presumably its origins lie in human color perception.

	return 0.30 * aRedComponent
		 + 0.59 * aGreenComponent
		 + 0.11 * aBlueComponent;
}
