//	KaleidoTile-MacOS-Utilities.c
//
//	A few platform-dependent functions that the platform-independent code will need.
//	(Along with a couple global variables whose use I'll need to ponder further.)
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#import <Cocoa/Cocoa.h>
#import "KaleidoTile-Common.h"
#import "KaleidoTileMacOSUtilities.h"


//	Define a SoundStopper object to release NSSound objects when they're done playing.
@interface SoundStopper : NSObject
{
}
- (id)initWithSound:(NSSound *)aSound;
- (void)sound:(NSSound *)sound didFinishPlaying:(BOOL)finishedPlaying;
@end
@implementation SoundStopper
- (id)initWithSound:(NSSound *)aSound
{
	self = [super init];
	if (self != nil)
	{
		[aSound retain];
		[self retain];
	}
	return self;
}
- (void)sound:(NSSound *)sound didFinishPlaying:(BOOL)finishedPlaying
{
	[sound setDelegate:nil];
	[sound release];
	[self release];
}
@end


NSString *GetLocalizedTextAsNSString(
	const char	*aKey)
{
	return [NSString stringWithUTF8String:GetLocalizedText(aKey)];
}


bool GetBasePath(
	char			*aPath,
	unsigned int	aPathSize)
{
	return [[[[NSBundle mainBundle] resourcePath] stringByAppendingString:@"/"]
		getCString:aPath maxLength:aPathSize encoding:NSUTF8StringEncoding];
}


ErrorText GetFileContents(
	const char		*aPathName,		//	input, zero-terminated UTF-8 string containing absolute path name
	unsigned int	*aNumRawBytes,	//	output, the file size in bytes
	Byte			**someRawBytes)	//	output, the file's contents as raw bytes
{
	//	Technical Note:  I had hoped that setlocale() + fopen() would
	//	let me read the file within the platform-independent part
	//	of the code, but alas on WinXP it works when aPathName 
	//	contains only ASCII characters but fails when aPathName 
	//	contains non-trivial Unicode characters, for example when
	//	I temporarily change "Textures" to "Textures日本".
	//	So I reluctantly resorted to platform-dependent code
	//	to read the file.

	NSData	*theFileContents;
	
	//	Read the file.
	theFileContents = [NSData dataWithContentsOfFile:[NSString stringWithUTF8String:aPathName]];
	if (theFileContents == nil)
		return "Couldn't get file contents as NSData in GetFileContents()";
	
	//	Note the number of bytes.
	*aNumRawBytes = [theFileContents length];
	if (*aNumRawBytes == 0)
		return "File contains 0 bytes in GetFileContents()";

	//	Copy the data.  The caller will own this copy, and must
	//	eventually free it with a call to FreeRawTextureBytes().
	*someRawBytes = GET_MEMORY(*aNumRawBytes);
	if (*someRawBytes == NULL)
	{
		*aNumRawBytes = 0;
		return "Couldn't allocate buffer in GetFileContents()";
	}
	[theFileContents getBytes:(void *)(*someRawBytes) length:(*aNumRawBytes)];
	
	//	Success!
	return NULL;
}

void FreeFileContents(
	unsigned int	*aNumRawBytes,
	Byte			**someRawBytes)
{
	*aNumRawBytes = 0;
	FREE_MEMORY_SAFELY(*someRawBytes);
}


void SetAlphaTextureFromString(
	unsigned int	aTextureName,
	const char		*aString,		//	null-terminated UTF-8 string
	unsigned int	aWidth,
	unsigned int	aHeight,
	char			*aFontName,		//	null-terminated ASCII string
	unsigned int	aFontSize,		//	height in pixels, excluding descent
	unsigned int	aFontDescent,	//	vertical space below baseline, in pixels
	bool			aCenteringFlag,	//	Center the text horizontally in the texture image?
	unsigned int	aMargin,		//	If aCenteringFlag == false, what horizontal inset to use?
	bool			aShadersFlag,
	ErrorText		*aFirstError)
{
	ErrorText				theErrorMessage			= NULL;
	Byte					*theAlphaBytes			= NULL;
	CGContextRef			theBitmapContext		= NULL;
	UniCharCount			theRunLengthArray[1]	= {kATSUToTextEnd};
	ATSUStyle				theStyleArray[1]		= {NULL};
	Fixed					theFontSize				= 0,	//	Overwritten below.
							theBaseline				= 0;	//	Overwritten below.
	ATSUFontID				theFontID				= 0;	//	Overwritten below.
	unsigned int			theNumValidStyleTags	= 0;	//	Overwritten below.
	ATSUAttributeTag		theStyleTags[2]			= {kATSUSizeTag,  kATSUFontTag      };
	UInt32					theStyleValueSizes[2]	= {sizeof(Fixed), sizeof(ATSUFontID)};
	ATSUAttributeValuePtr	theStyleValues[2]		= {&theFontSize,  &theFontID        };
	ATSUTextLayout			theATSUTextLayout		= NULL;
	ATSUAttributeTag		theLayoutTags[1]		= {kATSUCGContextTag};
	UInt32					theLayoutValueSizes[1]	= {sizeof(CGContextRef)};
	ATSUAttributeValuePtr	theLayoutValues[1]		= {&theBitmapContext};
	Rect					theImageRect			= {0, 0, 0, 0};
	NSString				*theString				= nil;
	unsigned int			theStringLength			= 0;
	unichar					*theStringUTF16			= NULL;
	float					theOffset				= 0;

	//	Assumes desired OpenGL context is already active.
	
	//	Note:  We could in theory set up the ATSUI structures once and for all, in advance,
	//	rather than creating them fresh each time SetAlphaTextureFromString() gets called.
	//	However,
	//		1.	the present code is fast enough (~3ms on my iMac) and, more importantly,
	//		2.	most of the time (~2ms on my iMac) is spent in the call to SetTextureImageFancy(),
	//	so there's little to gain by pre-initializing the ATSUI structures.
	//	Instead, let's keep the code simple, with everything right here in one place.

	//	OpenGL (and the underlying graphics card) likes textures with power-of-two dimensions.
	if ( ! IsPowerOfTwo(aWidth)
	 ||  ! IsPowerOfTwo(aHeight) )
	{
		theErrorMessage = "String texture dimensions must be powers of two.";
		goto CleanUpSetAlphaTextureFromString;
	}

	//	Allocate temporary memory for the bitmapped image.
	//	The image will be an alpha-only image with 1 byte per pixel (no RGB!).
	theAlphaBytes = (Byte *) GET_MEMORY( aWidth * aHeight * 1 );
	if (theAlphaBytes == NULL)
	{
		theErrorMessage = "Couldn't allocate theAlphaBytes.";
		goto CleanUpSetAlphaTextureFromString;
	}
	
	//	Create a Core Graphics context using theAlphaBytes.
	theBitmapContext = CGBitmapContextCreate(
							theAlphaBytes,
							aWidth,
							aHeight,
							8,
							aWidth*1,
							NULL,
							kCGImageAlphaOnly);
	if (theBitmapContext == NULL)
	{
		theErrorMessage = "Couldn't create theBitmapContext.";
		goto CleanUpSetAlphaTextureFromString;
	}
	
	//	Even though Core Graphics and ATSUI both use a consistent coordinate system
	//	with the y-coordinate running bottom-to-top, it seems that CGBitmapContextCreate()
	//	runs the rows of the buffer (theAlphaBytes) top-to-bottom, apparently
	//	for compatibility with the old QuickDraw coordinate system.  (For details, go to
	//		http://developer.apple.com/samplecode/TypeServicesForUnicode/listing1.html
	//	and search for "upside down".)
	//	To simulate a bottom-to-top image, let's flip the coordinates.
	//	(The other alternative, if this strategy ever becomes problematic,
	//	would be to manually swap the rows of the image after all drawing is complete.)
	CGContextScaleCTM(theBitmapContext, 1, -1);
	CGContextTranslateCTM(theBitmapContext, 0, -((float)aHeight));
	
	//	Clear the frame.
	CGContextClearRect(theBitmapContext, (CGRect) { {0, 0}, {aWidth, aHeight} } );

	//	Fill the frame with partial transparency, so we can see where we're drawing
	//	(for use during development only!).
//	CGContextSetRGBFillColor(theBitmapContext, 0.0, 0.0, 0.0, 0.25);
//	CGContextFillRect(theBitmapContext, (CGRect) { {0, 0}, {aWidth, aHeight} } );
	
	//	Set full opacity for drawing the text.
	CGContextSetRGBFillColor(theBitmapContext, 0.0, 0.0, 0.0, 1.0);

	//	ATSUI wants theFontSize in points, where 1pt = 1/72in.
	//	We are given aSize in pixels, with no notion of inches anywhere in sight.
	//	Traditionally one assumed 72 pixels/inch (so 1pt = 1px)
	//	but that assumption seems shaky.  Hope for the best.

	//	ATSUI wants the font height excluding the descent,
	//	so we must manually fudge things a bit to allow room for descenders.
	theFontSize	= FloatToFixed(aFontSize);
	theBaseline	= FloatToFixed(aFontDescent);
	
	//	Get an ATSUFontID for the desired font.
	//	If ATSUFindFontFromName() fails to find the requested font, 
	//	push onward without a font specification
	//	and let ATSUCreateTextLayoutWithTextPtr() fall back
	//	to its default font, Helvetica.
	(void) ATSUFindFontFromName(	aFontName,
									strlen(aFontName),
									kFontFullName,
									kFontNoPlatform,
									kFontNoScriptCode,
									kFontNoLanguageCode,
									&theFontID);
	theNumValidStyleTags = (theFontID != kATSUInvalidFontID) ? 2 : 1;

	//	Create a style that specifies the font size.
	//	We've already initialized theStyleValues to contain a pointer to theFontSize.
	if (ATSUCreateStyle(&theStyleArray[0]) != noErr)
	{
		theErrorMessage = "Couldn't create the ATSU Style object.";
		goto CleanUpSetAlphaTextureFromString;
	}
	if (ATSUSetAttributes(
			theStyleArray[0],
			theNumValidStyleTags,
			theStyleTags,
			theStyleValueSizes,
			theStyleValues)
		!= noErr)
	{
		theErrorMessage = "Couldn't set ATSU Style attributes.";
		goto CleanUpSetAlphaTextureFromString;
	}
	
	//	Convert aString from UTF-8 to UTF-16.
	theString		= [NSString stringWithUTF8String:aString];
	theStringLength	= [theString length];
	theStringUTF16	= (unichar *) GET_MEMORY(theStringLength * sizeof(unichar));
	if (theStringUTF16 == NULL)
	{
		theErrorMessage = "Couldn't allocate theStringUTF16.";
		goto CleanUpSetAlphaTextureFromString;
	}
	[theString getCharacters:theStringUTF16];

	//	Create the main text layout structure.
	if (ATSUCreateTextLayoutWithTextPtr(
			theStringUTF16,
			kATSUFromTextBeginning,
			kATSUToTextEnd,
			theStringLength,
			1,	//	1 style run
			theRunLengthArray,
			theStyleArray,
			&theATSUTextLayout)
		!= noErr)
	{
		theErrorMessage = "Couldn't create theATSUTextLayout.";
		goto CleanUpSetAlphaTextureFromString;
	}
	
	//	If the default font can't handle aString,
	//	let ATSUI find a font that can.
	(void) ATSUSetTransientFontMatching(theATSUTextLayout, true);
	
	//	Assign theBitmapContext to theATSUTextLayout.
	//	We've already initialized theLayoutValues to contain a pointer to theBitmapContext.
	if (ATSUSetLayoutControls(
			theATSUTextLayout,
			1,
			theLayoutTags,
			theLayoutValueSizes,
			theLayoutValues)
		!= noErr)
	{
		theErrorMessage = "ATSUSetLayoutControls() failed.";
		goto CleanUpSetAlphaTextureFromString;
	}
	
	//	Center the text?  Or use a predetermined margin?
	if (aCenteringFlag)
	{
		//	Measure the width of the string as it will be rendered,
		//	so we can center it correctly.
		if (ATSUMeasureTextImage(
				theATSUTextLayout,
				kATSUFromTextBeginning,
				kATSUToTextEnd,
				FloatToFixed(0),
				FloatToFixed(0),
				&theImageRect)
			!= noErr)
		{
			theErrorMessage = "ATSUMeasureTextImage() failed.";
			goto CleanUpSetAlphaTextureFromString;
		}
		
		//	Calculate the necessary offset.
		theOffset = 0.5 * (aWidth - theImageRect.right);
		if (theOffset < 0.0)
			theOffset = 0.0;
	}
	else
		theOffset = (float) aMargin;

	//	Draw the string into theBitmapContext.
	if (ATSUDrawText(
			theATSUTextLayout,
			kATSUFromTextBeginning,
			kATSUToTextEnd,
			FloatToFixed(theOffset),
			theBaseline)
		!= noErr)
	{
		theErrorMessage = "ATSUDrawText() failed.";
		goto CleanUpSetAlphaTextureFromString;
	}

	//	Convert the raw bitmap image to an OpenGL alpha-only texture.
	SetTextureImageFancy(aTextureName, aWidth, aHeight, 1, theAlphaBytes, aShadersFlag);

CleanUpSetAlphaTextureFromString:

	if (theATSUTextLayout != NULL)
		ATSUDisposeTextLayout(theATSUTextLayout);
	
	if (theStyleArray[0] != NULL)
		ATSUDisposeStyle(theStyleArray[0]);

	if (theBitmapContext != NULL)
		CGContextRelease(theBitmapContext);

	FREE_MEMORY_SAFELY(theStringUTF16);
	FREE_MEMORY_SAFELY(theAlphaBytes);

	if (theErrorMessage != NULL)
	{
		//	Substitute a transparent texture for the missing one.
		SetTextureImageFancy(aTextureName, 1, 1, 1, (Byte [1]){0x00}, aShadersFlag);
		
		//	Report this error iff an error report is desired
		//	and no earlier error has occurred.
		if (aFirstError != NULL
		 && *aFirstError == NULL)
		{
			*aFirstError = theErrorMessage;
		}

		//	Let the caller push on without the desired texture...
	}
}


bool GetImageToPaste(
	unsigned int	*aWidth,
	unsigned int	*aHeight,
	PixelRGBA		**aPixelArray)
{
	unsigned int		i;
	NSBitmapImageRep	*theImageRep;
	PixelRGBA			*theDstPixel;
	Byte				*theSrcData,
						*theSrcRow,
						*theSrcByte;
	unsigned int		theRowStride,
						thePixelStride;
	bool				theNoPremultFlag;
	unsigned int		theRow,
						theCol;
	float				theAlphaFraction;
	
	static bool		theFractionsAreInitialized	= false;
	static float	theFractions[256];

	//	One-time initialization.
	if ( ! theFractionsAreInitialized )
	{
		for (i = 0; i < 256; i++)
			theFractions[i] = i / 255.0;
		theFractionsAreInitialized = true;
	}

	//	Load the image from the pasteboard.
	theImageRep = [NSBitmapImageRep imageRepWithPasteboard:[NSPasteboard generalPasteboard]];
	if (theImageRep == nil)
		return false;

	//	Test for unexpected formats.
	if ([theImageRep samplesPerPixel] != ( [theImageRep hasAlpha] ? 4 : 3 ) )
	{
		ErrorMessage("Unexpected number of color channels per pixel (was expecting 4-component RGBA or 3-component RGB)", "Paste Error");
		return false;
	}
	if ([theImageRep bitmapFormat] & NSAlphaFirstBitmapFormat)
	{
		ErrorMessage("Unexpected alpha-first byte order", "Paste Error");
		return false;
	}
	if ([theImageRep bitmapFormat] & NSFloatingPointSamplesBitmapFormat)
	{
		ErrorMessage("Unexpected floating point format", "Paste Error");
		return false;
	}
	if ([theImageRep bitsPerSample] != 8)
	{
		ErrorMessage("Unexpected number of bits per color channel (was expecting 8 bits per channel)", "Paste Error");
		return false;
	}
	if ([theImageRep samplesPerPixel] > 1 && [theImageRep isPlanar])
	{
		ErrorMessage("Unexpected separation of color planes (was expecting contiguous RGB or RGBA values for each pixel)", "Paste Error");
		return false;
	}
	
	*aWidth		= [theImageRep pixelsWide];
	*aHeight	= [theImageRep pixelsHigh];

	*aPixelArray = GET_MEMORY((*aWidth) * (*aHeight) * sizeof(PixelRGBA));
	if (aPixelArray == NULL)
	{
		ErrorMessage("Couldn't get memory to copy custom texture from clipboard.", "Out of Memory");
		*aWidth		= 0;
		*aHeight	= 0;
		return false;
	}

	//	Copy the pixels from theImageRep into *aPixelArray.
	//
	//	Note #1
	//		theImageRep stores the scan lines top-down,
	//		but aPixelArray will want them bottom-up,
	//		so we must reverse the row order.
	//
	//	Note #2
	//		The row stride might include some padding.
	//
	theSrcData			= [theImageRep bitmapData];
	theRowStride		= [theImageRep bytesPerRow];
	thePixelStride		= [theImageRep samplesPerPixel];
	theNoPremultFlag	= ([theImageRep bitmapFormat] & NSAlphaNonpremultipliedBitmapFormat);
	theSrcRow			= &theSrcData[(*aHeight) * theRowStride];
	theDstPixel			= *aPixelArray;
	for (theRow = *aHeight; theRow-- > 0; )	//	order is moot here -- it's just a count
	{
		theSrcRow -= theRowStride;	//	note descending order
		theSrcByte = theSrcRow;
		
		for (theCol = *aWidth; theCol-- > 0; )
		{
			theDstPixel->r = *theSrcByte++;
			theDstPixel->g = *theSrcByte++;
			theDstPixel->b = *theSrcByte++;

			if (thePixelStride == 4)
			{
				theDstPixel->a = *theSrcByte++;
				
				if (theNoPremultFlag
				 && theDstPixel->a != 0xFF)
				{
					theAlphaFraction = theFractions[theDstPixel->a];
					theDstPixel->r *= theAlphaFraction;
					theDstPixel->g *= theAlphaFraction;
					theDstPixel->b *= theAlphaFraction;
				}
			}
			else	//	thePixelStride == 3
				theDstPixel->a = 0xFF;

			theDstPixel++;
		}
	}
	
	return true;
}


void FreeImageToPaste(PixelRGBA **aPixelArray)
{
	FREE_MEMORY_SAFELY(*aPixelArray);
}


void PlayWAV(
	char	*aRelativePath)	//	zero-terminated UTF-8 string with path separator '/'
{
	NSSound	*theSound	= nil;

	//	Create theSound.
	theSound = [[[NSSound alloc]
					initWithContentsOfFile:	[[[[NSBundle mainBundle] resourcePath]
							stringByAppendingString:@"/Sounds/"]
							stringByAppendingString:[NSString stringWithUTF8String:aRelativePath]]
					byReference:			YES] autorelease];
	
	//	We'll need to release theSound after it has finished playing.
	//	To know when it's finished, create a SoundStopper object
	//	to receive the sound:didFinishPlaying: message.
	//	Cocoa objects do not retain their delegates,
	//	so the SoundStopper must retain both theSound and itself,
	//	and must release both theSound and itself when
	//	theSound has finished playing.
	[theSound setDelegate:[[[SoundStopper alloc] initWithSound:theSound] autorelease]];
	
	//	Start theSound playing.
	[theSound play];
}


bool GetUserPrefBool(
	char	*aKey)
{
	//	KaleidoTileMacOSView's initWithFrame method should have already
	//	set a default value, so aKey's value should always exist.
	return [[NSUserDefaults standardUserDefaults] boolForKey:[NSString stringWithUTF8String:aKey]];
}

void SetUserPrefBool(
	char	*aKey,
	bool	aValue)
{
	[[NSUserDefaults standardUserDefaults] setBool:aValue forKey:[NSString stringWithUTF8String:aKey]];
}


void ErrorMessage(
	ErrorText	aMessage,	//	UTF8
	ErrorText	aTitle)		//	UTF8
{
	//	NSRunAlertPanel works better than NSBeginAlertSheet
	//	because the former blocks while the latter does not.
	//	This is especially important when displaying a message
	//	during the shutdown sequence (for example the memory-leak message),
	//	when the application would quit out from under any pending
	//	NSBeginAlertSheet() calls.

	NSRunAlertPanel(	[NSString stringWithUTF8String:aTitle  ],
						[NSString stringWithUTF8String:aMessage],
						@"OK",
						NULL,
						NULL);
}
