//	KaleidoTileMacOSView.m
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#import "KaleidoTileMacOSView.h"
#import "KaleidoTileMacOSUtilities.h"

//	When displaying the frame rate in the window's title bar,
//	how often should we update it?
#define FRAME_RATE_UPDATE_PERIOD	1000000		//	time interval in microseconds

//	Virtual key codes for the spacebar and arrow keys.
#define	     RETURN_KEY		 36
#define        SPACEBAR		 49
#define	     ESCAPE_KEY		 53
#define  LEFT_ARROW_KEY		123
#define	RIGHT_ARROW_KEY		124
#define  DOWN_ARROW_KEY		125
#define    UP_ARROW_KEY		126


@implementation KaleidoTileMacOSView


+ (NSOpenGLPixelFormat*)pixelFormatSupportingMultisampling:(bool)aMultisampleFlag
{
	NSOpenGLPixelFormatAttribute	thePlainPixelFormatAttributes[] =
									{
										NSOpenGLPFADoubleBuffer,

										NSOpenGLPFADepthSize,			0,
										NSOpenGLPFAAlphaSize,			8,

										(NSOpenGLPixelFormatAttribute)	0
									},
									theMultisamplePixelFormatAttributes[] =
									{
										NSOpenGLPFADoubleBuffer,

										NSOpenGLPFADepthSize,			0,
										NSOpenGLPFAAlphaSize,			8,

										//	Request multisampling support.
										NSOpenGLPFASampleBuffers,		1,
										NSOpenGLPFASamples,				8,	//	Just guess a large number and let the driver reduce it?

										(NSOpenGLPixelFormatAttribute)	0
									};
	
	return [[[NSOpenGLPixelFormat alloc] initWithAttributes:(aMultisampleFlag ? theMultisamplePixelFormatAttributes : thePlainPixelFormatAttributes)] autorelease];
}

- (id)initWithFrame:(NSRect)frameRect
{
	//	Pass an arbitrary seed to the random number generator,
	//	so SetUpModelData() can initialize the polyhedron
	//	with a different rotation velocity, etc., each time
	//	we start the program.
	srand(TickCount());

	//	Create a fallback set of user defaults (the "factory settings").
	//	These will get used if and only if the user has not provided his/her own values.
	[[NSUserDefaults standardUserDefaults] registerDefaults:
		[NSDictionary dictionaryWithObjectsAndKeys:
			@"yes", @"multisampling",
			@"yes", @"shaders",
			@"no",	@"fallback message given",
			nil
		]
	];

	//	Set harmless defaults for the ModelData,
	//	without assuming the presence of OpenGL or anything else.
	//	In particular, set all pointers to NULL so that
	//	windowShouldClose() can call ShutDownModelData() safely.
	SetUpModelData(&md);

	//	Initialize the superclass.
	//	Request a multisample buffer iff InitWindowData() set md.itsMultisamplingEnabledFlag to true.
	//	The graphics card may or may not support multisampling, but it's always safe to ask.
	self = [super initWithFrame: frameRect
					pixelFormat: [KaleidoTileMacOSView pixelFormatSupportingMultisampling: md.itsMultisamplingEnabledFlag ]];
	if (self != nil)
	{
		//	What time are we starting at?
		Microseconds( (UnsignedWide *) &itsLastRenderTime    );
		Microseconds( (UnsignedWide *) &itsLastFrameRateTime );
		
		//	Allocate and initialize the TIFF file directory name.
		itsLastImageDirectory	= [[NSMutableString stringWithCapacity:512] retain];
		[itsLastImageDirectory setString:NSHomeDirectory()];
		
		//	Allocate the saved window title (but don't set it just yet).
		itsDefaultWindowTitle	= [[NSMutableString stringWithCapacity:32] retain];
		
		//	Initialize itsTimer to nil, then...
		itsTimer = nil;
		
		//	...set the swap parameter and the timer.
		//	SetUpModelData() has already set a default value for itsSyncFlag.
		[self setRedrawWithSync:md.itsSyncFlag];
		
		//	Load custom cursors.
		//	When Mac OS X eventually moves to a resolution-independent interface,
		//	I'll probably want to either update these cursor images,
		//	or revert to Apple's built-in closedHandCursor, openHandCursor
		//	and pointingHandCursor.  For now I prefer these over Apple's defaults
		//	because they show a more internationally valid brown hand
		//	instead of the Apple cursors' lily-white hands.
		//	Also, I prefer their 32×32 size over the default 16×16 size,
		//	but that's not a big deal.
		itsOpenHandCursor		= [[self loadCustomCursor:@"BrownHandOpen.tiff"		] retain];
		itsClosedHandCursor		= [[self loadCustomCursor:@"BrownHandClosed.tiff"	] retain];
		itsPointingHandCursor	= [[self loadCustomCursor:@"BrownHandPointing.tiff"	] retain];
	}

	return self;
}

- (NSCursor *)loadCustomCursor:(NSString *)aCursorName
{
	return [[[NSCursor alloc]
			initWithImage:	[[[NSImage alloc] initWithContentsOfFile:
				[[[[NSBundle mainBundle] resourcePath]
					stringByAppendingString:@"/Cursors/"]
					stringByAppendingString:aCursorName]]
				autorelease]
			hotSpot:		(NSPoint){13.0,  2.0}] autorelease];
}

- (void)dealloc
{
	[itsTimer invalidate];
	[itsTimer release];
	
	[itsLastImageDirectory release];
	
	[itsDefaultWindowTitle release];
	
	[itsOpenHandCursor		release];
	[itsClosedHandCursor	release];
	[itsPointingHandCursor	release];

	[super dealloc];
}

- (void)awakeFromNib
{
	//	Zoom to the largest 5:4 window that fits on the screen.
	[[self window] zoom:self];
//	[[self window] setContentSize:(NSSize){800, 640}];	//	for screenshots for web site
	
	//	Record the default window title so we can restore it
	//	after displaying the frame rate.
	[itsDefaultWindowTitle setString:[[self window] title]];
}

- (NSRect)windowWillUseStandardFrame:(NSWindow *)window defaultFrame:(NSRect)newFrame
{
	NSRect	theContentRect,
			theFrameRect;
	CGFloat	theDesiredHeight;

	//	NSWindow's zoom method calls this delegate method
	//	to obtain the desired zoomed placement for the window.
	
	//	Technical Note:  We could have used 
	//	[[self window] setContentAspectRatio:(NSSize){5.0, 4.0}] to achieve
	//	a 5:4 content area, but once it's set, there's no way to disable it!
	//	That is, you could choose a new aspect ratio,
	//	but you couldn't disable the constraint entirely.
	//	See  www.mail-archive.com/cocoa-dev@lists.apple.com/msg26039.html
	//	for confirmation that someone else has encountered this same problem.
	//	In any case, this implementation of windowWillUseStandardFrame:defaultFrame:
	//	does what we need it to do, so all is well.
	
	theContentRect = [window contentRectForFrameRect:newFrame];

	//	The graphics code will need to divide up the viewport,
	//	and the viewport uses integer coordinates.
	//	We expect integer values here already, of course,
	//	so the floor() calls should be unnecessary.
	theContentRect.size.width  = floor(theContentRect.size.width );
	theContentRect.size.height = floor(theContentRect.size.height);
			
	switch (md.itsStereoMode)
	{
		case StereoNone:
		case StereoAnaglyphic:

			//	Align to convenient 5:4 integer dimensions.
			theContentRect.size.width  -= ((unsigned int) theContentRect.size.width ) % 5;
			theContentRect.size.height -= ((unsigned int) theContentRect.size.height) % 4;

			//	Restrict the content area to the largest enclosed rectangle
			//	that has proportions 5:4.
			if (theContentRect.size.width	> 5 * (theContentRect.size.height / 4))
				theContentRect.size.width	= 5 * (theContentRect.size.height / 4);
			if (theContentRect.size.height	> 4 * (theContentRect.size.width  / 5))
				theContentRect.size.height	= 4 * (theContentRect.size.width  / 5);
			
			break;
		
		case StereoGeoWall:
			
			//	We'd like a 10:4 aspect ratio, and we're willing to reduce
			//	the window's height to achieve it.  But we're not willing
			//	to reduce its width, because the GeoWall assumes the window
			//	fills the full width of the screen.
			theDesiredHeight = 4 * floor(theContentRect.size.width / 10);
			if (theContentRect.size.height	> theDesiredHeight)
			{
				theContentRect.origin.y		+= theContentRect.size.height - theDesiredHeight;
				theContentRect.size.height	 = theDesiredHeight;
			}
			
			break;
		
		default:	//	Will never occur, unless new stereo mode gets added.

			theContentRect = (NSRect){{0.0, 0.0}, {128.0, 128.0}};

			break;
	}
	
	theFrameRect = [window frameRectForContentRect:theContentRect];
	
	return theFrameRect;
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
	NSArray		*thePreferredLanguages;
	NSString	*thePreferredLanguage;
	bool		theShadersAreAvailable;

	//	Uncomment the following line in test versions.
//	ErrorMessage("...", "Pre-release copy for ...");

	//	Initialize the language code and dictionary to the user's default language.
	//
	//	Note #1.  preferredLocalizations determines language availability
	//	based on what localizations it finds in our program's application bundle.
	//	Thus, as a practical matter, localizing InfoPList.strings to a new language
	//	will bring that language to the system's attention.
	//	
	//	Note #2.  One would expect preferredLocalizations to return an NSArray containing
	//	all localizations supported by our mainBundle and acceptable to the user,
	//	listed in order of the user's preferences.  In practice, however,
	//	preferredLocalizations always returns a 1-element array, even when several
	//	acceptable languages are available.  Apple's documentation is evasive
	//	on this question, saying only that preferredLocalizations
	//
	//		Returns one or more localizations contained in the receiver’s bundle 
	//		that the receiver uses to locate resources based on the user’s preferences.
	//
	//	Fortunately preferredLocalizations reliably returns the user's most preferred localization,
	//	from among those supported.
	thePreferredLanguages	= [[NSBundle mainBundle] preferredLocalizations];
	thePreferredLanguage	= ([thePreferredLanguages count] > 0 ?			//	If a preferred language exists...
								[thePreferredLanguages objectAtIndex:0] :	//	...use it.
								@"en");										//	Otherwise fall back to English.
	[self changeLanguage:thePreferredLanguage];

	//	Does the OpenGL context support shaders?
	//	If not, tell the user that the fallback fixed-function code
	//	will be used and the quality of the image will be less.
	//	Display that message once, and then never again (for that user).
	[[self openGLContext] makeCurrentContext];
	theShadersAreAvailable = ShadersAreAvailable();
	[NSOpenGLContext clearCurrentContext];
	if ( ! theShadersAreAvailable
	  && ! GetUserPrefBool("fallback message given"))
	{
		ErrorMessage(GetLocalizedText("ErrorShadersMessage"), GetLocalizedText("ErrorShadersTitle"));
		SetUserPrefBool("fallback message given", true);
	}
}

- (BOOL)acceptsFirstResponder
{
	return YES;
}

- (void)resetCursorRects
{
	NSRect	theWholeContent,
			theTilingRect,
			theTilingRectL,
			theTilingRectR,
			theControlRect,
			theControlRectL,
			theControlRectR;
	CGFloat	theOriginX,
			theOriginY,
			theWidth,
			theHeight;

	theWholeContent	= [self bounds];
	theOriginX		= theWholeContent.origin.x;
	theOriginY		= theWholeContent.origin.y;
	theWidth		= theWholeContent.size.width;
	theHeight		= theWholeContent.size.height;

	if (md.itsStereoMode != StereoGeoWall)
	{
		theTilingRect	= (NSRect)
							{
								(NSPoint)
									{
										theOriginX,
										theOriginY
									},
								(NSSize)
									{
										0.8 * theWidth,
										theHeight
									}
							};
		theControlRect	= (NSRect)
							{
								(NSPoint)
									{
										theOriginX + 0.8 * theWidth,
										theOriginY
									},
								(NSSize)
									{
										0.2 * theWidth,
										theHeight
									}
							};
		
		[self addCursorRect:theTilingRect  cursor:itsOpenHandCursor    ];
		[self addCursorRect:theControlRect cursor:itsPointingHandCursor];
	}
	else	//	StereoGeoWall
	{
		theTilingRectL	= (NSRect)
							{
								(NSPoint)
									{
										theOriginX,
										theOriginY
									},
								(NSSize)
									{
										0.4 * theWidth,
										theHeight
									}
							};
		theControlRectL	= (NSRect)
							{
								(NSPoint)
									{
										theOriginX + 0.4 * theWidth,
										theOriginY
									},
								(NSSize)
									{
										0.1 * theWidth,
										theHeight
									}
							};
		theTilingRectR	= (NSRect)
							{
								(NSPoint)
									{
										theOriginX + 0.5 * theWidth,
										theOriginY
									},
								(NSSize)
									{
										0.4 * theWidth,
										theHeight
									}
							};
		theControlRectR	= (NSRect)
							{
								(NSPoint)
									{
										theOriginX + 0.9 * theWidth,
										theOriginY
									},
								(NSSize)
									{
										0.1 * theWidth,
										theHeight
									}
							};
		
		[self addCursorRect:theTilingRectL  cursor:itsOpenHandCursor    ];
		[self addCursorRect:theControlRectL cursor:itsPointingHandCursor];
		[self addCursorRect:theTilingRectR  cursor:itsOpenHandCursor    ];
		[self addCursorRect:theControlRectR cursor:itsPointingHandCursor];
	}
}

- (void)setRedrawWithSync:(bool)aSyncFlag
{
	GLint	theSwapValue;

	//	Record the new sync value.
	md.itsSyncFlag = aSyncFlag;

	//	Set the swap interval.
	theSwapValue = (aSyncFlag ? 1 : 0);
	[[self openGLContext] setValues: &theSwapValue forParameter: NSOpenGLCPSwapInterval];

	//	Set a timer for continuous redraw.
	//	The scene may get redrawn less often than the timer interval
	//	(especially if we're synchronizing to the monitor)
	//	but it won't get drawn more often.

	//		Out with the old...
	if (itsTimer != nil)
	{
		[itsTimer invalidate];
		[itsTimer release];
		itsTimer = nil;
	}

	//		...in with the new.
	itsTimer = [NSTimer
			scheduledTimerWithTimeInterval: (md.itsSyncFlag ? 0.01 : 0.001)
			target: self
			selector: @selector( pleaseRedraw )
			userInfo: nil
			repeats: YES];
	[itsTimer retain];
}

- (void)pleaseRedraw
{
	long long	theCurrentTime;
	double		theElapsedTime;

	//	Compute the time since the last redraw.
	Microseconds( (UnsignedWide *) &theCurrentTime );
	theElapsedTime = 0.000001 * (theCurrentTime - itsLastRenderTime);
	itsLastRenderTime = theCurrentTime;

	//	If the window is miniaturized, proceed no further.
	if ([[self window] isMiniaturized])
		return;

	//	Need to redraw the image?
	if (SimulationWantsUpdates(&md))
	{
		//	Update the simulation.
		SimulationUpdate(&md, theElapsedTime, 0.000001 * theCurrentTime);

		//	Request a redraw.
		[self setNeedsDisplay:YES];
	}
	
	//	If desired, update the frame rate display about once per second.
	if (gShowFrameRate
	 && theCurrentTime >= itsLastFrameRateTime + FRAME_RATE_UPDATE_PERIOD)
	{
		[[self window] setTitle:
			(SimulationWantsUpdates(&md) ?
				(theElapsedTime > 0.0001 ?
					[NSString localizedStringWithFormat:
						GetLocalizedTextAsNSString("fps"),
						(unsigned int)(1.0 / theElapsedTime)]
				:
					GetLocalizedTextAsNSString("fps2")
				)
			:
				[NSString localizedStringWithFormat:
					GetLocalizedTextAsNSString("fps"), 0]
			)
		];
		itsLastFrameRateTime = theCurrentTime;
	}
}

- (void)keyDown:(NSEvent *)anEvent
{
	//	Handle normal keystrokes in insertText (see below).
	//	Handle stuff like arrow keys here.

	switch ([anEvent keyCode])
	{
		case SPACEBAR:
			break;
		
		case ESCAPE_KEY:
			break;
		
		case LEFT_ARROW_KEY:
			break;

		case RIGHT_ARROW_KEY:
			break;

		case DOWN_ARROW_KEY:
			break;

		case UP_ARROW_KEY:
			break;

		default:

			//	Let interpretKeyEvents send plain text to insertText.
			[self interpretKeyEvents:[NSArray arrayWithObject:anEvent]];
			
			break;
	}
}

- (void)insertText:(id)aString
{
	NSString	*theNewText;
	unichar		theCharacter;

	//	The documentation promises only that aString will be either
	//	an NSString or an NSAttributedString. In practice it's an NSString
	//	for both the US English server and Hiragana server.
	if ([[aString class] isSubclassOfClass:[NSAttributedString class]])
	{
		theNewText = [aString string];
	}
	else
	if ([[aString class] isSubclassOfClass:[NSString class]])
	{
		theNewText = aString;
	}
	else
	{
		theNewText = @"?";
	}
	
	if ([theNewText length] > 0)
	{
		theCharacter = [theNewText characterAtIndex:0];

		if (theCharacter == 'a')	//	"secret" keyboard command to toggle anisotropic texture filtering
		{
			SetAnisotropicEnabledFlag(&md, ! md.itsAnisotropicEnabledFlag );
		}
		else
		if (theCharacter == 'f')	//	"secret" keyboard command to show frame rate
		{
			gShowFrameRate = ! gShowFrameRate;
			if ( ! gShowFrameRate )
				[[self window] setTitle:GetLocalizedTextAsNSString("WindowTitle")];
		}
		else
		if (theCharacter == 'm')	//	"secret" keyboard command to sync to the monitor
		{
			[self setRedrawWithSync:( ! md.itsSyncFlag )];
		}
	}
}

- (void)mouseDown:(NSEvent *)anEvent
{
	NSPoint	theMouseLocation;
	
	//	If the current cursor is an open hand cursor, push a closed hand cursor
	//	onto the cursor stack for the duration of the drag.  mouseUp will pop it off.
	//	If the current cursor is a pointing hand cursor, we want to leave it that way,
	//	but the simplest thing is to push it onto the cursor stack anyhow,
	//	to simplify the popping logic in mouseUp.
	[([NSCursor currentCursor] == itsOpenHandCursor ?
		itsClosedHandCursor : [NSCursor currentCursor]) push];

	//	Handle the mouseDown event.
	theMouseLocation = [self mouseLocation:anEvent];
	MouseDown(&md, theMouseLocation.x, theMouseLocation.y, [anEvent timestamp]);
}

- (void)mouseDragged:(NSEvent *)anEvent
{
	NSPoint	theMouseLocation;

	theMouseLocation = [self mouseLocation:anEvent];
	MouseMove(&md, theMouseLocation.x, theMouseLocation.y, [anEvent timestamp]);
}

- (void)mouseUp:(NSEvent *)anEvent
{
	NSPoint	theMouseLocation;

	theMouseLocation = [self mouseLocation:anEvent];
	MouseUp(&md, theMouseLocation.x, theMouseLocation.y, [anEvent timestamp]);
	
	[NSCursor pop];
}

- (NSPoint)mouseLocation:(NSEvent *)anEvent
{
	NSPoint	theMouseLocation;

	//	mouse location in window coordinates
	theMouseLocation = [anEvent locationInWindow];

	//	mouse location in view coordinates
	theMouseLocation = [self convertPoint:theMouseLocation fromView:nil];

	//	mouse location in [0,1] coordinates
	theMouseLocation.x /= [self bounds].size.width;
	theMouseLocation.y /= [self bounds].size.height;
	
	//	Adjust for side-by-side images.
	if (md.itsStereoMode == StereoGeoWall)
	{
		if (theMouseLocation.x <= 0.5)	//	map [0.0, 0.5] → [0.0, 1.0]
			theMouseLocation.x *= 2.0;
		else							//	map [0.5, 1.0] → [0.0, 1.0]
			theMouseLocation.x = 2.0*theMouseLocation.x - 1.0;
	}
	
	return theMouseLocation;
}

- (void)scrollWheel:(NSEvent *)anEvent
{
}

- (void)drawRect:(NSRect)bounds
{
	ErrorText	theError;

	if ((theError = SetUpGraphicsAsNeeded(&md)) != NULL)
	{
		ErrorMessage(theError, "OpenGL Setup Error");
		exit(1);
	}

	if ((theError = Render(&md)) != NULL)
	{
		ErrorMessage(theError, "OpenGL Rendering Error");
		exit(2);
	}

	[[self openGLContext] flushBuffer];
}

- (BOOL)validateMenuItem:(NSMenuItem *)aMenuItem
{
	SEL	theAction;
	
	theAction = [aMenuItem action];

	if (theAction == @selector(commandShowPlainImages:))
	{
		[aMenuItem setState:(md.itsPositiveImagesFlag	? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandShowReflectedImages:))
	{
		[aMenuItem setState:(md.itsNegativeImagesFlag	? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandCutAlongMirrorLines:))
	{
		[aMenuItem setState:(md.itsMirrorLinesFlag		? NSOnState : NSOffState)];
		return YES;
	}

	if (theAction == @selector(commandMakeTilesBigger:))
	{
		return md.itsTiling.itsGeometry == GeometryEuclidean
			&& md.itsTiling.itsEuclideanZoom < MAX_EUCLIDEAN_ZOOM;
	}
	if (theAction == @selector(commandMakeTilesSmaller:))
	{
		return md.itsTiling.itsGeometry == GeometryEuclidean
			&& md.itsTiling.itsEuclideanZoom > MIN_EUCLIDEAN_ZOOM;
	}

	if (theAction == @selector(commandSnapToArchimedeanSolids:))
	{
		[aMenuItem setState:(md.itsSnapToArchimedeanFlag	? NSOnState : NSOffState)];
		return YES;
	}

	if (theAction == @selector(commandViewMultisampling:))
	{
		[aMenuItem setState:(	md.itsMultisamplingAvailableFlag
							 && md.itsMultisamplingEnabledFlag	? NSOnState : NSOffState)];
		return md.itsMultisamplingAvailableFlag;
	}

	if (theAction == @selector(commandMesh2:))
	{
		if (md.itsTilingStyle == StyleCurved)
		{
			[aMenuItem setState:(md.itsMeshSize ==  2 ? NSOnState : NSOffState)];
			return YES;
		}
		else
		{
			[aMenuItem setState:NSOffState];
			return NO;
		}
	}
	if (theAction == @selector(commandMesh4:))
	{
		if (md.itsTilingStyle == StyleCurved)
		{
			[aMenuItem setState:(md.itsMeshSize ==  4 ? NSOnState : NSOffState)];
			return YES;
		}
		else
		{
			[aMenuItem setState:NSOffState];
			return NO;
		}
	}
	if (theAction == @selector(commandMesh8:))
	{
		if (md.itsTilingStyle == StyleCurved)
		{
			[aMenuItem setState:(md.itsMeshSize ==  8 ? NSOnState : NSOffState)];
			return YES;
		}
		else
		{
			[aMenuItem setState:NSOffState];
			return NO;
		}
	}
	if (theAction == @selector(commandMesh16:))
	{
		if (md.itsTilingStyle == StyleCurved)
		{
			[aMenuItem setState:(md.itsMeshSize == 16 ? NSOnState : NSOffState)];
			return YES;
		}
		else
		{
			[aMenuItem setState:NSOffState];
			return NO;
		}
	}

	if (theAction == @selector(commandTilingRadius40:))
	{
		if (md.itsTiling.itsGeometry == GeometryHyperbolic)
		{
			//	4.0 has a finite binary expansion, so it's OK to compare
			//	floating point numbers with zero error tolerance.
			[aMenuItem setState:(md.itsHyperbolicTilingRadius == 4.0 ? NSOnState : NSOffState)];
			return YES;
		}
		else
		{
			[aMenuItem setState:NSOffState];
			return NO;
		}
	}
	if (theAction == @selector(commandTilingRadius45:))
	{
		if (md.itsTiling.itsGeometry == GeometryHyperbolic)
		{
			//	4.5 has a finite binary expansion, so it's OK to compare
			//	floating point numbers with zero error tolerance.
			[aMenuItem setState:(md.itsHyperbolicTilingRadius == 4.5 ? NSOnState : NSOffState)];
			return YES;
		}
		else
		{
			[aMenuItem setState:NSOffState];
			return NO;
		}
	}
	if (theAction == @selector(commandTilingRadius50:))
	{
		if (md.itsTiling.itsGeometry == GeometryHyperbolic)
		{
			//	5.0 has a finite binary expansion, so it's OK to compare
			//	floating point numbers with zero error tolerance.
			[aMenuItem setState:(md.itsHyperbolicTilingRadius == 5.0 ? NSOnState : NSOffState)];
			return YES;
		}
		else
		{
			[aMenuItem setState:NSOffState];
			return NO;
		}
	}
	if (theAction == @selector(commandTilingRadius55:))
	{
		if (md.itsTiling.itsGeometry == GeometryHyperbolic)
		{
			//	5.5 has a finite binary expansion, so it's OK to compare
			//	floating point numbers with zero error tolerance.
			[aMenuItem setState:(md.itsHyperbolicTilingRadius == 5.5 ? NSOnState : NSOffState)];
			return YES;
		}
		else
		{
			[aMenuItem setState:NSOffState];
			return NO;
		}
	}
	if (theAction == @selector(commandTilingRadius60:))
	{
		if (md.itsTiling.itsGeometry == GeometryHyperbolic)
		{
			//	6.0 has a finite binary expansion, so it's OK to compare
			//	floating point numbers with zero error tolerance.
			[aMenuItem setState:(md.itsHyperbolicTilingRadius == 6.0 ? NSOnState : NSOffState)];
			return YES;
		}
		else
		{
			[aMenuItem setState:NSOffState];
			return NO;
		}
	}

	if (theAction == @selector(commandGraphicsFixedFunction:))
	{
		[aMenuItem setState:(! (md.itsShadersAvailableFlag
							 && md.itsShadersEnabledFlag)			? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandGraphicsShaders:))
	{
		[aMenuItem setState:(	md.itsShadersAvailableFlag
							 && md.itsShadersEnabledFlag			? NSOnState : NSOffState)];
		return md.itsShadersAvailableFlag;
	}

	if (theAction == @selector(commandViewStereoNone:))
	{
		[aMenuItem setState:(md.itsStereoMode == StereoNone			? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandViewStereoAnaglyphic:))
	{
		[aMenuItem setState:(md.itsStereoMode == StereoAnaglyphic	? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandViewStereoGeoWall:))
	{
		[aMenuItem setState:(md.itsStereoMode == StereoGeoWall		? NSOnState : NSOffState)];
		return YES;
	}

	if (theAction == @selector(commandLanguageCy:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "cy") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageEn:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "en") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageFr:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "fr") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageIt:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "it") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageJa:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "ja") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguagePt:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "pt") ? NSOnState : NSOffState)];
		return YES;
	}

	return YES;
}

- (IBAction)commandExportCopy:(id)sender
{
	NSData			*theTiff				= nil;
	NSPasteboard	*theGeneralPasteboard	= nil;

	theTiff = [self contentsAsTIFF];

	theGeneralPasteboard = [NSPasteboard generalPasteboard];
	
	[theGeneralPasteboard declareTypes: [NSArray arrayWithObject:NSTIFFPboardType] owner: nil];
	[theGeneralPasteboard setData: theTiff forType:NSTIFFPboardType];
}

- (IBAction)commandExportSave:(id)sender
{
	NSSavePanel	*theSavePanel	= nil;
	NSData		*theTiff		= nil;

	//	Create a save-a-file panel.
	theSavePanel = [NSSavePanel savePanel];
	[theSavePanel setAllowedFileTypes:[NSArray arrayWithObject:@"tiff"]];
	[theSavePanel setTitle:@"Save image as TIFF file"];

	//	Ask the user to choose a file name.
	if ([theSavePanel runModalForDirectory:itsLastImageDirectory file:
//	If the window represents a file, suggest <filename>.tiff as a default.
//			[[[[[self window] representedFilename] lastPathComponent] stringByDeletingPathExtension] stringByAppendingPathExtension:@"tiff"]
//	Otherwise suggest the window title, or even just pass nil to suggest nothing.
//			[itsDefaultWindowTitle stringByAppendingPathExtension:@"tiff"]
			nil
		] == NSFileHandlingPanelCancelButton)
		return;	//	user cancelled
	
	//	Remember the directory for next time.
	[itsLastImageDirectory setString:[theSavePanel directory]];

	//	Read the OpenGL image.
	theTiff = [self contentsAsTIFF];
	if (theTiff == nil)
		return;
	
	//	Save the data under the requested filename.
	if ([theTiff writeToFile:[theSavePanel filename] atomically:YES] == NO)
		ErrorMessage("Couldn't save image under requested filename.", "Save failed");
}

- (IBAction)commandChooseSymmetries:(id)sender
{
	//	Transfer the current (p,q,r) values into the NSPopUpButtons.
	[itsSymmetryP selectItemWithTag:md.itsTiling.itsPQR[0]];
	[itsSymmetryQ selectItemWithTag:md.itsTiling.itsPQR[1]];
	[itsSymmetryR selectItemWithTag:md.itsTiling.itsPQR[2]];

	//	Display the Symmetry Panel as a sheet.
	[NSApp beginSheet:itsSymmetryPanel modalForWindow:[self window] modalDelegate:nil didEndSelector:nil contextInfo:nil];
}

- (IBAction)updateSymmetryGroup:(id)sender
{
	ChangeTiling(&md,
		[[itsSymmetryP titleOfSelectedItem] intValue],
		[[itsSymmetryQ titleOfSelectedItem] intValue],
		[[itsSymmetryR titleOfSelectedItem] intValue]);
}

- (IBAction)closeSymmetryPanel:(id)sender
{
	//	Stop the Symmetry Panel's modal session and remove it from the window list,
	//	but don't destroy it.
	[NSApp endSheet:itsSymmetryPanel];
	[itsSymmetryPanel orderOut:self];
}

- (IBAction)commandShowPlainImages:(id)sender
{
	SetPositiveImagesFlag(&md, ! md.itsPositiveImagesFlag );
}

- (IBAction)commandShowReflectedImages:(id)sender
{
	SetNegativeImagesFlag(&md, ! md.itsNegativeImagesFlag );
}

- (IBAction)commandCutAlongMirrorLines:(id)sender
{
	SetMirrorLinesFlag(&md, ! md.itsMirrorLinesFlag );
}

- (IBAction)commandMakeTilesBigger:(id)sender
{
	SetBiggerTiles(&md);
}

- (IBAction)commandMakeTilesSmaller:(id)sender
{
	SetSmallerTiles(&md);
}

- (IBAction)commandSnapToArchimedeanSolids:(id)sender
{
	md.itsSnapToArchimedeanFlag = ! md.itsSnapToArchimedeanFlag;
}


//	The following three View menu items are synonyms for Preference Panel options,
//	so as well as changing our main settings, be sure to update the Preference Panel.
//	Under normal circumstances this wouldn't be necessary, but if the user happens
//	to have the Preference Panel open, we want to keep it valid.

- (IBAction)commandViewMultisampling:(id)sender
{
	SetMultisamplingEnabledFlag(&md, ! md.itsMultisamplingEnabledFlag );
	[itsPrefMultisampling setState:(md.itsMultisamplingEnabledFlag ? NSOnState : NSOffState)];
	[self replaceOpenGLContext];
}


- (IBAction)commandMesh2:(id)sender
{
	SetMeshSize(&md,  2);
}

- (IBAction)commandMesh4:(id)sender
{
	SetMeshSize(&md,  4);
}

- (IBAction)commandMesh8:(id)sender
{
	SetMeshSize(&md,  8);
}

- (IBAction)commandMesh16:(id)sender
{
	SetMeshSize(&md, 16);
}


- (IBAction)commandTilingRadius40:(id)sender
{
	SetHyperbolicTilingRadius(&md, 4.0);
}

- (IBAction)commandTilingRadius45:(id)sender
{
	SetHyperbolicTilingRadius(&md, 4.5);
}

- (IBAction)commandTilingRadius50:(id)sender
{
	SetHyperbolicTilingRadius(&md, 5.0);
}

- (IBAction)commandTilingRadius55:(id)sender
{
	SetHyperbolicTilingRadius(&md, 5.5);
}

- (IBAction)commandTilingRadius60:(id)sender
{
	SetHyperbolicTilingRadius(&md, 6.0);
}


- (IBAction)commandGraphicsFixedFunction:(id)sender
{
	SetShadersEnabledFlag(&md, false);
	[itsPrefShaders setState:NSOffState];
}

- (IBAction)commandGraphicsShaders:(id)sender
{
	SetShadersEnabledFlag(&md, true);
	[itsPrefShaders setState:NSOnState];
}


- (IBAction)commandViewStereoNone:(id)sender
{
	[self setStereoMode:StereoNone];
}

- (IBAction)commandViewStereoAnaglyphic:(id)sender
{
	[self setStereoMode:StereoAnaglyphic];
}

- (IBAction)commandViewStereoGeoWall:(id)sender
{
	[self setStereoMode:StereoGeoWall];
}

- (void)setStereoMode:(StereoMode)aStereoMode;
{
	//	Set the new mode.
	SetStereo3DMode(&md, aStereoMode);

	//	Enable or disable the stereo-specific fields
	//	in the Perspective section of the Preferences panel.
	[self setStereoLabels];

	//	Zoom to a large square window in StereoNone or StereoAnaglyphic modes,
	//	or to a full-width window in StereoGeoWall mode.
	if ( ! [[self window] isZoomed] )
		[[self window] zoom:self];
}


- (IBAction)commandLanguageCy:(id)sender
{
	[self changeLanguage:@"cy"];
}

- (IBAction)commandLanguageEn:(id)sender
{
	[self changeLanguage:@"en"];
}

- (IBAction)commandLanguageFr:(id)sender
{
	[self changeLanguage:@"fr"];
}

- (IBAction)commandLanguageIt:(id)sender
{
	[self changeLanguage:@"it"];
}

- (IBAction)commandLanguageJa:(id)sender
{
	[self changeLanguage:@"ja"];
}

- (IBAction)commandLanguagePt:(id)sender
{
	[self changeLanguage:@"pt"];
}


- (IBAction)commandHelpStudyQuestions:(id)sender
{
	//	The Mac OS X Help Book system is terrible:
	//	it's very difficult to get it accept different pages,
	//	won't accept multiple named pages in the same directory,
	//	tends to fail or crash at the least excuse,
	//	and caches so agressively that you can't get
	//	old versions of a Help Book to go away.
	//
	//	By contrast, sending a page to the default web browser
	//	is simple and effective.  A real breath of fresh air!
	[[NSWorkspace sharedWorkspace] openFile:[[[NSBundle mainBundle] resourcePath]
		stringByAppendingString:[NSString stringWithFormat:@"/Help/KaleidoTileStudyQuestions-%c%c.html", gLanguageCode[0], gLanguageCode[1]]]];
}


- (IBAction)setUpPreferencesPanel:(id)sender
{
	//	Transfer our private data to the Preference Panel's fields.
	
	//	Perspective Tab
	
	[itsPrefWindowWidth		setDoubleValue: 2.0 * md.itsWindowHalfWidth];
	[itsPrefWindowDistance	setDoubleValue:       md.itsWindowDistance ];
	[itsPrefEyeSeparation	setDoubleValue: 2.0 * md.itsEyeOffset      ];
	[self setStereoLabels];

	//	Advanced Tab
	
	[itsPrefFrameRate		setState:(gShowFrameRate					? NSOnState : NSOffState)];

	[itsPrefSyncToMonitor	setState:(md.itsSyncFlag					? NSOnState : NSOffState)];

	[itsPrefMultisampling	setEnabled: md.itsMultisamplingAvailableFlag ];
	[itsPrefMultisampling	setState:(	md.itsMultisamplingAvailableFlag
									 && md.itsMultisamplingEnabledFlag	? NSOnState : NSOffState)];

	[itsPrefShaders			setEnabled: md.itsShadersAvailableFlag ];
	[itsPrefShaders			setState:(	md.itsShadersAvailableFlag
									 && md.itsShadersEnabledFlag		? NSOnState : NSOffState)];

	//	Display the Preference Panel as a sheet.
	[NSApp beginSheet:itsPreferencesPanel modalForWindow:[self window] modalDelegate:nil didEndSelector:nil contextInfo:nil];
}


- (void) setStereoLabels
{
	//	Cocoa provides no way to enable/disable static text (!).
	//	See http://developer.apple.com/qa/qa2006/qa1461.html for discussion.
	//	So instead we must grey the static text ourselves.

	NSColor	*theStaticTextColor;
	BOOL	theEnabledFlag;
	
	if (md.itsStereoMode != StereoNone)
	{
		theStaticTextColor	= [NSColor controlTextColor];
		theEnabledFlag		= YES;
	}
	else
	{
		theStaticTextColor	= [NSColor disabledControlTextColor];
		theEnabledFlag		= NO;
	}
	
	[itsPrefEyeSeparation		setEnabled:theEnabledFlag];
	[itsPrefEyeSeparationLabel	setTextColor:theStaticTextColor];
	[itsPrefEyeSeparationUnits	setTextColor:theStaticTextColor];
}

- (IBAction)readOnePreference:(id)sender
{
	//	Perspective Tab

	if (sender == itsPrefWindowWidth)
	{
		md.itsWindowHalfWidth	= 0.5 * [itsPrefWindowWidth		doubleValue];
		[self setHeightFromWidth];
	}

	if (sender == itsPrefWindowDistance)
		md.itsWindowDistance	=       [itsPrefWindowDistance	doubleValue];

	if (sender == itsPrefEyeSeparation)
		md.itsEyeOffset			= 0.5 * [itsPrefEyeSeparation	doubleValue];


	//	Advanced Tab

	if (sender == itsPrefFrameRate)
	{
		gShowFrameRate = ([itsPrefFrameRate state] == NSOnState);
		if ( ! gShowFrameRate )
//			[[self window] setTitleWithRepresentedFilename:[[self window] representedFilename]];
			[[self window] setTitle:itsDefaultWindowTitle];
	}

	if (sender == itsPrefSyncToMonitor)
	{
		[self setRedrawWithSync:([itsPrefSyncToMonitor state] == NSOnState)];
	}

	if (sender == itsPrefMultisampling)
	{
		SetMultisamplingEnabledFlag(&md, [itsPrefMultisampling state] == NSOnState);
		[self replaceOpenGLContext];
	}

	if (sender == itsPrefShaders)
		SetShadersEnabledFlag(&md, [itsPrefShaders state] == NSOnState);
	

	//	OK Button
	//		shutDownPreferencesPanel wants us to test for modified text fields.
	if (sender == self)
	{
		md.itsWindowHalfWidth	= 0.5 * [itsPrefWindowWidth		doubleValue];
		md.itsWindowDistance	=       [itsPrefWindowDistance	doubleValue];
		md.itsEyeOffset			= 0.5 * [itsPrefEyeSeparation	doubleValue];

		[self setHeightFromWidth];
	}


	//	Request a redraw.
	md.itsRedrawRequestFlag = true;
}

- (void)setHeightFromWidth
{
	NSSize	theViewSize;
	double	theAspectRatio;

	theViewSize = [self bounds].size;

	//	To allow space for the control panel, the tiling itself
	//	fills only 4/5 of the view's width.  The variable itsWindowHalfWidth
	//	refers to that 4/5, not to the whole window width.
	if (theViewSize.width > 0.0 && theViewSize.height > 0.0)
		theAspectRatio = (5.0 * (double)theViewSize.height) / (4.0 * (double)theViewSize.width);
	else	//	should never occur
		theAspectRatio = 1.0;

	md.itsWindowHalfHeight = theAspectRatio * md.itsWindowHalfWidth;
}

- (void)replaceOpenGLContext
{
	NSOpenGLPixelFormat	*thePixelFormat;
	NSOpenGLContext		*theNewContext;
	
	//	Shut down the old OpenGL context.
	//	The old OpenGL context is about to be vaporized, 
	//	but calling ShutDownGraphicsAsNeeded() gives
	//	the program a chance to clear its references 
	//	to shader programs, textures and vertex buffer objects.
	[[self openGLContext] makeCurrentContext];
	ShutDownGraphicsAsNeeded(&md);
	[NSOpenGLContext clearCurrentContext];
	
	//	Get a new OpenGL context with the desired pixel format.
	thePixelFormat	= [KaleidoTileMacOSView pixelFormatSupportingMultisampling:md.itsMultisamplingEnabledFlag];
	theNewContext	= [[[NSOpenGLContext alloc] initWithFormat:thePixelFormat shareContext:nil] autorelease];
	[self setOpenGLContext:theNewContext];
	[theNewContext setView:self];
	
	//	Tell the new context about itsSyncFlag.
	[self setRedrawWithSync:md.itsSyncFlag];
	
	//	Reset everything.
	md.itsPreparedExtensions = false;
}

- (IBAction)shutDownPreferencesPanel:(id)sender
{
	//	Stop the Preference Panel's modal session and remove it from the window list,
	//	but don't destroy it.
	[NSApp endSheet:itsPreferencesPanel];
	[itsPreferencesPanel orderOut:self];

	//	The checkboxes and radio buttons all produce immediate results,
	//	and so do the text fields if the user hits tab or return.
	//	But if the user changes a text field and then hits OK,
	//	we have to explicitly ask that the text field be read.
	[self readOnePreference:self];
}


- (NSImage *)contentsAsImage
{
	NSBitmapImageRep	*theBitmap	= nil;
	NSImage				*theImage	= nil;
	
	//	Get a bitmap...
	theBitmap = [self contentsAsBitmapImageRep];
	
	//	...and wrap it as an NSImage.
	//
	//	Note:  The documentation for addRepresentation makes the vague comment that
	//	"After invoking this method, you may need to explicitly set features
	//	of the new image representation, such as the size, number of colors, and so on."
	//	but in practice it seems to work OK.
	theImage = [[[NSImage alloc] initWithSize:(NSSize){0.0, 0.0}] autorelease];
	[theImage addRepresentation:theBitmap];

	return theImage;
}


- (NSData *)contentsAsTIFF
{
	NSBitmapImageRep	*theBitmap	= nil;
	NSData				*theTIFF	= nil;
	
	//	Get a bitmap...
	theBitmap = [self contentsAsBitmapImageRep];

	//	...and convert it to TIFF format.
	theTIFF = [theBitmap TIFFRepresentationUsingCompression:NSTIFFCompressionLZW factor:0.0f];
	if (theTIFF == nil)
		ErrorMessage("Couldn't create TIFF representation.", "Image Copying Error");

	return theTIFF;
}


- (NSBitmapImageRep *)contentsAsBitmapImageRep
{
	ErrorText			theError	= NULL;
	NSRect				theBounds	= {{0.0f, 0.0f}, {0.0f, 0.0f}};
	NSBitmapImageRep	*theBitmap	= nil;
	unsigned char		*thePixels	= NULL;
	
	//	Note the view's bounds.
	theBounds = [self bounds];
	if (NSMinX(theBounds) != 0 || NSMinY(theBounds) != 0)
	{
		theError = "View does not have origin (0,0).";
		goto CleanUpContentsAsBitmapImageRep;
	}

	//	Create an empty bitmap of the correct dimensions.
	//	By default the NSBitmapImageRep assumes premultiplied alpha.
	theBitmap = [[[NSBitmapImageRep alloc]
		initWithBitmapDataPlanes:	NULL
		pixelsWide:					NSWidth(theBounds)
		pixelsHigh:					NSHeight(theBounds)
		bitsPerSample:				8
		samplesPerPixel:			4
		hasAlpha:					YES
		isPlanar:					NO
		colorSpaceName:				NSDeviceRGBColorSpace
		bytesPerRow:				NSWidth(theBounds) * 4
		bitsPerPixel:				32
		] autorelease];
	if (theBitmap == nil)
	{
		theError = "Couldn't allocate theBitmap.";
		goto CleanUpContentsAsBitmapImageRep;
	}
	
	//	Read the pixels into theBitmap.
	thePixels = [theBitmap bitmapData];
	[self lockFocus];
		theError = ReadRGBAPixels(NSWidth(theBounds), NSHeight(theBounds), thePixels);
	[self unlockFocus];
	if (theError != NULL)
		goto CleanUpContentsAsBitmapImageRep;
	
	//	OpenGL reads the image bottom-to-top, but Cocoa seems to be expecting
	//	it top-to-bottom.  So we must reverse the rows.
	[self invertRawImage: (unsigned int *)thePixels pixelsWide: NSWidth(theBounds) pixelsHigh: NSHeight(theBounds)];
	
CleanUpContentsAsBitmapImageRep:

	if (theError != NULL)
		ErrorMessage(theError, "Error Copying OpenGL Image from Window");

	return theBitmap;
}


- (void)invertRawImage: (unsigned int *)aPixelBuffer pixelsWide:(unsigned int)aWidth pixelsHigh:(unsigned int)aHeight
{
	unsigned int	theLowerRow,
					theUpperRow,
					theCount;
	
	//	Rather than thinking of a pixel as four 1-byte components {R,G,B,A},
	//	think of it as a single 4-byte integer.  The byte order,
	//	0xRRGGBBAA or 0xAABBGGRR, makes no difference here.
	unsigned int	*theLowerPixel,
					*theUpperPixel,
					theTempPixel;
	
	theLowerRow	= 0;
	theUpperRow	= aHeight - 1;

	while (theUpperRow > theLowerRow)
	{
		theLowerPixel = aPixelBuffer + theLowerRow*aWidth;
		theUpperPixel = aPixelBuffer + theUpperRow*aWidth;

		for (theCount = aWidth; theCount-- > 0; )
		{
			theTempPixel	= *theLowerPixel;
			*theLowerPixel	= *theUpperPixel;
			*theUpperPixel	= theTempPixel;
			theLowerPixel++;
			theUpperPixel++;
		}

		theLowerRow++;
		theUpperRow--;
	}
}


- (void)changeLanguage:(NSString *)aLanguageCode
{
	//	Record the new language code and initialize a dictionary.
	RefreshLocalization([aLanguageCode UTF8String]);

	//	Reconstruct the menu bar in the new language.
	[NSApp setMainMenu:[self buildLocalizedMenuBar]];
	
	//	Update the window title.
	[[self window] setTitle:GetLocalizedTextAsNSString("WindowTitle")];
	
	//	Update the text in the Control Panel and
	//	the displayed Archimedean polyhedron name (if any).
	md.itsPreparedTextures = false;

	//	Update the text in the Preferences Panel,
	//	whether or not it's currently visible.
	[self setPreferencePanelLabels];

	//	Update the text in the Symmetry Panel,
	//	whether or not it's currently visible.
	[self setSymmetryPanelLabels];

	//	Request a redraw.
	md.itsRedrawRequestFlag = true;
}

- (NSMenu *)buildLocalizedMenuBar
{
	NSMenu		*theMenuBar,
				*theMenu,
				*theSubmenu;
	NSMenuItem	*theItemWithKeyModifier;

	//	Construct all menus using the current language.

	//	Design Note:  A more traditional approach would be to maintain 
	//	a separate .nib file for each language, but it's a pain
	//	to maintain numerous parallel .nib files as the software evolves --
	//	one must either maintain them by hand or take on the added complexity
	//	of dealing with AppleGlot to automate the process.
	//	Personally I prefer to keep a program's functionality separated 
	//	from its language-specific strings.  Thus the menus (including all
	//	selectors and key equivalents) get defined here here in this
	//	language-independent function, while all language-specific strings 
	//	(including the text of each menu item) get stored in the language files.
	//	This not only makes program maintenance easier (because a change 
	//	to the menu bar happens in one place only) but also simplifies localization
	//	(because the translator works with a simple text-only file containing 
	//	the English phrases alongside their local language equivalents).
	
	//	Main menu bar
	theMenuBar = [[[NSMenu alloc] initWithTitle:@"main menu"] autorelease];
	
	//	Application menu
	//	(On Mac OS 10.6, title will be overridden by localized CFBundleName.
	//	But on 10.4 or 10.5, the title given here may persist.)
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"KaleidoTile"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("About KaleidoTile")
			action:@selector(orderFrontStandardAboutPanel:) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Preferences…")
			action:@selector(setUpPreferencesPanel:) keyEquivalent:@","];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Hide KaleidoTile")
			action:@selector(hide:) keyEquivalent:@"h"];
		theItemWithKeyModifier = [[NSMenuItem alloc]
					initWithTitle:GetLocalizedTextAsNSString("Hide Others")
					action:@selector(hideOtherApplications:) keyEquivalent:@"h"];
			[theItemWithKeyModifier setKeyEquivalentModifierMask:(NSCommandKeyMask | NSAlternateKeyMask)];
			[theMenu addItem:theItemWithKeyModifier];
			[theItemWithKeyModifier release];
			theItemWithKeyModifier = nil;
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Show All")
			action:@selector(unhideAllApplications:) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Quit KaleidoTile")
			action:@selector(terminate:) keyEquivalent:@"q"];
	theMenu = nil;
	
	//	Export menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"Export"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Copy Image")
			action:@selector(commandExportCopy:) keyEquivalent:@"c"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Save Image…")
			action:@selector(commandExportSave:) keyEquivalent:@"s"];
	theMenu = nil;
	
	//	View menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"View"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Choose Symmetries…")
			action:@selector(commandChooseSymmetries:) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Show Plain Images")
			action:@selector(commandShowPlainImages:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Show Reflected Images")
			action:@selector(commandShowReflectedImages:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Cut Along Mirror Lines")
			action:@selector(commandCutAlongMirrorLines:) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Make Tiles Bigger")
			action:@selector(commandMakeTilesBigger:) keyEquivalent:@"+"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Make Tiles Smaller")
			action:@selector(commandMakeTilesSmaller:) keyEquivalent:@"-"];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Snap to Archimedean Solids")
			action:@selector(commandSnapToArchimedeanSolids:) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Smooth Edges")
			action:@selector(commandViewMultisampling:) keyEquivalent:@""];
		theSubmenu = [self addToMenu:theMenu submenuWithTitle:"Mesh Refinement"];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Coarse (×2)")
				action:@selector(commandMesh2:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Medium (×4)")
				action:@selector(commandMesh4:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Fine (×8)")
				action:@selector(commandMesh8:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Extra Fine (×16)")
				action:@selector(commandMesh16:) keyEquivalent:@""];
		theSubmenu = nil;
		theSubmenu = [self addToMenu:theMenu submenuWithTitle:"Tiling Radius"];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("4.0")
				action:@selector(commandTilingRadius40:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("4.5")
				action:@selector(commandTilingRadius45:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("5.0")
				action:@selector(commandTilingRadius50:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("5.5")
				action:@selector(commandTilingRadius55:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("6.0")
				action:@selector(commandTilingRadius60:) keyEquivalent:@""];
		theSubmenu = nil;
		theSubmenu = [self addToMenu:theMenu submenuWithTitle:"Graphics Pipeline"];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Fixed-Function")
				action:@selector(commandGraphicsFixedFunction:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Shaders")
				action:@selector(commandGraphicsShaders:) keyEquivalent:@""];
		theSubmenu = nil;
		theSubmenu = [self addToMenu:theMenu submenuWithTitle:"Stereoscopic 3D"];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("None")
				action:@selector(commandViewStereoNone:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Anaglyphic")
				action:@selector(commandViewStereoAnaglyphic:) keyEquivalent:@""];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("GeoWall")
				action:@selector(commandViewStereoGeoWall:) keyEquivalent:@""];
		theSubmenu = nil;
	theMenu = nil;

	//	Language menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"Language"];
		//	To add a new language, please see the instructions
		//	in the file "To add a new language.txt".
		[theMenu addItemWithTitle:@"Cymraeg"
			action:@selector(commandLanguageCy:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"English"
			action:@selector(commandLanguageEn:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"Français"
			action:@selector(commandLanguageFr:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"Italiano"
			action:@selector(commandLanguageIt:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"日本語"
			action:@selector(commandLanguageJa:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"Português"
			action:@selector(commandLanguagePt:) keyEquivalent:@""];
	theMenu = nil;
	
	//	Help menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"Help"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("KaleidoTile Study Questions")
			action:@selector(commandHelpStudyQuestions:) keyEquivalent:@"?"];
	theMenu = nil;
	
	return theMenuBar;
}

- (NSMenu *)addToMenu:(NSMenu *)aParentMenu submenuWithTitle:(char *)aSubmenuTitle
{
	NSString	*theLocalizedTitle;
	NSMenuItem	*theNewParentMenuItem;
	NSMenu		*theNewMenu;

	//	Technical Note:  For top-level menus (i.e. those attached to the menu bar),
	//	theNewMenu's title gets displayed while theNewParentMenuItem's title gets ignored.
	//	For submenus, the reverse is true.  To be safe, pass the same title to both.

	//	Localize aSubmenuTitle to the current language.
	theLocalizedTitle = GetLocalizedTextAsNSString(aSubmenuTitle);

	//	Create a new item for aParentMenu, and a new menu to attach to it.
	theNewParentMenuItem	= [[NSMenuItem alloc] initWithTitle:theLocalizedTitle action:NULL keyEquivalent:@""];
	theNewMenu				= [[NSMenu alloc] initWithTitle:theLocalizedTitle];
	
	//	Attach theNewMenu to theNewParentMenuItem.
	[theNewParentMenuItem setSubmenu:theNewMenu];
	
	//	Add theNewParentMenuItem to aParentMenu.
	[aParentMenu addItem:theNewParentMenuItem];

	//	aParentMenu retains theNewParentMenuItem,
	//	so we can release theNewParentMenuItem.
	[theNewParentMenuItem release];
	
	//	aParentMenu retains theNewParentMenuItem,
	//	and theNewParentMenuItem retains theNewMenu,
	//	so we can release theNewMenu.
	[theNewMenu release];
	
	//	Pass the caller a weak reference to theNewMenu.
	return theNewMenu;
}

- (void)setPreferencePanelLabels
{
	[[itsPrefTabView tabViewItemAtIndex:0] setLabel:
		GetLocalizedTextAsNSString("PrefPerspective")];

	[[itsPrefTabView tabViewItemAtIndex:1] setLabel:
		GetLocalizedTextAsNSString("PrefAdvanced")];

	[itsPrefGeneralBox setTitle:
		GetLocalizedTextAsNSString("PrefGeneral")];

	[itsPrefWindowWidthLabel setStringValue:
		GetLocalizedTextAsNSString("PrefWindowWidth")];

	[itsPrefWindowDistanceLabel setStringValue:
		GetLocalizedTextAsNSString("PrefWindowDistance")];

	[itsPrefStereo3DBox setTitle:
		GetLocalizedTextAsNSString("PrefStereo3D")];

	[itsPrefEyeSeparationLabel setStringValue:
		GetLocalizedTextAsNSString("PrefEyeSpacing")];

	[itsPrefFrameRate setTitle:
		GetLocalizedTextAsNSString("PrefFrameRate")];

	[itsPrefSyncToMonitor setTitle:
		GetLocalizedTextAsNSString("PrefSync")];

	[itsPrefMultisampling setTitle:
		GetLocalizedTextAsNSString("PrefMultisampling")];

	[itsPrefShaders setTitle:
		GetLocalizedTextAsNSString("PrefShaders")];

	[itsPrefOK setTitle:
		GetLocalizedTextAsNSString("PrefOK")];
}

- (void)setSymmetryPanelLabels
{
	[itsSymmetryOK setTitle:GetLocalizedTextAsNSString("SymmetryOK")];
}


- (void)reshape
{
	ResizeGLView(	&md,
					(unsigned int) [self bounds].size.width,
					(unsigned int) [self bounds].size.height);

	//	Refresh the window.  You need to call update first to have it 
	//	come to terms with its context.  If not bizzare things happen.
	//	Adam found this solution at:
	//		http://developer.apple.com/qa/qa2001/qa1209.html
	//
	[self update];
	[self setNeedsDisplay:YES];
}

- (void)windowWillMiniaturize:(NSNotification *)aNotification
{
	//	When the user minimizes a window, Mac OS X creates a dock icon to represent it.
	//	Normally the icon is a low-resolution image of the window's contents,
	//	which Mac OS X creates automatically with no help from the programmer.
	//	Alas that automated process fails for OpenGL windows, so let's read
	//	the OpenGL image buffer ourselves and report its contents via setMiniwindowImage.
	[[self window] setMiniwindowImage:[self contentsAsImage]];
}

- (BOOL)windowShouldClose:(id)sender
{
	//	Caution:  Cocoa calls this method only when the user explictly clicks
	//	the window's close button.  When the user chooses Quit KaleidoTile
	//	we must call this method ourselves.  In the former case this method
	//	will get called twice, but that's OK.

	//	Free the ModelData here, rather than in dealloc(),
	//	so applicationWillTerminate() can reliably test for memory leaks.
	ShutDownModelData(&md);
	
	return YES;
}

- (BOOL)applicationShouldTerminateAfterLastWindowClosed:(NSApplication *)theApplication
{
	return YES;
}

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
{
	//	If the user chooses Quit from the menu,
	//	Cocoa will *not* ask windows whether they should close or not,
	//	so we must do it ourselves.
	if ([self windowShouldClose:self] == YES)
		return NSTerminateNow;
	else
		return NSTerminateCancel;
}

- (void)applicationWillTerminate:(NSNotification *)aNotification
{
	//	Free memory used for localized dictionary.
	RefreshLocalization("--");

	//	We should be clean.
	if (gMemCount != 0)
		ErrorMessage("Memory allocated does not equal memory freed.\r\nPlease report this error to\r\n\twww.geometrygames.org/contact.html", "Memory Leak");
}

@end
