//	KaleidoTile-Win32-WndProc.c
//
//	Provides MainWndProc(), DrawingWndProc() and supporting functions
//	for the Win32 version of KaleidoTile.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "KaleidoTile-Win32.h"
#include "KaleidoTileIDs.h"
#include <commdlg.h>	//	for Save dialog
#include <math.h>		//	for sqrt()
#include <stdio.h>		//	for sprintf()
#include <shellapi.h>	//	for ShellExecute()
#include <commctrl.h>	//	for tab control in preferences panel
#include "GeometryGames-Win32-LoadEntryPoints.h"

//	What is the minimum acceptable size for the window's client area?
#define MIN_CLIENT_SIZE	256


//	Command IDs

#define IDC_FILE_EXIT						0x0000

#define IDC_EXPORT_COPY						0x0100
#define IDC_EXPORT_SAVE						0x0101

#define IDC_VIEW_SYMMETRIES					0x0200
#define IDC_VIEW_POSITIVE_IMAGES			0x0201
#define IDC_VIEW_NEGATIVE_IMAGES			0x0202
#define IDC_VIEW_MIRROR_LINES				0x0203
#define IDC_VIEW_BIGGER_TILES				0x0204
#define IDC_VIEW_SMALLER_TILES				0x0205
#define IDC_VIEW_SNAP						0x0206
#define IDC_VIEW_MULTISAMPLING				0x0207
#define IDC_VIEW_MESH						0x0210
#define IDC_VIEW_MESH02						0x0211
#define IDC_VIEW_MESH04						0x0212
#define IDC_VIEW_MESH08						0x0213
#define IDC_VIEW_MESH16						0x0214
#define IDC_VIEW_TILING_RADIUS				0x0220
#define IDC_VIEW_TILING_RADIUS_40			0x0221
#define IDC_VIEW_TILING_RADIUS_45			0x0222
#define IDC_VIEW_TILING_RADIUS_50			0x0223
#define IDC_VIEW_TILING_RADIUS_55			0x0224
#define IDC_VIEW_TILING_RADIUS_60			0x0225
#define IDC_VIEW_GRAPHICS_FIXED_FUNCTION	0x0230
#define IDC_VIEW_GRAPHICS_SHADERS			0x0231
#define	IDC_VIEW_STEREO_NONE				0x0240
#define	IDC_VIEW_STEREO_ANAGLYPHIC			0x0241
#define	IDC_VIEW_STEREO_GEOWALL				0x0242

#define IDC_LANGUAGE_CY						0x0300
#define IDC_LANGUAGE_EN						0x0301
#define IDC_LANGUAGE_FR						0x0302
#define IDC_LANGUAGE_IT						0x0303
#define IDC_LANGUAGE_JA						0x0304
#define IDC_LANGUAGE_PT						0x0305

#define IDC_HELP_STUDY_QUESTIONS			0x0400
#define IDC_HELP_PREFERENCES				0x0401
#define IDC_HELP_ABOUT						0x0402


//	Support for WGL extensions.

#define WGL_DRAW_TO_WINDOW_ARB         0x2001
#define WGL_ACCELERATION_ARB           0x2003
#define WGL_SUPPORT_OPENGL_ARB         0x2010
#define WGL_DOUBLE_BUFFER_ARB          0x2011
#define WGL_COLOR_BITS_ARB             0x2014
#define WGL_ALPHA_BITS_ARB             0x201B
#define WGL_DEPTH_BITS_ARB             0x2022
#define WGL_STENCIL_BITS_ARB           0x2023
#define WGL_FULL_ACCELERATION_ARB      0x2027
#define WGL_SAMPLE_BUFFERS_ARB         0x2041
#define WGL_SAMPLES_ARB                0x2042

typedef const char	*(WINAPI * PFNWGLGETEXTENSIONSSTRINGARBPROC)(HDC hdc);
typedef BOOL		(WINAPI * PFNWGLCHOOSEPIXELFORMATARBPROC)(HDC hdc, const int *piAttribIList, const FLOAT *pfAttribFList, UINT nMaxFormats, int *piFormats, UINT *nNumFormats);
typedef BOOL		(WINAPI * PFNWGLSWAPINTERVALEXTPROC )(int interval);

static  PFNWGLSWAPINTERVALEXTPROC	wglSwapInterval	= NULL;


//	Remember the Preferences Panel's preferred page.
static unsigned int	gPreferencesPreferredPage = 0;


static bool				SetUpWindowData(HWND aWindow);
static void				ShutDownWindowData(WindowData *wd);
static bool				SetUpOpenGL(WindowData *wd);
static void				ShutDownOpenGL(WindowData *wd);
static bool				SetUpDrawingPanel(WindowData *wd);
static bool				SetUpDrawingPanelWithPixelFormat(WindowData *wd, unsigned int aPixelFormat);
static unsigned int		ChooseMultisamplePixelFormat(WindowData *wd);
static void				ShutDownDrawingPanel(WindowData *wd);
static bool				WholeWordSubstring(const char *aString, const char *aPotentialSubstring);
static void				MenuPrepare(WindowData *wd, HMENU aMenu);
static void				MenuCommand(WindowData *wd, WORD aCommand);
static BOOL CALLBACK	PostCloseMessage(HWND aWindow, LPARAM anUnusedParameter);
static void				CopyTheImage(WindowData *wd);
static void				SaveTheImage(WindowData *wd);
static bool				CopyDirectoryName(WCHAR *aPathName, DWORD aPathNameLength, WCHAR *aDirectoryName, DWORD aDirectoryNameLength);
static HANDLE			FetchTheImage(WindowData *wd);
static void				ChangeLanguage(const char aLanguageCode[3]);
static void				RefreshMirroring(HWND aWindow);
static HMENU			BuildLocalizedMenuBar(void);
static void				OpenHelpBook(void);
static void				SetPrefPanelLabels(HWND aPreferencesDialog);
static void				SwapPreferencesPanel(WindowData *wd, HWND aPreferencesDialog, unsigned int aPageIndex);
static INT_PTR CALLBACK	PreferencesDialogProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);
static INT_PTR CALLBACK PrefsPerspectiveDialogProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);
static void				EnableStereoFields(HWND aDialog, bool aStereoFlag);
static void				FormatUnsignedInteger32(WCHAR *aBuffer, unsigned int aBufferLength, unsigned int aNumber);
static void				UnformatUnsignedInteger32(WCHAR *aBuffer, unsigned int aBufferLength, unsigned int *aNumber);
static INT_PTR CALLBACK PrefsAdvancedDialogProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);
static INT_PTR CALLBACK	SymmetryDialogProc(HWND hDlg, UINT msg, WPARAM wparam, LPARAM lParam);
static void				SetSync(WindowData *wd);
static void				SetMinWindowSize(HWND aWindow, POINT *aMinSize);
static void				ResizeDrawingPanel(WindowData *wd, unsigned int aWidth, unsigned int aHeight);
static void				DisplayChange(WindowData *wd);
static void				GetMouseLocation(ModelData *md, HWND aDrawingWindow, LPARAM lParam, double *theMouseH, double *theMouseV);
static void				MaintainCursor(double aMouseH, WPARAM wParam);
static void				PaintWindow(WindowData *wd);


//	Callback function to process messages in the main window.
LRESULT CALLBACK MainWndProc(
	HWND	hWnd,
	UINT	uMsg,
	WPARAM	wParam,
	LPARAM	lParam)
{
	WindowData	*wd = NULL;

	//	For all messages except WM_CREATE and WM_DESTROY, locate the WindowData.
	//	If we confirm once and for all that wd != NULL, then
	//	the routines we call in the message loop can safely rely on it.
	if (hWnd != NULL)
	{
		if (uMsg != WM_CREATE
		 && uMsg != WM_DESTROY)
		{
			wd = (WindowData *) GetWindowLongPtr(hWnd, GWLP_USERDATA);
			if (wd == NULL)
				return DefWindowProc(hWnd, uMsg, wParam, lParam);
		}
		else
			wd = NULL;
	}
	else
		return 0;	//	should never occur

	//	Process the message.
	switch (uMsg)
	{
		case WM_CREATE:
			if (SetUpWindowData(hWnd))
				return 0;	//	success -- continue normally
			else
				return -1;	//	failure -- abort window creation

		case WM_CLOSE:
			//	Shut down our window data, including itsRenderingContext,
			//	itsDeviceContext and itsDrawingPanel.  Otherwise Windows will destroy
			//	child windows prematurely, leaving itsRenderingContext invalid.
			ShutDownWindowData(wd);
			DestroyWindow(hWnd);
			return 0;

//		case WM_DESTROY:
//			return 0;

//		case WM_KEYDOWN:
//			switch (wParam)	//	virtual key code
//			{
//				//	Note:  Return and Escape may be handled either here or in WM_CHAR,
//				//	but arrow keys don't generate characters so they must be handled here.
//				case VK_UP:
//				case VK_DOWN:
//				case VK_LEFT:
//				case VK_RIGHT:
//				case VK_RETURN:
//				case VK_ESCAPE:
//				case ' ':
//					break;
//			}
//			return 0;
//
//		case WM_CHAR:
//			switch (wParam)	//	character code
//			{
//				case 0x1B:	//	escape (ESC)
//					break;
//
//				case 0x0D:	//	return (CR)
//					break;
//
//				case ' ':
//					break;
//
//				case 'a':
//					break;
//			}
//			return 0;

//		case WM_MOUSEWHEEL:
//			return 0;

		case WM_INITMENUPOPUP:
			MenuPrepare(wd, (HMENU) wParam);
			return 0;

		case WM_COMMAND:
			MenuCommand(wd, LOWORD(wParam));
			return 0;

//		case WM_EXITMENULOOP:
//		case WM_EXITSIZEMOVE:
//		case WM_ACTIVATE:
//			//	Warning:  wd will be NULL when a WM_ACTIVATE event
//			//	arrives during window destruction.
//			return 0;

		case WM_GETMINMAXINFO:
			SetMinWindowSize(hWnd, &((MINMAXINFO*)lParam)->ptMinTrackSize);
			return 0;

		case WM_SIZE:
			MoveWindow(wd->itsDrawingPanel, 0, 0, LOWORD(lParam), HIWORD(lParam), false);
			return 0;

		case WM_DISPLAYCHANGE:
			DisplayChange(wd);
			return 0;

		default:
			return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}
}


//	Callback function to process messages in the drawing panel.
LRESULT CALLBACK DrawingWndProc(
	HWND	aDrawingWindow,
	UINT	uMsg,
	WPARAM	wParam,
	LPARAM	lParam)
{
	WindowData	*wd;
	double		theMouseH,
				theMouseV;

	//	If we confirm once and for all that wd != NULL, then the routines
	//	we call in the message switch can safely rely on it.
	//	See MainWndProc() for an explanation of why the WindowData
	//	isn't available at WM_DESTROY time.
	if (aDrawingWindow != NULL)
	{
		if (uMsg != WM_CREATE
		 && uMsg != WM_DESTROY)
		{
			wd = (WindowData *) GetWindowLongPtr(GetParent(aDrawingWindow), GWLP_USERDATA);
			if (wd == NULL)
				return DefWindowProc(aDrawingWindow, uMsg, wParam, lParam);
		}
		else
			wd = NULL;
	}
	else
		return 0;	//	should never occur

	switch (uMsg)
	{
		case WM_CREATE:
			return 0;

		case WM_LBUTTONDOWN:
			SetCapture(aDrawingWindow);
			GetMouseLocation(&wd->md, aDrawingWindow, lParam, &theMouseH, &theMouseV);
			MouseDown(&wd->md, theMouseH, theMouseV, 0.001 * GetTickCount());
			MaintainCursor(theMouseH, wParam);
			return 0;

		case WM_MOUSEMOVE:
			GetMouseLocation(&wd->md, aDrawingWindow, lParam, &theMouseH, &theMouseV);
			MaintainCursor(theMouseH, wParam);
			if (GetCapture() == aDrawingWindow)
				MouseMove(&wd->md, theMouseH, theMouseV, 0.001 * GetTickCount());
			return 0;

		case WM_LBUTTONUP:
			ReleaseCapture();
			GetMouseLocation(&wd->md, aDrawingWindow, lParam, &theMouseH, &theMouseV);
			MouseUp(&wd->md, theMouseH, theMouseV, 0.001 * GetTickCount());
			MaintainCursor(theMouseH, wParam);
			return 0;

		case WM_PAINT:
			PaintWindow(wd);
			return 0;

		case WM_SIZE:
			ResizeDrawingPanel(wd, LOWORD(lParam), HIWORD(lParam));
			return 0;

		default:
			return DefWindowProc(aDrawingWindow, uMsg, wParam, lParam);
	}
}


static bool SetUpWindowData(HWND aWindow)
{
	WindowData	*wd;
	bool		theShadersAreAvailable;

	//	Attempt to allocate the WindowData and -- whether or not the call
	//	succeeds -- attach the resulting pointer to the window.
	wd = GET_MEMORY(sizeof(WindowData));
	SetWindowLongPtr(aWindow, GWLP_USERDATA, (LONG_PTR) wd);

	//	If the memory allocation failed, return false
	//	to tell Windows to abort window construction.
	if (wd == NULL)
	{
		ErrorMessage("Couldn't get memory for WindowData.", "SetUpWindowData() Error");
		return false;
	}

	//	Record the window.
	wd->itsWindow = aWindow;

	//	Set pointers to NULL for robust error handling.
	wd->itsDrawingPanel			= NULL;
	wd->itsDeviceContext		= NULL;
	wd->itsRenderingContext		= NULL;
	wd->itsPreferencesPane		= NULL;

	//	Initialize the model's internal data.
	SetUpModelData(&wd->md);

	//	Set up OpenGL.
	if ( ! SetUpOpenGL(wd) )
		goto FailedToSetUpData;

	//	Does the OpenGL context support shaders?
	//	If not, tell the user that the fallback fixed-function code
	//	will be used and the quality of the image will be less.
	//	Display that message once, and then never again (for that user).
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	theShadersAreAvailable = ShadersAreAvailable();
	wglMakeCurrent(NULL, NULL);
	if ( ! theShadersAreAvailable
	  && ! GetUserPrefBool("fallback message given"))
	{
		ErrorMessage(GetLocalizedText("ErrorShadersMessage"), GetLocalizedText("ErrorShadersTitle"));
		SetUserPrefBool("fallback message given", true);
	}

	//	Report success.
	return true;

FailedToSetUpData:

	//	Clean up.
	ShutDownWindowData(wd);	//	wd is known to be non-NULL here

	//	Report failure.
	ErrorMessage("Failed to set up window.", "SetUpWindowData() Error");
	return false;
}


static void ShutDownWindowData(WindowData *wd)
{
	ShutDownOpenGL(wd);

	ShutDownModelData(&wd->md);

	//	Once all dependent data have been cleaned up,
	//	then free the WindowData itself.
	SetWindowLongPtr(wd->itsWindow, GWLP_USERDATA, 0);
	FREE_MEMORY(wd);
}


static bool SetUpOpenGL(WindowData *wd)
{
	bool	theLoadResult;

	//	Create a drawing panel, get a device context and
	//	attach an OpenGL rendering context.
	if ( ! SetUpDrawingPanel(wd) )
		return false;

	//	On Win32 we must explicitly load functions from OpenGL 1.2 or later.
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	theLoadResult = LoadOpenGLFunctions();
	wglMakeCurrent(NULL, NULL);	//	Deactivate context before displaying error message.
	if ( ! theLoadResult )
	{
		ErrorMessage("This program cannot continue because your computer's graphics library does not support \"Vertex Buffer Objects\".", "OpenGL 1.5 Required");
		ExitProcess(5);
	}
	
	//	Load an entry point for wglSwapInterval, if it's present.
	//	Caution:  My mother's Dell with Intel Extreme Graphics 
	//	does not support WGL_EXT_swap_control but nevertheless returns a non-NULL
	//	(but ultimately useless) entry point for wglGetProcAddress("wglSwapIntervalEXT").
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	wglSwapInterval = (PFNWGLSWAPINTERVALEXTPROC) wglGetProcAddress("wglSwapIntervalEXT");
	wglMakeCurrent(NULL, NULL);

	//	Set sync-to-monitor as desired.
	SetSync(wd);

	//	The following flags should already be false
	//	(either from the original window initialization
	//	or from a call to ShutDownOpenGL()), but reset them
	//	here just to be safe.
	wd->md.itsPreparedExtensions	= false;
	wd->md.itsPreparedShaders		= false;
	wd->md.itsPreparedTextures		= false;

	return true;
}


static void ShutDownOpenGL(WindowData *wd)
{
	//	Shut down application-specific OpenGL shaders, textures and vertex buffer objects.
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	ShutDownGraphicsAsNeeded(&wd->md);
	wglMakeCurrent(NULL, NULL);

	//	Shut down the drawing panel along with its OpenGL context.
	ShutDownDrawingPanel(wd);
}


static bool SetUpDrawingPanel(WindowData *wd)
{
	unsigned int	thePixelFormat;

	//	The multisampling extension presents a chicken-and-egg problem:
	//	we need to use it to set up the OpenGL rendering context,
	//	but we can't find out whether it's available -- or even get
	//	a pixel format -- until after the rendering context has been created
	//	and made current.

	//	Set up an OpenGL drawing panel with a plain (non-multisampling) buffer.
	if (SetUpDrawingPanelWithPixelFormat(wd, 0) != true)
		return false;

	//	Try for a multi-sample pixel format if the user wants one.
	//	Note:  For sure we'll have at least the default value of
	//	itsMultisamplingEnabledFlag, but at start-up we won't yet
	//	have tested or set itsMultisamplingAvailableFlag.
	//	We hope it's available, but if it's not ChooseMultisamplePixelFormat()
	//	will fail harmlessly.
	if (wd->md.itsMultisamplingEnabledFlag)
	{
		//	Use the plain context to access the WGL_ARB_multisample extension
		//	and select a multisample-capable pixel format.
		thePixelFormat = ChooseMultisamplePixelFormat(wd);

		//	If a multisample-capable pixel format is available,
		//	destroy the plain drawing panel and replace it
		//	with a multisample-capable panel.
		if (thePixelFormat != 0)
		{
			ShutDownDrawingPanel(wd);
			
			if (SetUpDrawingPanelWithPixelFormat(wd, thePixelFormat) != true)
			{
				ErrorMessage("Failed to create multisample-capable OpenGL context.  Falling back to plain OpenGL context.", "Full-scene antialiasing not working");
				wd->md.itsMultisamplingEnabledFlag = false;
				SetUserPrefBool("multisampling", false);
				if (SetUpDrawingPanelWithPixelFormat(wd, 0) != true)
					return false;
			}
		}
	}

	return true;
}


static bool SetUpDrawingPanelWithPixelFormat(
	WindowData		*wd,
	unsigned int	aPixelFormat)	//	needed only for multisampling;  otherwise pass 0
{
	ErrorText				theErrorMessage	= NULL,
							theErrorTitle	= "SetUpDrawingPanel() failed";	//	default title, may be overridden
	RECT					theClientRect;
	PIXELFORMATDESCRIPTOR	thePixelFormatDescriptor =
	{
		sizeof(PIXELFORMATDESCRIPTOR),	//	size
		1,								//	version number
		  PFD_DRAW_TO_WINDOW			//	format
		| PFD_SUPPORT_OPENGL
	//	| PFD_STEREO	//	could be useful in the future, but not needed for anaglyphic stereo 3D
		| PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,					//	preferred format is rgba (not palette)
		32, 							//	preferred color buffer depth (including alpha bits)
		0, 0, 0, 0, 0, 0,				//	color bits ignored
		8,								//	alpha buffer depth
		0,								//	shift bit ignored
		0,								//	no accumulation buffer
		0, 0, 0, 0, 					//	accumulation bits ignored
		0, 								//	no Z-buffer
		0,								//	no stencil buffer
		0,								//	no auxiliary buffer
		0,								//	ignored
		PFD_MAIN_PLANE, 				//	underlay/overlay planes
		0,								//	ignored
		0,								//	underlay color
		0								//	ignored
	};
	
	if (wd->itsDrawingPanel != NULL)
	{
		theErrorMessage = "Window has pre-existing drawing panel.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}
	if (wd->itsDeviceContext != NULL)
	{
		theErrorMessage = "Window has pre-existing device context.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}
	if (wd->itsRenderingContext != NULL)
	{
		theErrorMessage = "Window has pre-existing rendering context.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	//	You'd think you could create itsDrawingPanel once and for all,
	//	and then replace its OpenGL rendering context as needed,
	//	but in practice that doesn't work.  So instead we must
	//	re-create itsDrawingPanel whenever we need to change
	//	the OpenGL rendering context (for example to change
	//	to a multisample-capable pixel format).
	//
	//	Caution:  CreateWindow() send a WM_SIZE message 
	//	which leads to an unwanted ResizeDrawingPanel() call.
	GetClientRect(wd->itsWindow, &theClientRect);
	wd->itsDrawingPanel = CreateWindow(
		DRAWING_WINDOW_CLASS_NAME,
		NULL,
		WS_CHILD | WS_VISIBLE,
		theClientRect.left,	//	should be 0
		theClientRect.top,	//	should be 0
		theClientRect.right - theClientRect.left,
		theClientRect.bottom - theClientRect.top,
		wd->itsWindow, 0, GetModuleHandle(NULL), NULL);
	if (wd->itsDrawingPanel == NULL)
	{
		theErrorMessage = "Can't create a window for OpenGL graphics.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	wd->itsDeviceContext = GetDC(wd->itsDrawingPanel);
	if (wd->itsDeviceContext == NULL)
	{
		theErrorMessage = "Can't get a device context.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	if (aPixelFormat == 0)
	{
		aPixelFormat = ChoosePixelFormat(wd->itsDeviceContext, &thePixelFormatDescriptor);
		if (aPixelFormat == 0)
		{
			theErrorMessage = "Can't get a pixel format.";
			goto CleanUpSetUpDrawingPanelWithPixelFormat;
		}
	}

	if (DescribePixelFormat(wd->itsDeviceContext, aPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &thePixelFormatDescriptor))
	{
		if (thePixelFormatDescriptor.iPixelType	!= PFD_TYPE_RGBA
		 || (	thePixelFormatDescriptor.cColorBits	!= 24	//	Microsoft's documentation claims that cColorBits
		 	 && thePixelFormatDescriptor.cColorBits	!= 32))	//	excludes the alpha bits, but in reality it includes them.
		{
			theErrorMessage	= GetLocalizedText("ErrorColorMessage");
			theErrorTitle	= GetLocalizedText("ErrorColorTitle");
			goto CleanUpSetUpDrawingPanelWithPixelFormat;
		}

		if (thePixelFormatDescriptor.cAlphaBits	!= 8)
		{
			theErrorMessage = GetLocalizedText("ErrorAlphaMessage");
			theErrorTitle	= GetLocalizedText("ErrorAlphaTitle");
			goto CleanUpSetUpDrawingPanelWithPixelFormat;
		}
	}

	if ( ! SetPixelFormat(wd->itsDeviceContext, aPixelFormat, &thePixelFormatDescriptor) )
	{
		theErrorMessage = "Can't set the pixel format.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	wd->itsRenderingContext = wglCreateContext(wd->itsDeviceContext);
	if (wd->itsRenderingContext == NULL)
	{
		theErrorMessage = "Can't get an OpenGL rendering context.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	//	Let's do a "dry run" to test itsRenderingContext right here and now.
	//	That way we won't need to clutter up the rest of the code with
	//	error checking on wglMakeCurrent().  In the highly unlikely event
	//	that wglMakeCurrent() succeeds now and fails later on, nothing
	//	too terrible should happen because when wglMakeCurrent() fails
	//	to set the requested context it instead sets wglMakeCurrent(NULL, NULL),
	//	meaning that all subsequent OpenGL calls will be ignored.
	if ( ! wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext) )
	{
		theErrorMessage = "Can't make OpenGL rendering context current.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	//	The generic Windows OpenGL driver ignores 3D graphics hardware.
	//	In practice this means that a computer from, say, Dell will run
	//	OpenGL fine because Dell provides a good driver, but a standard
	//	installation from a Windows XP CD will run OpenGL in software emulation
	//	(no doubt because Microsoft wants Microsoft-only DirectX graphics
	//	to run well while platform-independent OpenGL graphics runs poorly).
	//	I don't know any direct way to query whether hardware acceleration
	//	is active, but while our rendering context is current we can at least
	//	check the renderer name, and if it's "GDI Generic" notify the user.
	if (strcmp(GetRendererName(), "GDI Generic") == 0)
	{
		theErrorMessage = GetLocalizedText("ErrorDriverMessage");
		theErrorTitle	= GetLocalizedText("ErrorDriverTitle");
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

CleanUpSetUpDrawingPanelWithPixelFormat:

	//	Clear the OpenGL context before calling ErrorMessage().
	wglMakeCurrent(NULL, NULL);

	if (theErrorMessage != NULL)
	{
		ErrorMessage(theErrorMessage, theErrorTitle);
		ShutDownDrawingPanel(wd);
		return false;
	}
	else
		return true;
}


static unsigned int ChooseMultisamplePixelFormat(WindowData *wd)
{
	signed int							thePixelFormat;
	PFNWGLGETEXTENSIONSSTRINGARBPROC	wglGetExtensionsStringARB;
	const char							*theWGLExtensionsList;	//	always ASCII, never Unicode
	PFNWGLCHOOSEPIXELFORMATARBPROC		wglChoosePixelFormatARB;
	signed int							thePixelFormatIntAttributes[] =
	{
		WGL_DRAW_TO_WINDOW_ARB,	true,
		WGL_SUPPORT_OPENGL_ARB,	true,
		WGL_DOUBLE_BUFFER_ARB,	true,
		WGL_ACCELERATION_ARB,	WGL_FULL_ACCELERATION_ARB,
		WGL_COLOR_BITS_ARB,		32,	//	including alpha bits (contrary to some online documentation)
		WGL_ALPHA_BITS_ARB,		8,
		WGL_DEPTH_BITS_ARB,		0,
		WGL_STENCIL_BITS_ARB,	0,
		WGL_SAMPLE_BUFFERS_ARB,	true,
		WGL_SAMPLES_ARB,		8,	//	the code below assumes the number of samples is at array index 19
		0,						0
	};
	float								thePixelFormatFloatAttributes[] = {0.0f, 0.0f};
	bool								theFormatIsValid;
	unsigned int						theNumPixelFormats;

	//	If we encounter an error before finding a pixel format, we'll return 0.
	thePixelFormat = 0;

	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);

	wglGetExtensionsStringARB = (PFNWGLGETEXTENSIONSSTRINGARBPROC) wglGetProcAddress("wglGetExtensionsStringARB");
	if (wglGetExtensionsStringARB == NULL)
		goto CleanUpChooseMultisamplePixelFormat;

	theWGLExtensionsList = wglGetExtensionsStringARB(wd->itsDeviceContext);
	if (theWGLExtensionsList == NULL)
		goto CleanUpChooseMultisamplePixelFormat;

	if ( ! WholeWordSubstring(theWGLExtensionsList, "WGL_ARB_multisample" )
	 ||  ! WholeWordSubstring(theWGLExtensionsList, "WGL_ARB_pixel_format") )
		goto CleanUpChooseMultisamplePixelFormat;

	wglChoosePixelFormatARB = (PFNWGLCHOOSEPIXELFORMATARBPROC) wglGetProcAddress("wglChoosePixelFormatARB");
	if (wglChoosePixelFormatARB == NULL)
		goto CleanUpChooseMultisamplePixelFormat;

	//	Try for 8-sample multisampling, then 6 samples, then 4, then 2, then give up.
	//	Note:  Some applications might want to voluntarily use fewer than 8 samples,
	//	to keep memory usage down.  Conversely, someday we might want to be more
	//	ambitious and try for more than 8 samples.
	do
	{
		theFormatIsValid = wglChoosePixelFormatARB(
								wd->itsDeviceContext,
								thePixelFormatIntAttributes,
								thePixelFormatFloatAttributes,
								1,
								&thePixelFormat,
								&theNumPixelFormats)
						&&
							theNumPixelFormats >= 1;

		//	Make sure the WGL_SAMPLES_ARB parameter sits where we think it sits,
		//	then decrement its value by 2.
		if (thePixelFormatIntAttributes[18] != WGL_SAMPLES_ARB)
			goto CleanUpChooseMultisamplePixelFormat;
		thePixelFormatIntAttributes[19] -= 2;

	} while ( ! theFormatIsValid && thePixelFormatIntAttributes[19] > 0);

	if ( ! theFormatIsValid )
	{
		thePixelFormat = 0;
		goto CleanUpChooseMultisamplePixelFormat;
	}

CleanUpChooseMultisamplePixelFormat:

	wglMakeCurrent(NULL, NULL);

	return thePixelFormat;
}


static void ShutDownDrawingPanel(WindowData *wd)
{
	wglMakeCurrent(NULL, NULL);	//	redundant, but safe

	if (wd->itsRenderingContext != NULL)
	{
		wglDeleteContext(wd->itsRenderingContext);
		wd->itsRenderingContext = NULL;
	}

	if (wd->itsDeviceContext != NULL)
	{
		ReleaseDC(wd->itsDrawingPanel, wd->itsDeviceContext);
		wd->itsDeviceContext = NULL;
	}

	if (wd->itsDrawingPanel != NULL)
	{
		DestroyWindow(wd->itsDrawingPanel);
		wd->itsDrawingPanel = NULL;
	}
}


static bool WholeWordSubstring(
	const char	*aString,				//	null-terminated
	const char	*aPotentialSubstring)	//	null-terminated
{
	const char	*thePotentialMatch,
				*s,
				*p;
	bool		theMatchFlag;

	//	Does aPotentialSubstring occur as a complete space-delimited word in aString?

	//	Irrelevant technical detail:  an empty aPotentialSubstring
	//	(consisting of a terminating zero alone) counts as a substring
	//	of any string, even an empty one.

	//	Consider the first character in each space-delimited word of aString
	//	as the starting point for a potential match.
	thePotentialMatch = aString;
	while (*thePotentialMatch == ' ')
		thePotentialMatch++;
	do
	{
		//	Compare each character in aPotentialSubstring
		//	(excluding the terminating zero) to thePotentialMatch.
		s = thePotentialMatch;
		p = aPotentialSubstring;
		theMatchFlag = true;
		while (theMatchFlag && *p != 0)
		{
			if (*p++ != *s++)
				theMatchFlag = false;
		}

		//	If a space or terminating 0 follows a match, return true.
		if (theMatchFlag
		 && (*s == ' ' || *s == 0))
			return true;

		//	Skip to the beginning of the next whole word in thePotentialMatch.
		//		Skip by non-spaces.
		while (*thePotentialMatch != ' ' && *thePotentialMatch != 0)
			thePotentialMatch++;
		//		Skip by intervening spaces.
		while (*thePotentialMatch == ' ')
			thePotentialMatch++;

		//	Continue with the loop to examine the next whole word (if any).

	} while (*thePotentialMatch != 0);

	return false;
}


static void MenuPrepare(
	WindowData	*wd,
	HMENU		aMenu)
{
	CheckMenuItem(aMenu, IDC_VIEW_POSITIVE_IMAGES,
		wd->md.itsPositiveImagesFlag ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_NEGATIVE_IMAGES,
		wd->md.itsNegativeImagesFlag ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_MIRROR_LINES,
		wd->md.itsMirrorLinesFlag    ? MF_CHECKED : MF_UNCHECKED);

	EnableMenuItem(aMenu, IDC_VIEW_BIGGER_TILES,
			wd->md.itsTiling.itsGeometry == GeometryEuclidean
		 && wd->md.itsTiling.itsEuclideanZoom < MAX_EUCLIDEAN_ZOOM ? MF_ENABLED : MF_GRAYED);
	EnableMenuItem(aMenu, IDC_VIEW_SMALLER_TILES,
			wd->md.itsTiling.itsGeometry == GeometryEuclidean
		 && wd->md.itsTiling.itsEuclideanZoom > MIN_EUCLIDEAN_ZOOM ? MF_ENABLED : MF_GRAYED);

	CheckMenuItem(aMenu, IDC_VIEW_SNAP,
		wd->md.itsSnapToArchimedeanFlag ? MF_CHECKED : MF_UNCHECKED);

	EnableMenuItem(aMenu, IDC_VIEW_MULTISAMPLING,
		wd->md.itsMultisamplingAvailableFlag ? MF_ENABLED : MF_GRAYED);
	CheckMenuItem(aMenu, IDC_VIEW_MULTISAMPLING,
		(wd->md.itsMultisamplingAvailableFlag && wd->md.itsMultisamplingEnabledFlag)
			? MF_CHECKED : MF_UNCHECKED);

	EnableMenuItem(aMenu, IDC_VIEW_MESH,
		wd->md.itsTilingStyle == StyleCurved ? MF_ENABLED : MF_GRAYED);
	CheckMenuItem(aMenu, IDC_VIEW_MESH02,
		wd->md.itsMeshSize ==  2 ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_MESH04,
		wd->md.itsMeshSize ==  4 ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_MESH08,
		wd->md.itsMeshSize ==  8 ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_MESH16,
		wd->md.itsMeshSize == 16 ? MF_CHECKED : MF_UNCHECKED);

	EnableMenuItem(aMenu, IDC_VIEW_TILING_RADIUS,
		wd->md.itsTiling.itsGeometry == GeometryHyperbolic ? MF_ENABLED : MF_GRAYED);
	CheckMenuItem(aMenu, IDC_VIEW_TILING_RADIUS_40,
		wd->md.itsHyperbolicTilingRadius == 4.0 ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_TILING_RADIUS_45,
		wd->md.itsHyperbolicTilingRadius == 4.5 ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_TILING_RADIUS_50,
		wd->md.itsHyperbolicTilingRadius == 5.0 ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_TILING_RADIUS_55,
		wd->md.itsHyperbolicTilingRadius == 5.5 ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_TILING_RADIUS_60,
		wd->md.itsHyperbolicTilingRadius == 6.0 ? MF_CHECKED : MF_UNCHECKED);

	EnableMenuItem(aMenu, IDC_VIEW_GRAPHICS_FIXED_FUNCTION,
		MF_ENABLED);
	CheckMenuItem(aMenu, IDC_VIEW_GRAPHICS_FIXED_FUNCTION,
		! (wd->md.itsShadersAvailableFlag && wd->md.itsShadersEnabledFlag) ? MF_CHECKED : MF_UNCHECKED);
	EnableMenuItem(aMenu, IDC_VIEW_GRAPHICS_SHADERS,
		wd->md.itsShadersAvailableFlag ? MF_ENABLED : MF_GRAYED);
	CheckMenuItem(aMenu, IDC_VIEW_GRAPHICS_SHADERS,
		  (wd->md.itsShadersAvailableFlag && wd->md.itsShadersEnabledFlag) ? MF_CHECKED : MF_UNCHECKED);

	CheckMenuItem(aMenu, IDC_VIEW_STEREO_NONE,
		wd->md.itsStereoMode == StereoNone			? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_STEREO_ANAGLYPHIC,
		wd->md.itsStereoMode == StereoAnaglyphic	? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_VIEW_STEREO_GEOWALL,
		wd->md.itsStereoMode == StereoGeoWall		? MF_CHECKED : MF_UNCHECKED);

	EnableMenuItem(aMenu, IDC_LANGUAGE_CY, IsValidLocale(MAKELCID(MAKELANGID(LANG_WELSH,      SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
	EnableMenuItem(aMenu, IDC_LANGUAGE_EN, IsValidLocale(MAKELCID(MAKELANGID(LANG_ENGLISH,    SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
	EnableMenuItem(aMenu, IDC_LANGUAGE_FR, IsValidLocale(MAKELCID(MAKELANGID(LANG_FRENCH,     SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
	EnableMenuItem(aMenu, IDC_LANGUAGE_IT, IsValidLocale(MAKELCID(MAKELANGID(LANG_ITALIAN,    SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
	EnableMenuItem(aMenu, IDC_LANGUAGE_JA, IsValidLocale(MAKELCID(MAKELANGID(LANG_JAPANESE,   SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
	EnableMenuItem(aMenu, IDC_LANGUAGE_PT, IsValidLocale(MAKELCID(MAKELANGID(LANG_PORTUGUESE, SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);

	CheckMenuItem(aMenu, IDC_LANGUAGE_CY, SAME_LANGUAGE_CODE(gLanguageCode, "cy") ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_LANGUAGE_EN, SAME_LANGUAGE_CODE(gLanguageCode, "en") ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_LANGUAGE_FR, SAME_LANGUAGE_CODE(gLanguageCode, "fr") ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_LANGUAGE_IT, SAME_LANGUAGE_CODE(gLanguageCode, "it") ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_LANGUAGE_JA, SAME_LANGUAGE_CODE(gLanguageCode, "ja") ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(aMenu, IDC_LANGUAGE_PT, SAME_LANGUAGE_CODE(gLanguageCode, "pt") ? MF_CHECKED : MF_UNCHECKED);
}


static void MenuCommand(
	WindowData	*wd,
	WORD		aCommand)
{
	switch (aCommand)
	{
		//	file menu

		case IDC_FILE_EXIT:
			//	Send each window a WM_CLOSE message.
			//	If all windows comply, our message loop in WinMain() will terminate.
			EnumThreadWindows(GetCurrentThreadId(), PostCloseMessage, 0);
			break;

		//	export menu

		case IDC_EXPORT_COPY:
			CopyTheImage(wd);
			break;

		case IDC_EXPORT_SAVE:
			SaveTheImage(wd);
			break;
		
		//	view menu

		case IDC_VIEW_SYMMETRIES:
			DialogBoxParam(	GetModuleHandle(NULL),
							(void *) IDD_SYMMETRY_DIALOG,
							wd->itsWindow,
							(DLGPROC) SymmetryDialogProc,
							(LPARAM) wd);
			break;

		case IDC_VIEW_POSITIVE_IMAGES:
			SetPositiveImagesFlag(&wd->md, ! wd->md.itsPositiveImagesFlag );
			break;

		case IDC_VIEW_NEGATIVE_IMAGES:
			SetNegativeImagesFlag(&wd->md, ! wd->md.itsNegativeImagesFlag );
			break;

		case IDC_VIEW_MIRROR_LINES:
			SetMirrorLinesFlag(&wd->md, ! wd->md.itsMirrorLinesFlag );
			break;
		
		case IDC_VIEW_BIGGER_TILES:
			SetBiggerTiles(&wd->md);
			break;
		
		case IDC_VIEW_SMALLER_TILES:
			SetSmallerTiles(&wd->md);
			break;

		case IDC_VIEW_SNAP:
			wd->md.itsSnapToArchimedeanFlag = ! wd->md.itsSnapToArchimedeanFlag;
			break;

		case IDC_VIEW_MULTISAMPLING:
			SetMultisamplingEnabledFlag(&wd->md, ! wd->md.itsMultisamplingEnabledFlag );
			DisplayChange(wd);
			break;

		case IDC_VIEW_MESH02:	SetMeshSize(&wd->md,  2);	break;
		case IDC_VIEW_MESH04:	SetMeshSize(&wd->md,  4);	break;
		case IDC_VIEW_MESH08:	SetMeshSize(&wd->md,  8);	break;
		case IDC_VIEW_MESH16:	SetMeshSize(&wd->md, 16);	break;

		case IDC_VIEW_TILING_RADIUS_40:	SetHyperbolicTilingRadius(&wd->md, 4.0);	break;
		case IDC_VIEW_TILING_RADIUS_45:	SetHyperbolicTilingRadius(&wd->md, 4.5);	break;
		case IDC_VIEW_TILING_RADIUS_50:	SetHyperbolicTilingRadius(&wd->md, 5.0);	break;
		case IDC_VIEW_TILING_RADIUS_55:	SetHyperbolicTilingRadius(&wd->md, 5.5);	break;
		case IDC_VIEW_TILING_RADIUS_60:	SetHyperbolicTilingRadius(&wd->md, 6.0);	break;

		case IDC_VIEW_GRAPHICS_FIXED_FUNCTION:	SetShadersEnabledFlag(&wd->md, false);	break;
		case IDC_VIEW_GRAPHICS_SHADERS:			SetShadersEnabledFlag(&wd->md, true );	break;

		case IDC_VIEW_STEREO_NONE:			SetStereo3DMode(&wd->md, StereoNone			);	break;
		case IDC_VIEW_STEREO_ANAGLYPHIC:	SetStereo3DMode(&wd->md, StereoAnaglyphic	);	break;
		case IDC_VIEW_STEREO_GEOWALL:		SetStereo3DMode(&wd->md, StereoGeoWall		);	break;
	
		//	language menu

		case IDC_LANGUAGE_CY:	ChangeLanguage("cy");	break;
		case IDC_LANGUAGE_EN:	ChangeLanguage("en");	break;
		case IDC_LANGUAGE_FR:	ChangeLanguage("fr");	break;
		case IDC_LANGUAGE_IT:	ChangeLanguage("it");	break;
		case IDC_LANGUAGE_JA:	ChangeLanguage("ja");	break;
		case IDC_LANGUAGE_PT:	ChangeLanguage("pt");	break;

		//	help menu

		case IDC_HELP_STUDY_QUESTIONS:
			OpenHelpBook();
			break;

		case IDC_HELP_PREFERENCES:
			DialogBoxParam(	GetModuleHandle(NULL),
							(void *) IDD_PREFERENCES_DIALOG,
							wd->itsWindow,
							(DLGPROC) PreferencesDialogProc,
							(LPARAM) wd);
			break;

		case IDC_HELP_ABOUT:
			//	Building a fancy dialog box with graphics is tricky
			//	because of the conflict between the "dialog layout units"
			//	and the pixel dimensions of bitmaps.  A simple but humble
			//	solution relays basic information in a MessageBox().
			ErrorMessage(GetLocalizedText("AboutBoxMessage"), GetLocalizedText("AboutBoxTitle"));
			break;
	}
}


static BOOL CALLBACK PostCloseMessage(
	HWND	aWindow,
	LPARAM	anUnusedParameter)
{
	UNUSED_PARAMETER(anUnusedParameter);

	//	Skip windows that aren't ours.
	//
	//	Note:  On my system the Input Method Environment (IME) creates
	//	secret windows that shadow the main window.  They are top-level windows,
	//	so EnumThreadWindows() finds them.  Specifically, each main window owns
	//	a secret window of class "IME", which in turn owns another secret window
	//	of class "MSCTFIME UI".
	if (GetClassLong(aWindow, GCW_ATOM) == gMainWindowClass)
		PostMessage(aWindow, WM_CLOSE, 0, 0);

	return true;	//	keep going
}


static void CopyTheImage(WindowData *wd)
{
	ErrorText	theErrorMessage		= NULL;
	HANDLE		theDibHandle		= NULL;
	bool		theClipboardIsOpen	= false;

	//	Read the image as a DIB.
	theDibHandle = FetchTheImage(wd);
	if (theDibHandle == NULL)
	{
		theErrorMessage = "Couldn't read image to copy to clipboard.";
		goto CleanUpCopyTheImage;
	}

	//	Transfer the DIB to the clipboard.

	if (OpenClipboard(NULL) == 0)
	{
		theErrorMessage = "Couldn't open clipboard.";
		goto CleanUpCopyTheImage;
	}
	theClipboardIsOpen = true;

	if (EmptyClipboard() == 0)
	{
		theErrorMessage = "Couldn't delete previous clipboard contents.";
		goto CleanUpCopyTheImage;
	}

	if (SetClipboardData(CF_DIB, theDibHandle) == NULL)
	{
		theErrorMessage = "Couldn't transfer image to clipboard.";
		goto CleanUpCopyTheImage;
	}
	//	The clipboard now owns the DIB.
	theDibHandle = NULL;

CleanUpCopyTheImage:

	if (theClipboardIsOpen)
		CloseClipboard();

	if (theDibHandle != NULL)
		GlobalFree(theDibHandle);

	if (theErrorMessage != NULL)
		ErrorMessage(theErrorMessage, "Copy Error");
}


static void SaveTheImage(WindowData *wd)
{
	ErrorText			theErrorMessage		= NULL;
	WCHAR				theFileName[1024]	= {0};
	OPENFILENAME		theFileInfo			= {sizeof(OPENFILENAME),
		NULL, NULL, L"32-bit Bitmap (*.bmp)\0*.bmp\0", NULL, 0, 0,
		theFileName, BUFFER_LENGTH(theFileName), NULL, 0,
		gLastImageDirectory, NULL, OFN_OVERWRITEPROMPT, 0, 0,
		L"bmp", 0L, NULL, NULL};
	HANDLE				theDibHandle		= NULL;
	unsigned int		theDIBSize			= 0;
	HANDLE				theFile				= NULL;
	BITMAPFILEHEADER	theFileHeader		= {0x4D42, 0, 0, 0, 0};
	DWORD				theNumBytes			= 0;
	Byte				*theDib				= NULL;

	//	Prompt the user for a file name.
	//	The extension will be .bmp (see theFileInfo above).
	if (GetSaveFileName(&theFileInfo) == 0)
	{
		//	The user cancelled or an error occurred.
		goto CleanUpSaveTheImage;
	}

	//	Remember the directory name for next time.
	CopyDirectoryName(	theFileName,
						BUFFER_LENGTH(theFileName),
						gLastImageDirectory,
						BUFFER_LENGTH(gLastImageDirectory));

	//	Read the image as a DIB.
	theDibHandle = FetchTheImage(wd);
	if (theDibHandle == NULL)
	{
		theErrorMessage = "Couldn't read image to save to file.";
		goto CleanUpSaveTheImage;
	}
	
	//	How big is the DIB?
	theDIBSize = GlobalSize(theDibHandle);
	if (theDIBSize == 0)
	{
		theErrorMessage = "Couldn't get image size.";
		goto CleanUpSaveTheImage;
	}
	
	//	Open the output file.
	theFile = CreateFile(	theFileName,
							GENERIC_WRITE,
							0,
							NULL,
							CREATE_ALWAYS,
							FILE_ATTRIBUTE_NORMAL,
							NULL);
	if (theFile == NULL)
	{
		theErrorMessage = "Couldn't create file to save image.";
		goto CleanUpSaveTheImage;
	}
	
	//	Save the DIB to the file.
	
	//		First the BITMAPFILEHEADER.
	theFileHeader.bfSize	= sizeof(BITMAPFILEHEADER) + theDIBSize;
	theFileHeader.bfOffBits	= sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	if (WriteFile(theFile, &theFileHeader, sizeof(BITMAPFILEHEADER), &theNumBytes, NULL) == 0
	 || theNumBytes != sizeof(BITMAPFILEHEADER))
	{
		theErrorMessage = "Couldn't write bitmap file header.";
		goto CleanUpSaveTheImage;
	}
	
	//		Then the DIB itself.
	theDib = (Byte *) GlobalLock(theDibHandle);
	if (theDib == NULL)
	{
		theErrorMessage = "Couldn't lock the image memory to save to file.";
		goto CleanUpSaveTheImage;
	}
	if (WriteFile(theFile, theDib, theDIBSize, &theNumBytes, NULL) == 0
	 || theNumBytes != theDIBSize)
	{
		theErrorMessage = "Couldn't write DIB to file.";
		goto CleanUpSaveTheImage;
	}
	GlobalUnlock(theDibHandle);
	theDib = NULL;

CleanUpSaveTheImage:

	if (theDib != NULL)
		GlobalUnlock(theDibHandle);

	if (theFile != NULL)
		CloseHandle(theFile);

	if (theDibHandle != NULL)
		GlobalFree(theDibHandle);

	if (theErrorMessage != NULL)
		ErrorMessage(theErrorMessage, "Save Error");
}


static bool CopyDirectoryName(
	WCHAR	*aPathName,				//	input, 	NULL-terminated string of WCHARs
	DWORD	aPathNameLength,		//	input, 	size of input  buffer in WCHARs, including terminating zero
	WCHAR	*aDirectoryName,		//	output, NULL-terminated string of WCHARs
	DWORD	aDirectoryNameLength)	//	input,	size of output buffer in WCHARs, including terminating zero
{
	//	Copy aPathName to aDirectoryName, excluding the file name itself.
	
	DWORD	thePosition,
			i;
	
	//	Will aDirectoryName accommodate a null string in case of error?
	if (aDirectoryName == NULL
	 || aDirectoryNameLength == 0)
	{
		//	Return *without* setting an empty output string.
		return false;
	}
	else
	{
		//	Preinitialize an empty output string.
		aDirectoryName[0] = 0;
	}

	//	Find the terminating zero at the end of the input path name.
	for (thePosition = 0; thePosition < aPathNameLength; thePosition++)
		if (aPathName[thePosition] == 0)
			break;
	if (thePosition == aPathNameLength)
		return false;	//	no terminating zero
	
	//	Back up to the last backslash character.
	while (thePosition > 0
	    && aPathName[thePosition] != '\\')
	{
		thePosition--;
	}
	if (thePosition == 0)
		return false;	//	no backslash found
	
	if (aDirectoryNameLength < thePosition + 2)
		return false;	//	output buffer too small
	
	//	Copy the directory name from aPathName to aDirectoryName.
	for (i = 0; i <= thePosition; i++)
		aDirectoryName[i] = aPathName[i];
	
	//	Append a terminating zero.
	aDirectoryName[thePosition + 1] = 0;
	
	//	Success
	return true;
}


static HANDLE FetchTheImage(WindowData *wd)
{
	ErrorText			theErrorMessage			= NULL;
	RECT				theClientRect			= {0,0,0,0};
	unsigned int		theImageWidth			= 0,
						theImageHeight			= 0,
						theImageSize			= 0,
						theHeaderSize			= 0,
						theTotalSize			= 0;
	HANDLE				theDibHandle			= NULL;
	Byte				*theDib					= NULL;
	BITMAPINFOHEADER	*theHeader				= NULL;
	Byte				*theImage				= NULL;
	unsigned int		r						= 0,
						c						= 0;
	Byte				*thePixel				= NULL,
						theSwap					= 0;

	//	How big is the image?
	if (GetClientRect(wd->itsDrawingPanel, &theClientRect) == 0
	 || theClientRect.left != 0
	 || theClientRect.top  != 0
	 || (theImageWidth  = theClientRect.right)  == 0
	 || (theImageHeight = theClientRect.bottom) == 0)
	{
		theErrorMessage = "Couldn't read window size.";
		goto CleanUpFetchTheImage;
	}

	//	How many bytes do we have?
	//	(If we were reading RGB instead of RGBA we'd need to pad the rows
	//	to insure a row-stride that's a multiple of four, but for RGBA
	//	it's automatic.  Unfortunately the alpha channel will be ignored,
	//	but let's include the alpha values anyhow.)
	theImageSize = theImageWidth * theImageHeight * 4;

	//	Create a DIB.

	theHeaderSize	= sizeof(BITMAPINFOHEADER);
	theTotalSize	= theHeaderSize + theImageSize;

	theDibHandle = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, theTotalSize);
	if (theDibHandle == NULL)
	{
		theErrorMessage = "Not enough memory to copy image.";
		goto CleanUpFetchTheImage;
	}

	theDib = (Byte *) GlobalLock(theDibHandle);
	if (theDib == NULL)
	{
		theErrorMessage = "Couldn't lock the image memory to write image data.";
		goto CleanUpFetchTheImage;
	}
	theHeader	= (BITMAPINFOHEADER *) theDib;
	theImage	= theDib + theHeaderSize;

	theHeader->biSize			= sizeof(BITMAPINFOHEADER);
	theHeader->biWidth			= theImageWidth;
	theHeader->biHeight			= theImageHeight;
	theHeader->biPlanes			= 1;
	theHeader->biBitCount		= 32;
	theHeader->biCompression	= BI_RGB;
	theHeader->biSizeImage		= 0;
	theHeader->biXPelsPerMeter	= 0;
	theHeader->biYPelsPerMeter	= 0;
	theHeader->biClrUsed		= 0;
	theHeader->biClrImportant	= 0;
	
	//	The back buffer may or may not contain the current image,
	//	depending on whether the driver swaps pointers to two buffers
	//	or whether it copies the contents of fixed back buffer
	//	to a fixed front buffer.  To play it safe, ReadRGBAPixels()
	//	always reads the front buffer.
	//	The only problem with this is that we end up copying
	//	the Export menu's fade effect.  To work around that,
	//	sleep for 1sec before reading the buffer.
//	PaintWindow(wd);	unnecessary
	Sleep(1000);

	//	Read the pixels as RGBA.
	//
	//	Note:  Microsoft's BITMAP format does not support transparency --
	//	the last byte of each pixel gets ignored!
	//	ReadRGBAPixels() returns pixels with premultiplied alpha,
	//	which the BITMAP format interprets as fading to black.
	//	Given the lack of transparency support, this is the best
	//	outcome we could have hoped for.
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	ReadRGBAPixels(theImageWidth, theImageHeight, theImage);
	wglMakeCurrent(NULL, NULL);

	//	Swap the byte order manually to conform to .bmp conventions.
	thePixel = theImage;
	for (r = 0; r < theImageHeight; r++)
		for (c = 0; c < theImageWidth; c++)
		{
			theSwap		= thePixel[0];
			thePixel[0]	= thePixel[2];
			thePixel[2]	= theSwap;
			
			thePixel += 4;
		}

	//	Unlock the handle.
	GlobalUnlock(theDibHandle);
	theDib = NULL;

CleanUpFetchTheImage:

	if (theDib != NULL)
		GlobalUnlock(theDibHandle);

	if (theErrorMessage != NULL)
	{
		if (theDibHandle != NULL)
		{
			GlobalFree(theDibHandle);
			theDibHandle = NULL;
		}

		ErrorMessage(theErrorMessage, "Image Export Error");
	}
	
	return theDibHandle;
}


static void ChangeLanguage(
	const char	aLanguageCode[3])
{
	RefreshLocalization(aLanguageCode);
	EnumThreadWindows(GetCurrentThreadId(), RefreshLanguage, 0);
}


BOOL CALLBACK RefreshLanguage(
	HWND	aWindow,
	LPARAM	anUnusedParameter)
{
	WCHAR		theWindowTitle[128];
	HMENU		theOldMenu,
				theNewMenu;
	WindowData	*wd				= NULL;

	UNUSED_PARAMETER(anUnusedParameter);

	//	Skip windows that aren't ours.
	//
	//	Note:  On my system the Input Method Environment (IME) creates
	//	secret windows that shadow the main window.  They are top-level windows,
	//	so EnumThreadWindows() finds them.  Specifically, each main window owns
	//	a secret window of class "IME", which in turn owns another secret window
	//	of class "MSCTFIME UI".
	if (GetClassLong(aWindow, GCW_ATOM) != gMainWindowClass)
		return true;

	//	Reset the window's title.
	SetWindowText(aWindow, GetLocalizedTextAsWCHAR("WindowTitle", theWindowTitle, BUFFER_LENGTH(theWindowTitle)));

	//	Reset the mirroring.
	RefreshMirroring(aWindow);

	//	Replace the window's old menu with a new menu in the new language.
	//	Note:  LoadMenu() creates a private menu for each window,
	//	even though they are all based on the same resource.
	//	Thus we may safely translate the new menu and destroy the old one.
	theOldMenu = GetMenu(aWindow);
	theNewMenu = BuildLocalizedMenuBar();
	SetMenu(aWindow, theNewMenu);
	DestroyMenu(theOldMenu);
	DrawMenuBar(aWindow);
	
	//	Ask that textures be reloaded,
	//	so the control panel labels will appear
	//	in the new language.
	wd = (WindowData *) GetWindowLongPtr(aWindow, GWLP_USERDATA);
	wd->md.itsPreparedTextures = false;
	wd->md.itsRedrawRequestFlag = true;

	return true;	//	keep going
}


static void RefreshMirroring(HWND aWindow)
{
	//	Reset the window layout, left-to-right (LTR) or right-to-left (RTL).

	if (LanguageReadsRightToLeft())
	{
		//	Mirror the main window.
		//	Mirror child windows only if you want to -- it isn't necessary.
		SetWindowLongPtr(	aWindow,
							GWL_EXSTYLE,
							GetWindowLongPtr(aWindow, GWL_EXSTYLE) |  WS_EX_LAYOUTRTL);
	}
	else
	{
		SetWindowLongPtr(	aWindow,
							GWL_EXSTYLE,
							GetWindowLongPtr(aWindow, GWL_EXSTYLE) & ~WS_EX_LAYOUTRTL);
	}
}


static HMENU BuildLocalizedMenuBar(void)
{
	HMENU			theMainMenu		= NULL,
					theSubMenu		= NULL,
					theSubSubMenu	= NULL;
	WCHAR			theBuffer[2048];
	MENUITEMINFO	theMenuItemInfo;

	theMainMenu = CreateMenu();
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("File", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_FILE_EXIT, GetLocalizedTextAsWCHAR("Exit", theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubMenu = NULL;
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("Export", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_EXPORT_COPY, GetLocalizedTextAsWCHAR("Copy Image",  theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_EXPORT_SAVE, GetLocalizedTextAsWCHAR("Save Image…", theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubMenu = NULL;
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("View", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_SYMMETRIES,      GetLocalizedTextAsWCHAR("Choose Symmetries…",         theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                        NULL                                                                                      );
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_POSITIVE_IMAGES, GetLocalizedTextAsWCHAR("Show Plain Images",          theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_NEGATIVE_IMAGES, GetLocalizedTextAsWCHAR("Show Reflected Images",      theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_MIRROR_LINES,    GetLocalizedTextAsWCHAR("Cut Along Mirror Lines",     theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                        NULL                                                                                      );
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_BIGGER_TILES,    GetLocalizedTextAsWCHAR("Make Tiles Bigger",          theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_SMALLER_TILES,   GetLocalizedTextAsWCHAR("Make Tiles Smaller",         theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                        NULL                                                                                      );
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_SNAP,            GetLocalizedTextAsWCHAR("Snap to Archimedean Solids", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                        NULL                                                                                      );
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_MULTISAMPLING,   GetLocalizedTextAsWCHAR("Smooth Edges",               theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubSubMenu = CreateMenu();
		//	We want the "Mesh Refinement" item to have a menu ID
		//	(so it can be disabled for tilings of the Euclidean plane)
		//	but also have a subsubmenu (so it can be used!).
		//	This requires two function calls:
		//	AppendMenu() will set the menu ID, then
		//	SetMenuItemInfo() will add the subsubmenu.
		AppendMenu(theSubMenu, MF_STRING, IDC_VIEW_MESH, GetLocalizedTextAsWCHAR("Mesh Refinement", theBuffer, BUFFER_LENGTH(theBuffer)));
		theMenuItemInfo.cbSize		= sizeof(MENUITEMINFO);
		theMenuItemInfo.fMask		= MIIM_SUBMENU;
		theMenuItemInfo.hSubMenu	= theSubSubMenu;
		SetMenuItemInfo(theSubMenu, IDC_VIEW_MESH, false, &theMenuItemInfo);
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_MESH02, GetLocalizedTextAsWCHAR("Coarse (×2)",      theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_MESH04, GetLocalizedTextAsWCHAR("Medium (×4)",      theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_MESH08, GetLocalizedTextAsWCHAR("Fine (×8)",        theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_MESH16, GetLocalizedTextAsWCHAR("Extra Fine (×16)", theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubSubMenu = NULL;
	theSubSubMenu = CreateMenu();
		//	We want the "Tiling Radius" item to have a menu ID
		//	(so it can be disabled for non-hyperbolic tilings)
		//	but also have a subsubmenu (so it can be used!).
		//	This requires two function calls:
		//	AppendMenu() will set the menu ID, then
		//	SetMenuItemInfo() will add the subsubmenu.
		AppendMenu(theSubMenu, MF_STRING, IDC_VIEW_TILING_RADIUS, GetLocalizedTextAsWCHAR("Tiling Radius", theBuffer, BUFFER_LENGTH(theBuffer)));
		theMenuItemInfo.cbSize		= sizeof(MENUITEMINFO);
		theMenuItemInfo.fMask		= MIIM_SUBMENU;
		theMenuItemInfo.hSubMenu	= theSubSubMenu;
		SetMenuItemInfo(theSubMenu, IDC_VIEW_TILING_RADIUS, false, &theMenuItemInfo);
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_TILING_RADIUS_40, GetLocalizedTextAsWCHAR("4.0", theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_TILING_RADIUS_45, GetLocalizedTextAsWCHAR("4.5", theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_TILING_RADIUS_50, GetLocalizedTextAsWCHAR("5.0", theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_TILING_RADIUS_55, GetLocalizedTextAsWCHAR("5.5", theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_TILING_RADIUS_60, GetLocalizedTextAsWCHAR("6.0", theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubSubMenu = NULL;
	theSubSubMenu = CreateMenu();
		AppendMenu(theSubMenu, MF_POPUP | MF_STRING, (UINT_PTR)theSubSubMenu, GetLocalizedTextAsWCHAR("Graphics Pipeline", theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_GRAPHICS_FIXED_FUNCTION, GetLocalizedTextAsWCHAR("Fixed-Function", theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_GRAPHICS_SHADERS,        GetLocalizedTextAsWCHAR("Shaders",        theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubSubMenu = NULL;
	theSubSubMenu = CreateMenu();
		AppendMenu(theSubMenu, MF_POPUP | MF_STRING, (UINT_PTR)theSubSubMenu, GetLocalizedTextAsWCHAR("Stereoscopic 3D", theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_STEREO_NONE,       GetLocalizedTextAsWCHAR("None",       theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_STEREO_ANAGLYPHIC, GetLocalizedTextAsWCHAR("Anaglyphic", theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_VIEW_STEREO_GEOWALL,    GetLocalizedTextAsWCHAR("GeoWall",    theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubSubMenu = NULL;
	theSubMenu = NULL;
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("Language", theBuffer, BUFFER_LENGTH(theBuffer)));
	//	To add a new language, please see the instructions
	//	in the file "To add a new language.txt".
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_CY, L"Cymraeg"  );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_EN, L"English"  );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_FR, L"Français" );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_IT, L"Italiano" );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_JA, L"日本語"      );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_PT, L"Português");
	theSubMenu = NULL;
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("Help", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_HELP_STUDY_QUESTIONS, GetLocalizedTextAsWCHAR("KaleidoTile Study Questions", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_HELP_PREFERENCES,     GetLocalizedTextAsWCHAR("Preferences…",                theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                        NULL                                                                                       );
	AppendMenu(theSubMenu, MF_STRING,    IDC_HELP_ABOUT,           GetLocalizedTextAsWCHAR("About KaleidoTile…",          theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubMenu = NULL;
	
	return theMainMenu;
}


static void OpenHelpBook(void)
{
	char	theFileName[64]			= "",
			theFilePath[2048]		= "";
	WCHAR	theFilePathUTF16[2048]	= L"";

	sprintf(theFileName, "KaleidoTileStudyQuestions-%c%c.html", gLanguageCode[0], gLanguageCode[1]);

	if (GetAbsolutePath("Help", theFileName, theFilePath, BUFFER_LENGTH(theFilePath)) == NULL
	 && MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, theFilePath, -1, theFilePathUTF16, BUFFER_LENGTH(theFilePathUTF16)) != 0)
	{
		ShellExecute(	NULL,
						L"open",
						theFilePathUTF16,
						NULL,
						NULL,
						SW_SHOWNORMAL);
	}
}


static INT_PTR CALLBACK	PreferencesDialogProc(
	HWND	hDlg,
	UINT	msg,
	WPARAM	wParam,
	LPARAM	lParam)
{
	WindowData		*wd				= NULL;
	HWND			theTabControl;
	unsigned int	i;
	
	if (msg != WM_INITDIALOG)
	{
		wd = (WindowData *) GetWindowLongPtr(hDlg, GWLP_USERDATA);
		if (wd == NULL)	//	should never occur for messages that interest us
			return false;	//	message ignored
	}

	switch (msg)
	{
		case WM_INITDIALOG:

			//	lParam points to the parent window's WindowData.
			//	Stash a copy in the dialog box as well.
			wd = (WindowData *) lParam;
			SetWindowLongPtr(hDlg, GWLP_USERDATA, (LONG_PTR) wd);
			RefreshMirroring(hDlg);
			if (wd == NULL)	//	should never occur
				return true;	//	set focus normally

			theTabControl = GetDlgItem(hDlg, IDC_PREFS_TAB_CONTROL);

			for (i = 0; i < 2; i++)
				SendMessage(theTabControl, TCM_INSERTITEM, i, (LPARAM) &(TCITEM){0, 0, 0, NULL, 0, -1, 0});

			SendMessage(theTabControl, TCM_SETCURSEL, gPreferencesPreferredPage, 0);

			SetPrefPanelLabels(hDlg);

			SwapPreferencesPanel(wd, hDlg, gPreferencesPreferredPage);

			return true;	//	set focus normally

		case WM_DESTROY:
			wd->itsPreferencesPane = NULL;
			SetWindowLong(hDlg, DWL_MSGRESULT, 0);	//	return value
			return true;	//	message processed

		case WM_NOTIFY:
			switch (((LPNMHDR)lParam)->code)
			{
				case TCN_SELCHANGE:
					gPreferencesPreferredPage = SendMessage( ((LPNMHDR)lParam)->hwndFrom, TCM_GETCURSEL, 0, 0 );
					SwapPreferencesPanel(wd, hDlg, gPreferencesPreferredPage);
					break;
			}
			//	no specific return value required
			return true;	//	message processed

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDCANCEL:	//	the Close Box sends IDCANCEL
					EndDialog(hDlg, true);
					break;
			}
			SetWindowLong(hDlg, DWL_MSGRESULT, 0);	//	return value
			return true;	//	message processed

		default:
			return false;	//	message ignored
	}
}


static void SetPrefPanelLabels(
	HWND	aPreferencesDialog)
{
	WCHAR	theCaption[128]	= L"",
			theLabel[64]	= L"";
	HWND	theTabControl	= NULL;

	//	Set the window caption.
	SetWindowText(aPreferencesDialog,
		GetLocalizedTextAsWCHAR("PrefTitle", theCaption, BUFFER_LENGTH(theCaption)));

	//	Set the tab labels.
	theTabControl = GetDlgItem(aPreferencesDialog, IDC_PREFS_TAB_CONTROL);
	SendMessage(theTabControl, TCM_SETITEM, 0, (LPARAM) &(TCITEM){TCIF_TEXT, 0, 0, 
		GetLocalizedTextAsWCHAR("PrefPerspective", theLabel, BUFFER_LENGTH(theLabel)), 0, -1, 0});
	SendMessage(theTabControl, TCM_SETITEM, 1, (LPARAM) &(TCITEM){TCIF_TEXT, 0, 0, 
		GetLocalizedTextAsWCHAR("PrefAdvanced",    theLabel, BUFFER_LENGTH(theLabel)), 0, -1, 0});
}


static void SwapPreferencesPanel(
	WindowData		*wd,
	HWND			aPreferencesDialog,
	unsigned int	aPageIndex)
{
	static unsigned int	theResourceId[2] =
						{
							IDD_PREFS_PERSPECTIVE_DIALOG,
							IDD_PREFS_ADVANCED_DIALOG
						};
	static DLGPROC		thePaneProcedure[2] =
						{
							PrefsPerspectiveDialogProc,
							PrefsAdvancedDialogProc
						};

	if (wd->itsPreferencesPane != NULL)
	{
		DestroyWindow(wd->itsPreferencesPane);
		wd->itsPreferencesPane = NULL;
	}

	wd->itsPreferencesPane = CreateDialogParam(	GetModuleHandle(NULL),
												(LPCTSTR) theResourceId[aPageIndex],
												aPreferencesDialog,
												thePaneProcedure[aPageIndex],
												(LPARAM) wd);
	InvalidateRect(aPreferencesDialog, NULL, true);
}


static INT_PTR CALLBACK PrefsPerspectiveDialogProc(
	HWND	hDlg,
	UINT	msg,
	WPARAM	wParam,
	LPARAM	lParam)
{
	WindowData		*wd;
	WCHAR			theBuffer[11],
					theLabel[256];
	unsigned int	theNumber;
	double			theAspectRatio;

	//	Technical note:  Using integer fields for the window width and height
	//	works fine if you don't resize the window.  But if you do resize the window
	//	you get a non-integer width and height.  In and of itself that's OK.
	//	But when you call up this dialog it writes and then re-reads the parameters,
	//	in effect forcing the window width and height to the nearest integers,
	//	which causes a slight jump.  The solution, of course, would be text boxes
	//	that allowed floating point values.

	switch (msg)
	{
		case WM_INITDIALOG:

			//	lParam contains a pointer to the parent window's WindowData.
			//	Stash a copy in this dialog box as well.
			wd = (WindowData *) lParam;
			SetWindowLongPtr(hDlg, GWLP_USERDATA, (LONG_PTR) wd);
			if (wd == NULL)	//	should never occur
				return true;	//	set focus normally

			//	Initialize the stereo 3D checkbox and
			//	enable corresponding text fields accordingly.
			EnableStereoFields(hDlg, wd->md.itsStereoMode != StereoNone);

			//	Copy our internal variables into the dialog's edit fields.

			FormatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), (unsigned int) (2 * wd->md.itsWindowHalfWidth));
			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_WINDOW_WIDTH   ), theBuffer);

			FormatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), (unsigned int) wd->md.itsWindowDistance);
			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_WINDOW_DISTANCE), theBuffer);

			FormatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), (unsigned int) (2 * wd->md.itsEyeOffset));
			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_EYE_SPACING    ), theBuffer);

			//	Initialize the labels in the current language.

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_GENERAL_LABEL),
				GetLocalizedTextAsWCHAR("PrefGeneral",        theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_WINDOW_WIDTH_LABEL),
				GetLocalizedTextAsWCHAR("PrefWindowWidth",    theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_WINDOW_DISTANCE_LABEL),
				GetLocalizedTextAsWCHAR("PrefWindowDistance", theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_STEREO_LABEL),
				GetLocalizedTextAsWCHAR("PrefStereo3D",       theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_EYE_SPACING_LABEL),
				GetLocalizedTextAsWCHAR("PrefEyeSpacing",     theLabel, BUFFER_LENGTH(theLabel)));

			return true;	//	set focus normally

		case WM_COMMAND:

			wd = (WindowData *) GetWindowLongPtr(hDlg, GWLP_USERDATA);
			if (wd != NULL)
			{
				switch (LOWORD(wParam))
				{
					case IDC_PREFS_PERSPECTIVE_WINDOW_WIDTH:

						if (wd->md.itsWindowHalfHeight > 0.0)	//	should always be the case
							theAspectRatio = wd->md.itsWindowHalfWidth / wd->md.itsWindowHalfHeight;
						else
							theAspectRatio = 1.0;	//	emergency use only
				
						GetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_WINDOW_WIDTH   ), theBuffer, BUFFER_LENGTH(theBuffer));
						UnformatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), &theNumber);
						wd->md.itsWindowHalfWidth	= ((double) theNumber) / 2;
						wd->md.itsWindowHalfHeight	= wd->md.itsWindowHalfWidth / theAspectRatio;

						break;

					case IDC_PREFS_PERSPECTIVE_WINDOW_DISTANCE:
						GetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_WINDOW_DISTANCE), theBuffer, BUFFER_LENGTH(theBuffer));
						UnformatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), &theNumber);
						wd->md.itsWindowDistance = theNumber;
						break;

					case IDC_PREFS_PERSPECTIVE_EYE_SPACING:
						GetWindowText(GetDlgItem(hDlg, IDC_PREFS_PERSPECTIVE_EYE_SPACING    ), theBuffer, BUFFER_LENGTH(theBuffer));
						UnformatUnsignedInteger32(theBuffer, BUFFER_LENGTH(theBuffer), &theNumber);
						wd->md.itsEyeOffset = ((double) theNumber) / 2;
						break;
				}
				PaintWindow(wd);
			}
			SetWindowLong(hDlg, DWL_MSGRESULT, 0);	//	return value
			return true;	//	message processed

		default:
			return false;	//	message ignored
	}
}


static void EnableStereoFields(
	HWND	aDialog,
	bool	aStereoFlag)
{
	EnableWindow(GetDlgItem(aDialog, IDC_PREFS_PERSPECTIVE_EYE_SPACING_LABEL),		aStereoFlag);
	EnableWindow(GetDlgItem(aDialog, IDC_PREFS_PERSPECTIVE_EYE_SPACING),			aStereoFlag);
	EnableWindow(GetDlgItem(aDialog, IDC_PREFS_PERSPECTIVE_EYE_SPACING_UNIT),		aStereoFlag);
}


static void FormatUnsignedInteger32(
	WCHAR			*aBuffer,
	unsigned int	aBufferLength,
	unsigned int	aNumber)
{
	unsigned int	theNumDigits;
	WCHAR			theString[10];

	//	Format aNumber into local memory, with digits reversed.
	//	At most ten decimal digits will be required for a 32-bit unsigned integer.
	theNumDigits = 0;
	while (aNumber)
	{
		theString[theNumDigits] = (WCHAR)('0' + aNumber % 10);
		aNumber /= 10;
		theNumDigits++;
	}

	//	Logically zero is a 0-digit number,
	//	but let's make a special case an write it as a 1-digit number.
	if (theNumDigits == 0)
	{
		theString[0] = (WCHAR) '0';
		theNumDigits = 1;
	}

	//	Is the input buffer OK?
	if (aBuffer == NULL)
	{
		ErrorMessage("NULL buffer passed to FormatUnsignedInteger32().", "Bad Parameters");
		return;
	}
	if (aBufferLength < theNumDigits + 1)
	{
		ErrorMessage("Buffer too small in FormatUnsignedInteger32().", "Bad Parameters");
		return;
	}

	//	Copy the digits from theString into aBuffer, suppressing leading zeros.
	while (theNumDigits-- > 0)
		*aBuffer++ = theString[theNumDigits];
	*aBuffer = 0;	//	write a terminating 0
}


static void UnformatUnsignedInteger32(
	WCHAR			*aBuffer,
	unsigned int	aBufferLength,
	unsigned int	*aNumber)
{
	WCHAR	*theEnd;	//	will point to the first character past the end of aBuffer

	//	Initialize.
	*aNumber = 0;
	theEnd = aBuffer + aBufferLength;

	//	Scan past any initial white space and/or other junk.
	//	A negative sign is also intentionally ignored.
	while (aBuffer < theEnd
		&& (*aBuffer < ((WCHAR)'0') || *aBuffer > ((WCHAR)'9')))
	{
		aBuffer++;
	}

	//	Read the digits, if any.
	while (aBuffer < theEnd
		&& (*aBuffer >= ((WCHAR)'0') && *aBuffer <= ((WCHAR)'9')))
	{
		*aNumber *= 10;
		*aNumber += *aBuffer - ((WCHAR)'0');
		aBuffer++;
	}
}


static INT_PTR CALLBACK PrefsAdvancedDialogProc(
	HWND	hDlg,
	UINT	msg,
	WPARAM	wParam,
	LPARAM	lParam)
{
	WindowData	*wd;
	WCHAR		theLabel[256];

	switch (msg)
	{
		case WM_INITDIALOG:

			//	lParam contains a pointer to the parent window's WindowData.
			//	Stash a copy in this dialog box as well.
			wd = (WindowData *) lParam;
			SetWindowLongPtr(hDlg, GWLP_USERDATA, (LONG_PTR) wd);
			if (wd == NULL)	//	should never occur
				return true;	//	set focus normally

			//	Initialize the checkboxes.

			SendMessage(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_FRAME_RATE),
							BM_SETCHECK,
							gShowFrameRate ? BST_CHECKED : BST_UNCHECKED,
							0);

			EnableWindow(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_SYNC_TO_MONITOR),
							wglSwapInterval != NULL);	//	wglSwapInterval may be non-NULL but ineffective
			SendMessage(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_SYNC_TO_MONITOR),
							BM_SETCHECK,
							(	wglSwapInterval
							 && wd->md.itsSyncFlag	)
								? BST_CHECKED : BST_UNCHECKED,
							0);

			EnableWindow(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_MULTISAMPLING),
							wd->md.itsMultisamplingAvailableFlag);
			SendMessage(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_MULTISAMPLING),
							BM_SETCHECK,
							(	wd->md.itsMultisamplingAvailableFlag
							 && wd->md.itsMultisamplingEnabledFlag	)
								? BST_CHECKED : BST_UNCHECKED,
							0);

			EnableWindow(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_SHADERS),
							wd->md.itsShadersAvailableFlag);
			SendMessage(	GetDlgItem(hDlg, IDC_PREFS_ADVANCED_SHADERS),
							BM_SETCHECK,
							(	wd->md.itsShadersAvailableFlag
							 && wd->md.itsShadersEnabledFlag	)
								? BST_CHECKED : BST_UNCHECKED,
							0);

			//	Initialize the labels in the current language.

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_ADVANCED_FRAME_RATE),
				GetLocalizedTextAsWCHAR("PrefFrameRate",     theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_ADVANCED_SYNC_TO_MONITOR),
				GetLocalizedTextAsWCHAR("PrefSync",          theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_ADVANCED_MULTISAMPLING),
				GetLocalizedTextAsWCHAR("PrefMultisampling", theLabel, BUFFER_LENGTH(theLabel)));

			SetWindowText(GetDlgItem(hDlg, IDC_PREFS_ADVANCED_SHADERS),
				GetLocalizedTextAsWCHAR("PrefShaders",       theLabel, BUFFER_LENGTH(theLabel)));

			return true;	//	set focus normally

		case WM_COMMAND:

			wd = (WindowData *) GetWindowLongPtr(hDlg, GWLP_USERDATA);
			if (wd != NULL)
			{
				switch (LOWORD(wParam))
				{
					case IDC_PREFS_ADVANCED_FRAME_RATE:
						gShowFrameRate = ! gShowFrameRate;
						EnumThreadWindows(	GetCurrentThreadId(),
											SetWindowTitle,
											(LPARAM) (gShowFrameRate ? L"…" : NULL));
						break;

					case IDC_PREFS_ADVANCED_SYNC_TO_MONITOR:
						wd->md.itsSyncFlag = ! wd->md.itsSyncFlag;
						SetSync(wd);
						break;

					case IDC_PREFS_ADVANCED_MULTISAMPLING:
						SetMultisamplingEnabledFlag(&wd->md, ! wd->md.itsMultisamplingEnabledFlag );
						DisplayChange(wd);
						break;

					case IDC_PREFS_ADVANCED_SHADERS:
						SetShadersEnabledFlag(&wd->md, ! wd->md.itsShadersEnabledFlag );
						break;
				}
				PaintWindow(wd);
			}
			SetWindowLong(hDlg, DWL_MSGRESULT, 0);	//	return value
			return true;	//	message processed

		default:
			return false;	//	message ignored
	}
}


static INT_PTR CALLBACK SymmetryDialogProc(
	HWND	hDlg,
	UINT	msg,
	WPARAM	wparam,
	LPARAM	lParam)
{
	WindowData		*wd;
	HWND			theEditField;
	WCHAR			theBuffer[64],
					theMenuChoice[2];
	unsigned int	theNumber[3],
					i,
					j;

	switch (msg)
	{
		case WM_INITDIALOG:

			//	lParam will contain a pointer to the parent window's WindowData.
			//	Stash a copy in the dialog box as well.
			wd = (WindowData *) lParam;
			SetWindowLongPtr(hDlg, GWLP_USERDATA, (LONG_PTR) wd);
			if (wd == NULL)	//	should never occur
				return true;

			//	Set the dialog's title and buttons in the current language.
			SetWindowText(hDlg,
				GetLocalizedTextAsWCHAR("SymmetryTitle",  theBuffer, BUFFER_LENGTH(theBuffer)));
			SetWindowText(GetDlgItem(hDlg, IDOK),
				GetLocalizedTextAsWCHAR("SymmetryOK",     theBuffer, BUFFER_LENGTH(theBuffer)));
			SetWindowText(GetDlgItem(hDlg, IDCANCEL),
				GetLocalizedTextAsWCHAR("SymmetryCancel", theBuffer, BUFFER_LENGTH(theBuffer)));

			//	Copy our internal variables into the dialog's edit fields.
			for (i = 0; i < 3; i++)
			{
				theEditField = GetDlgItem(hDlg, IDC_SYMMETRY_BASE + i);
				for (j = 2; j <= 9; j++)
				{
					theMenuChoice[0] = L'0' + j;
					theMenuChoice[1] = 0;
					SendMessage(theEditField, CB_ADDSTRING, 0, (LPARAM)theMenuChoice);
				}
				SendMessage(theEditField, CB_SETCURSEL, wd->md.itsTiling.itsPQR[i] - 2, 0);
			}

			return true;

		case WM_COMMAND:

			switch (LOWORD(wparam))
			{
				case IDOK:

					wd = (WindowData *) GetWindowLongPtr(hDlg, GWLP_USERDATA);

					if (wd != NULL)
					{
						//	Copy the dialog's edit fields back into our internal variables.
						for (i = 0; i < 3; i++)
						{
							theEditField = GetDlgItem(hDlg, IDC_SYMMETRY_BASE + i);
							theNumber[i] = 2 + SendMessage(theEditField, CB_GETCURSEL, 0, 0);

							//	Shouldn't be necessary, but just to be safe...
							if (theNumber[i] < 2)
								theNumber[i] = 2;
							if (theNumber[i] > 9)
								theNumber[i] = 9;
						}

						//	Replace the tiling.
						ChangeTiling(&wd->md, theNumber[0], theNumber[1], theNumber[2]);
					}

					//	Dismiss the dialog.
					EndDialog(hDlg, true);

					return true;

				case IDCANCEL:
					EndDialog(hDlg, true);
					return true;

				default:
					return false;
			}

		default:
			return false;
	}
}


static void SetSync(WindowData *wd)
{
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	if (wglSwapInterval != NULL)	//	wglSwapInterval may be non-NULL but ineffective
		wglSwapInterval(wd->md.itsSyncFlag ? 1 : 0);
	wglMakeCurrent(NULL, NULL);
}


static void SetMinWindowSize(
	HWND	aWindow,
	POINT	*aMinSize)
{
	RECT	theRect;

	// Prevent the client area from going smaller than MIN_CLIENT_SIZE.
	SetRect(&theRect, 0, 0, MIN_CLIENT_SIZE, MIN_CLIENT_SIZE);
	AdjustWindowRect(&theRect, (DWORD) GetWindowLongPtr(aWindow, GWL_STYLE), true);
	aMinSize->x = theRect.right - theRect.left;
	aMinSize->y = theRect.bottom - theRect.top;
}


static void ResizeDrawingPanel(
	WindowData		*wd,
	unsigned int	aWidth,
	unsigned int	aHeight)
{
	//	When SetUpDrawingPanel() creates a temporary dummy window,
	//	the CreateWindow() call sends a WM_SIZE message which leads here.
	//	Ignore this unwanted call, so we don't screw up the window measurements.
	if (wd->itsDeviceContext == NULL || wd->itsRenderingContext == NULL)
		return;
	
	//	Let OpenGL resize the viewport.
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	ResizeGLView(&wd->md, aWidth, aHeight);
	wglMakeCurrent(NULL, NULL);

	//	Request a refresh.
	InvalidateRect(wd->itsDrawingPanel, NULL, false);
}


static void DisplayChange(WindowData *wd)
{
	//	Maybe the user changed the color depth, screen size or multisampling,
	//	or maybe it's just that the screen saver had been running
	//	and is now passing control back to our program.
	//	Either way, we need to refresh our OpenGL context, shaders,
	//	textures and vertex buffer objects.

	//	Free the old OpenGL context along with its shaders, textures
	//	and vertex buffer objects.
	ShutDownOpenGL(wd);

	//	Create a new OpenGL context with new shaders, textures
	//	and vertex buffer objects.
	SetUpOpenGL(wd);
}


static void GetMouseLocation(
	ModelData	*md,
	HWND		aDrawingWindow,
	LPARAM		lParam,			//	input location in drawing panel coordinates
	double		*theMouseH,		//	output horizontal coordinate ∈ [0,1] (directed rightward)
	double		*theMouseV)		//	output  vertical  coordinate ∈ [0,1] (directed upward)
{
	signed short int	hh,
						vv;
	RECT				theRect;

	//	Split lParam into its horizontal and vertical parts.
	hh = (signed short int) LOWORD(lParam);
	vv = (signed short int) HIWORD(lParam);

	//	Fetch the client rectangle.
	GetClientRect(aDrawingWindow, &theRect);

	//	Compute coordinates in range [0,1] with origin at the lower left corner.
	*theMouseH = (double)( hh - theRect.left ) / (double)(theRect.right - theRect.left);
	*theMouseV = (double)(theRect.bottom - vv) / (double)(theRect.bottom - theRect.top);
	
	//	Adjust for side-by-side images.
	if (md->itsStereoMode == StereoGeoWall)
	{
		if (*theMouseH <= 0.5)	//	map [0.0, 0.5] → [0.0, 1.0]
			*theMouseH *= 2.0;
		else					//	map [0.5, 1.0] → [0.0, 1.0]
			*theMouseH = 2.0*(*theMouseH) - 1.0;
	}
}


static void MaintainCursor(
	double		aMouseH,	//	horizontal mouse coordinate in range [0,1]
							//		already adjusted for itsStereoMode == StereoGeoWall
	WPARAM		wParam)
{
	unsigned int	theCursorID;

	if (aMouseH < 0.8) 
	{
		if (wParam & MK_LBUTTON)
			theCursorID = IDU_CURSOR_GRAB;
		else
			theCursorID = IDU_CURSOR_FLAT;
	}
	else
	{
		//	Ignore the "touching cursor" and use only the "pointing cursor".
//		if (wParam & MK_LBUTTON)
//			theCursorID = IDU_CURSOR_TOUCH;
//		else
			theCursorID = IDU_CURSOR_POINT;
	}

	SetCursor(LoadCursor(GetModuleHandle(NULL), (void *) theCursorID));
}


static void PaintWindow(WindowData *wd)
{
	ErrorText	theSetUpError				= NULL,
				theRenderError				= NULL;
	HDC			theSavedDeviceContext		= NULL;
	HGLRC		theSavedRenderingContext	= NULL;

	//	In the normal flow of events there's no need to save and restore
	//	the rendering context.  However, we may get "unexpected" WM_PAINT
	//	messages -- for example if a MessageBox takes it upon itself
	//	to send WM_PAINT messages to underlying windows -- so play it safe
	//	and leave the rendering context as we found it.

	theSavedDeviceContext		= wglGetCurrentDC();		//	typically NULL
	theSavedRenderingContext	= wglGetCurrentContext();	//	typically NULL
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
		theSetUpError = SetUpGraphicsAsNeeded(&wd->md);
		if (theSetUpError == NULL)
			theRenderError = Render(&wd->md);
		if (theSetUpError == NULL && theRenderError == NULL)
			SwapBuffers(wd->itsDeviceContext);
	wglMakeCurrent(theSavedDeviceContext, theSavedRenderingContext);

	ValidateRect(wd->itsDrawingPanel, NULL);

	if (theSetUpError != NULL)
	{
		ErrorMessage(theSetUpError, "OpenGL Setup Error");
		ExitProcess(17);
	}
	if (theRenderError != NULL)
	{
		ErrorMessage(theRenderError, "OpenGL Rendering Error");
		ExitProcess(18);
	}
}
